@@ -1,4 +1,5 @@
 #!/usr/bin/python3
+# SPDX-License-Identifier: GPL-3.0-or-later
 import json
 import time
 import sys
@@ -36,12 +37,14 @@ while time.time() < end_time:
         except KeyError:
             pass
 
-        if state == "errored":
-            exit("Travis CI Result: ERRORED!")
-        elif state == "passed":
+        if state == "passed":
             exit("Travis CI Result: PASSED!", code=0)
+        elif state == "created" or state == "started":
+            pass
+        else:
+            exit("Travis CI Result: {}!".format(state.upper()))
     else:
-        exit("API Response Code: {code}".format(response.status_code), code=2)
+        exit("API Response Code: {}".format(response.status_code), code=2)
     time.sleep(POLL_DELAY)
 
 exit("Timed out!")
