@@ -1,4 +1,6 @@
 -- unload modules which are not related to this test
+-- SPDX-License-Identifier: GPL-3.0-or-later
+
 if ta_signal_query then
         modules.unload('ta_signal_query')
 end
@@ -31,8 +33,8 @@ event.cancel(ev)
 ev = event.after(0, function () return 1 end)
 
 
--- Import fake root zone; avoid interference with configured KEYFILE_DEFAULT.
-trust_anchors.keyfile_default = nil
+-- Import fake root zone; avoid interference with configured keyfile_default.
+trust_anchors.remove('.')
 trust_anchors.add('. IN DS 48409 8 2 3D63A0C25BCE86621DE63636F11B35B908EFE8E9381E0E3E9DEFD89EA952C27D')
 
 local function check_answer(desc, qname, qtype, expected_rcode)
