@@ -1,22 +1,11 @@
 /*  Copyright (C) 2015-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 #include <contrib/cleanup.h>
 #include <ccan/json/json.h>
 #include <ccan/asprintf/asprintf.h>
+#include <dlfcn.h>
 #include <uv.h>
 #include <unistd.h>
 #include <grp.h>
@@ -24,8 +13,13 @@
 #include <sys/param.h>
 #include <libzscanner/scanner.h>
 
+#include <lua.h>
+#include <lualib.h>
+#include <lauxlib.h>
+#include "daemon/bindings/impl.h"
+
+#include "kresconfig.h"
 #include "daemon/engine.h"
-#include "daemon/bindings.h"
 #include "daemon/ffimodule.h"
 #include "lib/nsrep.h"
 #include "lib/cache/api.h"
@@ -48,41 +42,26 @@
 	#endif
 #endif
 
-/** @internal Compatibility wrapper for Lua < 5.2 */
-#if LUA_VERSION_NUM < 502
-#define lua_rawlen(L, obj) lua_objlen((L), (obj))
-#endif
-
 /**@internal Maximum number of incomplete TCP connections in queue.
-* Default is from Redis and Apache. */
+* Default is from empirical testing - in our case, more isn't necessarily better.
+* See https://gitlab.labs.nic.cz/knot/knot-resolver/-/merge_requests/968
+* */
 #ifndef TCP_BACKLOG_DEFAULT
-#define TCP_BACKLOG_DEFAULT 511
+#define TCP_BACKLOG_DEFAULT 128
 #endif
 
-/** @internal Annotate for static checkers. */
-KR_NORETURN int lua_error (lua_State *L);
-
 /* Cleanup engine state every 5 minutes */
 const size_t CLEANUP_TIMER = 5*60*1000;
 
 /* Execute byte code */
 #define l_dobytecode(L, arr, len, name) \
 	(luaL_loadbuffer((L), (arr), (len), (name)) || lua_pcall((L), 0, LUA_MULTRET, 0))
-/** Load file in a sandbox environment. */
-#define l_dosandboxfile(L, filename) \
-	(luaL_loadfile((L), (filename)) || engine_pcall((L), 0))
 
 /*
  * Global bindings.
  */
+struct args *the_args;
 
-/** Register module callback into Lua world. */
-#define REGISTER_MODULE_CALL(L, module, cb, name) do { \
-	lua_pushlightuserdata((L), (module)); \
-	lua_pushlightuserdata((L), (cb)); \
-	lua_pushcclosure((L), l_trampoline, 2); \
-	lua_setfield((L), -2, (name)); \
-	} while (0)
 
 /** Print help and available commands. */
 static int l_help(lua_State *L)
@@ -100,7 +79,6 @@ static int l_help(lua_State *L)
 		"resolve(name, type[, class, flags, callback])\n    resolve query, callback when it's finished\n"
 		"todname(name)\n    convert name to wire format\n"
 		"tojson(val)\n    convert value to JSON\n"
-		"map(expr)\n    run expression on all workers\n"
 		"net\n    network configuration\n"
 		"cache\n    network configuration\n"
 		"modules\n    modules configuration\n"
@@ -130,31 +108,25 @@ static bool update_privileges(int uid, int gid)
 static int l_setuser(lua_State *L)
 {
 	int n = lua_gettop(L);
-	if (n < 1 || !lua_isstring(L, 1)) {
-		lua_pushliteral(L, "user(user[, group)");
-		lua_error(L);
-	}
+	if (n < 1 || !lua_isstring(L, 1))
+		lua_error_p(L, "user(user[, group])");
+
 	/* Fetch UID/GID based on string identifiers. */
 	struct passwd *user_pw = getpwnam(lua_tostring(L, 1));
-	if (!user_pw) {
-		lua_pushliteral(L, "invalid user name");
-		lua_error(L);
-	}
+	if (!user_pw)
+		lua_error_p(L, "invalid user name");
 	int uid = user_pw->pw_uid;
 	int gid = getgid();
 	if (n > 1 && lua_isstring(L, 2)) {
 		struct group *group_pw = getgrnam(lua_tostring(L, 2));
-		if (!group_pw) {
-			lua_pushliteral(L, "invalid group name");
-			lua_error(L);
-		}
+		if (!group_pw)
+			lua_error_p(L, "invalid group name");
 		gid = group_pw->gr_gid;
 	}
 	/* Drop privileges */
 	bool ret = update_privileges(uid, gid);
 	if (!ret) {
-		lua_pushstring(L, strerror(errno));
-		lua_error(L);
+		lua_error_maybe(L, errno);
 	}
 	lua_pushboolean(L, ret);
 	return 1;
@@ -217,17 +189,13 @@ static int l_hostname(lua_State *L)
 		lua_pushstring(L, engine_get_hostname(engine));
 		return 1;
 	}
-	if ((lua_gettop(L) != 1) || !lua_isstring(L, 1)) {
-		lua_pushstring(L, "hostname takes at most one parameter: (\"fqdn\")");
-		lua_error(L);
-	}
+	if ((lua_gettop(L) != 1) || !lua_isstring(L, 1))
+		lua_error_p(L, "hostname takes at most one parameter: (\"fqdn\")");
 
-	if (engine_set_hostname(engine, lua_tostring(L, 1)) != 0) {
-		lua_pushstring(L, "setting hostname failed");
-		lua_error(L);
-	}
+	if (engine_set_hostname(engine, lua_tostring(L, 1)) != 0)
+		lua_error_p(L, "setting hostname failed");
 
-	lua_pushstring(L, engine_get_hostname(engine));	
+	lua_pushstring(L, engine_get_hostname(engine));
 	return 1;
 }
 
@@ -238,120 +206,6 @@ static int l_package_version(lua_State *L)
 	return 1;
 }
 
-char *engine_get_moduledir(struct engine *engine) {
-	return engine->moduledir;
-}
-
-int engine_set_moduledir(struct engine *engine, const char *moduledir) {
-	if (!engine || !moduledir) {
-		return kr_error(EINVAL);
-	}
-
-	char *new_moduledir = strdup(moduledir);
-	if (!new_moduledir) {
-		return kr_error(ENOMEM);
-	}
-	if (engine->moduledir) {
-		free(engine->moduledir);
-	}
-	engine->moduledir = new_moduledir;
-
-	/* Use module path for including Lua scripts */
-	char l_paths[MAXPATHLEN] = { 0 };
-	#pragma GCC diagnostic push
-	#pragma GCC diagnostic ignored "-Wformat" /* %1$ is not in C standard */
-	/* Save original package.path to package._path */
-	snprintf(l_paths, MAXPATHLEN - 1,
-		 "if package._path == nil then package._path = package.path end\n"
-		 "package.path = '%1$s/?.lua;%1$s/?/init.lua;'..package._path\n"
-		 "if package._cpath == nil then package._cpath = package.cpath end\n"
-		 "package.cpath = '%1$s/?%2$s;'..package._cpath\n",
-		 new_moduledir, LIBEXT);
-	#pragma GCC diagnostic pop
-
-	int ret = l_dobytecode(engine->L, l_paths, strlen(l_paths), "");
-	if (ret != 0) {
-		lua_pop(engine->L, 1);
-		return ret;
-	}
-	return 0;
-}
-
-/** Return hostname. */
-static int l_moduledir(lua_State *L)
-{
-	struct engine *engine = engine_luaget(L);
-	if (lua_gettop(L) == 0) {
-		lua_pushstring(L, engine_get_moduledir(engine));
-		return 1;
-	}
-	if ((lua_gettop(L) != 1) || !lua_isstring(L, 1)) {
-		lua_pushstring(L, "moduledir takes at most one parameter: (\"directory\")");
-		lua_error(L);
-	}
-
-	if (engine_set_moduledir(engine, lua_tostring(L, 1)) != 0) {
-		lua_pushstring(L, "setting moduledir failed");
-		lua_error(L);
-	}
-
-	lua_pushstring(L, engine_get_moduledir(engine));
-	return 1;
-}
-
-/** @internal for l_trustanchor: */
-static void ta_add(zs_scanner_t *zs)
-{
-	map_t *ta = zs->process.data;
-	if (!ta)
-		return;
-	if (kr_ta_add(ta, zs->r_owner, zs->r_type, zs->r_ttl, zs->r_data, zs->r_data_length))
-		zs->process.data = NULL; /* error signalling */
-}
-/** Enable/disable trust anchor. */
-static int l_trustanchor(lua_State *L)
-{
-	struct engine *engine = engine_luaget(L);
-	const char *anchor = lua_tostring(L, 1);
-	bool enable = lua_isboolean(L, 2) ? lua_toboolean(L, 2) : true;
-	if (!anchor || strlen(anchor) == 0) {
-		return 0;
-	}
-	/* If disabling, parse the owner string only. */
-	if (!enable) {
-		knot_dname_t *owner = knot_dname_from_str(NULL, anchor, KNOT_DNAME_MAXLEN);
-		if (!owner) {
-			lua_pushstring(L, "invalid trust anchor owner");
-			lua_error(L);
-		}
-		lua_pushboolean(L, kr_ta_del(&engine->resolver.trust_anchors, owner) == 0);
-		free(owner);
-		return 1;
-	}
-
-	/* Parse the record */
-	zs_scanner_t *zs = malloc(sizeof(*zs));
-	if (!zs || zs_init(zs, ".", 1, 0) != 0) {
-		free(zs);
-		lua_pushstring(L, "not enough memory");
-		lua_error(L);
-	}
-	zs_set_processing(zs, ta_add, NULL, &engine->resolver.trust_anchors);
-	bool ok = zs_set_input_string(zs, anchor, strlen(anchor)) == 0
-		&& zs_parse_all(zs) == 0;
-	ok = ok && zs->process.data; /* reset to NULL on error in ta_add */
-
-	zs_deinit(zs);
-	free(zs);
-	/* Report errors */
-	if (!ok) {
-		lua_pushstring(L, "failed to process trust anchor RR");
-		lua_error(L);
-	}
-	lua_pushboolean(L, true);
-	return 1;
-}
-
 /** Load root hints from zonefile. */
 static int l_hint_root_file(lua_State *L)
 {
@@ -364,8 +218,7 @@ static int l_hint_root_file(lua_State *L)
 		if (!file) {
 			file = ROOTHINTS;
 		}
-		lua_push_printf(L, "error when opening '%s': %s", file, err);
-		lua_error(L);
+		lua_error_p(L, "error when opening '%s': %s", file, err);
 	} else {
 		lua_pushboolean(L, true);
 		return 1;
@@ -436,7 +289,7 @@ static void l_unpack_json(lua_State *L, JsonNode *table)
 		if (node->key) {
 			lua_setfield(L, -2, node->key);
 		} else {
-			lua_rawseti(L, -2, lua_rawlen(L, -2) + 1);
+			lua_rawseti(L, -2, lua_objlen(L, -2) + 1);
 		}
 	}
 }
@@ -504,18 +357,14 @@ static int l_tojson(lua_State *L)
 
 static int l_fromjson(lua_State *L)
 {
-	if (lua_gettop(L) != 1 || !lua_isstring(L, 1)) {
-		lua_pushliteral(L, "a JSON string is required");
-		lua_error(L);
-	}
+	if (lua_gettop(L) != 1 || !lua_isstring(L, 1))
+		lua_error_p(L, "a JSON string is required");
 
 	const char *json_str = lua_tostring(L, 1);
 	JsonNode *root_node = json_decode(json_str);
 
-	if (!root_node) {
-		lua_pushliteral(L, "invalid JSON string");
-		lua_error(L);
-	}
+	if (!root_node)
+		lua_error_p(L, "invalid JSON string");
 	l_unpack_json(L, root_node);
 	json_delete(root_node);
 
@@ -524,14 +373,15 @@ static int l_fromjson(lua_State *L)
 
 /** @internal Throw Lua error if expr is false */
 #define expr_checked(expr) \
-	if (!(expr)) { lua_pushboolean(L, false); lua_rawseti(L, -2, lua_rawlen(L, -2) + 1); continue; }
+	if (!(expr)) { lua_pushboolean(L, false); lua_rawseti(L, -2, lua_objlen(L, -2) + 1); continue; }
 
 static int l_map(lua_State *L)
 {
-	if (lua_gettop(L) != 1 || !lua_isstring(L, 1)) {
-		lua_pushliteral(L, "map('string with a lua expression')");
-		lua_error(L);
-	}
+	/* We don't kr_log_deprecate() here for now.  Plan: after --forks gets *removed*,
+	 * kill internal uses of map() (e.g. from daf module) and add deprecation here.
+	 * Alternatively we might (attempt to) implement map() in another way. */
+	if (lua_gettop(L) != 1 || !lua_isstring(L, 1))
+		lua_error_p(L, "map('string with a lua expression')");
 
 	struct engine *engine = engine_luaget(L);
 	const char *cmd = lua_tostring(L, 1);
@@ -565,12 +415,12 @@ static int l_map(lua_State *L)
 				lua_pushlstring(L, rbuf, rlen);
 			}
 			json_delete(root_node);
-			lua_rawseti(L, -2, lua_rawlen(L, -2) + 1);
+			lua_rawseti(L, -2, lua_objlen(L, -2) + 1);
 			continue;
 		}
 		/* Didn't respond */
 		lua_pushboolean(L, false);
-		lua_rawseti(L, -2, lua_rawlen(L, -2) + 1);
+		lua_rawseti(L, -2, lua_objlen(L, -2) + 1);
 	}
 	return 1;
 }
@@ -578,55 +428,6 @@ static int l_map(lua_State *L)
 #undef expr_checked
 
 
-/** Trampoline function for module properties. */
-static int l_trampoline(lua_State *L)
-{
-	struct kr_module *module = lua_touserdata(L, lua_upvalueindex(1));
-	void* callback = lua_touserdata(L, lua_upvalueindex(2));
-	struct engine *engine = engine_luaget(L);
-	if (!module) {
-		lua_pushstring(L, "module closure missing upvalue");
-		lua_error(L);
-	}
-
-	/* Now we only have property callback or config,
-	 * if we expand the callables, we might need a callback_type.
-	 */
-	const char *args = NULL;
-	auto_free char *cleanup_args = NULL;
-	if (lua_gettop(L) > 0) {
-		if (lua_istable(L, 1) || lua_isboolean(L, 1)) {
-			cleanup_args = l_pack_json(L, 1);
-			args = cleanup_args;
-		} else {
-			args = lua_tostring(L, 1);
-		}
-	}
-	#pragma GCC diagnostic push
-	#pragma GCC diagnostic ignored "-Wpedantic" /* void* vs. function pointer */
-	if (callback == module->config) {
-		module->config(module, args);
-	} else {
-		kr_prop_cb *prop = (kr_prop_cb *)callback;
-		auto_free char *ret = prop(engine, module, args);
-		if (!ret) { /* No results */
-			return 0;
-		}
-		JsonNode *root_node = json_decode(ret);
-		if (root_node) {
-			l_unpack_json(L, root_node);
-		} else {
-			lua_pushstring(L, ret);
-		}
-		json_delete(root_node);
-		return 1;
-	}
-	#pragma GCC diagnostic pop
-
-	/* No results */
-	return 0;
-}
-
 /*
  * Engine API.
  */
@@ -648,12 +449,12 @@ static int init_resolver(struct engine *engine)
 	knot_edns_init(engine->resolver.opt_rr, KR_EDNS_PAYLOAD, 0, KR_EDNS_VERSION, engine->pool);
 	/* Use default TLS padding */
 	engine->resolver.tls_padding = -1;
-	/* Empty init; filled via ./lua/config.lua */
+	/* Empty init; filled via ./lua/postconfig.lua */
 	kr_zonecut_init(&engine->resolver.root_hints, (const uint8_t *)"", engine->pool);
 	/* Open NS rtt + reputation cache */
-	lru_create(&engine->resolver.cache_rtt, LRU_RTT_SIZE, engine->pool, NULL);
-	lru_create(&engine->resolver.cache_rep, LRU_REP_SIZE, engine->pool, NULL);
-	lru_create(&engine->resolver.cache_cookie, LRU_COOKIES_SIZE, engine->pool, NULL);
+	lru_create(&engine->resolver.cache_rtt, LRU_RTT_SIZE, NULL, NULL);
+	lru_create(&engine->resolver.cache_rep, LRU_REP_SIZE, NULL, NULL);
+	lru_create(&engine->resolver.cache_cookie, LRU_COOKIES_SIZE, NULL, NULL);
 
 	/* Load basic modules */
 	engine_register(engine, "iterate", NULL, NULL);
@@ -671,7 +472,6 @@ static int init_state(struct engine *engine)
 		return kr_error(ENOMEM);
 	}
 	/* Initialize used libraries. */
-	lua_gc(engine->L, LUA_GCSTOP, 0);
 	luaL_openlibs(engine->L);
 	/* Global functions */
 	lua_pushcfunction(engine->L, l_help);
@@ -682,14 +482,10 @@ static int init_state(struct engine *engine)
 	lua_setglobal(engine->L, "hostname");
 	lua_pushcfunction(engine->L, l_package_version);
 	lua_setglobal(engine->L, "package_version");
-	lua_pushcfunction(engine->L, l_moduledir);
-	lua_setglobal(engine->L, "moduledir");
 	lua_pushcfunction(engine->L, l_verbose);
 	lua_setglobal(engine->L, "verbose");
 	lua_pushcfunction(engine->L, l_setuser);
 	lua_setglobal(engine->L, "user");
-	lua_pushcfunction(engine->L, l_trustanchor);
-	lua_setglobal(engine->L, "trustanchor");
 	lua_pushcfunction(engine->L, l_hint_root_file);
 	lua_setglobal(engine->L, "_hint_root_file");
 	lua_pushliteral(engine->L, libknot_SONAME);
@@ -704,6 +500,13 @@ static int init_state(struct engine *engine)
 	lua_setglobal(engine->L, "map");
 	lua_pushlightuserdata(engine->L, engine);
 	lua_setglobal(engine->L, "__engine");
+	/* Random number generator */
+	lua_getfield(engine->L, LUA_GLOBALSINDEX, "math");
+	lua_getfield(engine->L, -1, "randomseed");
+	lua_remove(engine->L, -2);
+	lua_Number seed = kr_rand_bytes(sizeof(lua_Number));
+	lua_pushnumber(engine->L, seed);
+	lua_call(engine->L, 1, 0);
 	return kr_ok();
 }
 
@@ -727,7 +530,7 @@ static void init_measurement(struct engine *engine)
 		"})\n"
 		"jit.off()\n", statspath
 	);
-	assert(ret > 0);
+	assert(ret > 0); (void)ret;
 
 	ret = luaL_loadstring(engine->L, snippet);
 	assert(ret == 0);
@@ -735,6 +538,33 @@ static void init_measurement(struct engine *engine)
 	free(snippet);
 }
 
+int init_lua(struct engine *engine) {
+	if (!engine) {
+		return kr_error(EINVAL);
+	}
+
+	/* Use libdir path for including Lua scripts */
+	char l_paths[MAXPATHLEN] = { 0 };
+	#pragma GCC diagnostic push
+	#pragma GCC diagnostic ignored "-Wformat" /* %1$ is not in C standard */
+	/* Save original package.path to package._path */
+	snprintf(l_paths, MAXPATHLEN - 1,
+		 "if package._path == nil then package._path = package.path end\n"
+		 "package.path = '%1$s/?.lua;%1$s/?/init.lua;'..package._path\n"
+		 "if package._cpath == nil then package._cpath = package.cpath end\n"
+		 "package.cpath = '%1$s/?%2$s;'..package._cpath\n",
+		 LIBDIR, LIBEXT);
+	#pragma GCC diagnostic pop
+
+	int ret = l_dobytecode(engine->L, l_paths, strlen(l_paths), "");
+	if (ret != 0) {
+		lua_pop(engine->L, 1);
+		return ret;
+	}
+	return 0;
+}
+
+
 int engine_init(struct engine *engine, knot_mm_t *pool)
 {
 	if (engine == NULL) {
@@ -760,16 +590,23 @@ int engine_init(struct engine *engine, knot_mm_t *pool)
 	/* Initialize network */
 	network_init(&engine->net, uv_default_loop(), TCP_BACKLOG_DEFAULT);
 
+	/* Initialize lua */
+	ret = init_lua(engine);
+	if (ret != 0) {
+		engine_deinit(engine);
+		return ret;
+	}
+
 	return ret;
 }
 
+/** Unregister a (found) module */
 static void engine_unload(struct engine *engine, struct kr_module *module)
 {
-	/* Unregister module */
-	auto_free char *name = strdup(module->name);
-	kr_module_unload(module);
+	auto_free char *name = module->name ? strdup(module->name) : NULL;
+	kr_module_unload(module); /* beware: lua/C mix, could be confusing */
 	/* Clear in Lua world, but not for embedded modules ('cache' in particular). */
-	if (name && !kr_module_embedded(name)) {
+	if (name && !kr_module_get_embedded(name)) {
 		lua_pushnil(engine->L);
 		lua_setglobal(engine->L, name);
 	}
@@ -781,30 +618,34 @@ void engine_deinit(struct engine *engine)
 	if (engine == NULL) {
 		return;
 	}
+	if (!engine->L) {
+		assert(false);
+		return;
+	}
+	/* Only close sockets and services; no need to clean up mempool. */
 
-	/* Only close sockets and services,
-	 * no need to clean up mempool. */
-	network_deinit(&engine->net);
+	/* Network deinit is split up.  We first need to stop listening,
+	 * then we can unload modules during which we still want
+	 * e.g. the endpoint kind registry to work (inside ->net),
+	 * and this registry deinitization uses the lua state. */
+	network_close_force(&engine->net);
+	for (size_t i = 0; i < engine->ipc_set.len; ++i) {
+		close(engine->ipc_set.at[i]);
+	}
+	for (size_t i = 0; i < engine->modules.len; ++i) {
+		engine_unload(engine, engine->modules.at[i]);
+	}
 	kr_zonecut_deinit(&engine->resolver.root_hints);
 	kr_cache_close(&engine->resolver.cache);
 
-	/* The lru keys are currently malloc-ated and need to be freed. */
+	/* The LRUs are currently malloc-ated and need to be freed. */
 	lru_free(engine->resolver.cache_rtt);
 	lru_free(engine->resolver.cache_rep);
 	lru_free(engine->resolver.cache_cookie);
 
-	/* Clear IPC pipes */
-	for (size_t i = 0; i < engine->ipc_set.len; ++i) {
-		close(engine->ipc_set.at[i]);
-	}
-
-	/* Unload modules and engine. */
-	for (size_t i = 0; i < engine->modules.len; ++i) {
-		engine_unload(engine, engine->modules.at[i]);
-	}
-	if (engine->L) {
-		lua_close(engine->L);
-	}
+	network_deinit(&engine->net);
+	ffimodule_deinit(engine->L);
+	lua_close(engine->L);
 
 	/* Free data structures */
 	array_clear(engine->modules);
@@ -813,15 +654,10 @@ void engine_deinit(struct engine *engine)
 	kr_ta_clear(&engine->resolver.trust_anchors);
 	kr_ta_clear(&engine->resolver.negative_anchors);
 	free(engine->hostname);
-	free(engine->moduledir);
 }
 
 int engine_pcall(lua_State *L, int argc)
 {
-#if LUA_VERSION_NUM >= 502
-	lua_getglobal(L, "_SANDBOX");
-	lua_setupvalue(L, -(2 + argc), 1);
-#endif
 	return lua_pcall(L, argc, LUA_MULTRET, 0);
 }
 
@@ -859,37 +695,28 @@ int engine_ipc(struct engine *engine, const char *expr)
 int engine_load_sandbox(struct engine *engine)
 {
 	/* Init environment */
-	static const char sandbox_bytecode[] = {
-		#include "daemon/lua/sandbox.inc"
-	};
-	if (l_dobytecode(engine->L, sandbox_bytecode, sizeof(sandbox_bytecode), "init") != 0) {
+	int ret = luaL_dofile(engine->L, LIBDIR "/sandbox.lua");
+	if (ret != 0) {
 		fprintf(stderr, "[system] error %s\n", lua_tostring(engine->L, -1));
 		lua_pop(engine->L, 1);
 		return kr_error(ENOEXEC);
 	}
-	return kr_ok();
+	ret = ffimodule_init(engine->L);
+	return ret;
 }
 
 int engine_loadconf(struct engine *engine, const char *config_path)
 {
 	assert(config_path != NULL);
-	int ret = l_dosandboxfile(engine->L, config_path);
-	if (ret != 0) {
-		fprintf(stderr, "%s\n", lua_tostring(engine->L, -1));
-		lua_pop(engine->L, 1);
-	}
-	return ret;
-}
 
-int engine_load_defaults(struct engine *engine)
-{
-	/* Load defaults */
-	static const char config_bytecode[] = {
-		#include "daemon/lua/config.inc"
-	};
-	int ret = l_dobytecode(engine->L, config_bytecode, sizeof(config_bytecode), "config");
+	char cwd[PATH_MAX];
+	get_workdir(cwd, sizeof(cwd));
+	kr_log_verbose("[system] loading config '%s' (workdir '%s')\n", config_path, cwd);
+
+	int ret = luaL_dofile(engine->L, config_path);
 	if (ret != 0) {
-		fprintf(stderr, "%s\n", lua_tostring(engine->L, -1));
+		fprintf(stderr, "[system] error while loading config: "
+			"%s (workdir '%s')\n", lua_tostring(engine->L, -1), cwd);
 		lua_pop(engine->L, 1);
 	}
 	return ret;
@@ -897,12 +724,8 @@ int engine_load_defaults(struct engine *engine)
 
 int engine_start(struct engine *engine)
 {
-	/* Clean up stack and restart GC */
+	/* Clean up stack */
 	lua_settop(engine->L, 0);
-	lua_gc(engine->L, LUA_GCCOLLECT, 0);
-	lua_gc(engine->L, LUA_GCSETSTEPMUL, 50);
-	lua_gc(engine->L, LUA_GCSETPAUSE, 400);
-	lua_gc(engine->L, LUA_GCRESTART, 0);
 
 	return kr_ok();
 }
@@ -915,38 +738,6 @@ void engine_stop(struct engine *engine)
 	uv_stop(uv_default_loop());
 }
 
-/** Register module properties in Lua environment, if any. */
-static int register_properties(struct engine *engine, struct kr_module *module)
-{
-	#pragma GCC diagnostic push
-	#pragma GCC diagnostic ignored "-Wpedantic" /* casts in lua_pushlightuserdata() */
-	if (!module->config && !module->props) {
-		return kr_ok();
-	}
-	lua_newtable(engine->L);
-	if (module->config != NULL) {
-		REGISTER_MODULE_CALL(engine->L, module, module->config, "config");
-	}
-
-	const struct kr_prop *p = module->props == NULL ? NULL : module->props();
-	for (; p && p->name; ++p) {
-		if (p->cb != NULL) {
-			REGISTER_MODULE_CALL(engine->L, module, p->cb, p->name);
-		}
-	}
-	lua_setglobal(engine->L, module->name);
-	#pragma GCC diagnostic pop
-
-	/* Register module in Lua env */
-	lua_getglobal(engine->L, "modules_register");
-	lua_getglobal(engine->L, module->name);
-	if (engine_pcall(engine->L, 1) != 0) {
-		lua_pop(engine->L, 1);
-	}
-
-	return kr_ok();
-}
-
 /** @internal Find matching module */
 static size_t module_find(module_array_t *mod_list, const char *name)
 {
@@ -964,6 +755,7 @@ static size_t module_find(module_array_t *mod_list, const char *name)
 int engine_register(struct engine *engine, const char *name, const char *precedence, const char* ref)
 {
 	if (engine == NULL || name == NULL) {
+		assert(!EINVAL);
 		return kr_error(EINVAL);
 	}
 	/* Make sure module is unloaded */
@@ -982,24 +774,57 @@ int engine_register(struct engine *engine, const char *name, const char *precede
 	if (!module) {
 		return kr_error(ENOMEM);
 	}
-	module->data = engine;
-	int ret = kr_module_load(module, name, engine->moduledir);
-	/* Load Lua module if not a binary */
-	if (ret == kr_error(ENOENT)) {
+	module->data = engine; /*< some outside modules may still use this value */
+
+	int ret = kr_module_load(module, name, LIBDIR "/kres_modules");
+	if (ret == 0) {
+		/* We have a C module, loaded and init() was called.
+		 * Now we need to prepare the lua side. */
+		lua_State *L = engine->L;
+		lua_getglobal(L, "modules_create_table_for_c");
+		lua_pushpointer(L, module);
+		if (lua_isnil(L, -2)) {
+			/* When loading the three embedded modules, we don't
+			 * have the "modules_*" lua function yet, but fortunately
+			 * we don't need it there.  Let's just check they're embedded.
+			 * TODO: solve this better *without* breaking stuff. */
+			lua_pop(L, 2);
+			if (module->lib != RTLD_DEFAULT) {
+				ret = kr_error(1);
+				lua_pushliteral(L, "missing modules_create_table_for_c()");
+			}
+		} else {
+			ret = engine_pcall(L, 1);
+		}
+		if (ret) {
+			kr_log_error("[system] internal error when loading C module %s: %s\n",
+					module->name, lua_tostring(L, -1));
+			lua_pop(L, 1);
+			assert(false); /* probably not critical, but weird */
+		}
+
+	} else if (ret == kr_error(ENOENT)) {
+		/* No luck with C module, so try to load and .init() lua module. */
 		ret = ffimodule_register_lua(engine, module, name);
+		if (ret != 0) {
+			kr_log_error("[system] failed to load module '%s'\n", name);
+		}
+
 	} else if (ret == kr_error(ENOTSUP)) {
 		/* Print a more helpful message when module is linked against an old resolver ABI. */
-		fprintf(stderr, "[system] module '%s' links to unsupported ABI, please rebuild it\n", name);
+		kr_log_error("[system] module '%s' links to unsupported ABI, please rebuild it\n", name);
 	}
+
 	if (ret != 0) {
-		free(module);
+		engine_unload(engine, module);
 		return ret;
 	}
+
+	/* Push to the right place in engine->modules */
 	if (array_push(engine->modules, module) < 0) {
 		engine_unload(engine, module);
 		return kr_error(ENOMEM);
 	}
-	/* Evaluate precedence operator */
 	if (precedence) {
 		struct kr_module **arr = mod_list->at;
 		size_t emplacement = mod_list->len;
@@ -1017,7 +842,7 @@ int engine_register(struct engine *engine, const char *name, const char *precede
 		}
 	}
 
-	return register_properties(engine, module);
+	return kr_ok();
 }
 
 int engine_unregister(struct engine *engine, const char *name)
@@ -1033,18 +858,6 @@ int engine_unregister(struct engine *engine, const char *name)
 	return kr_error(ENOENT);
 }
 
-void engine_lualib(struct engine *engine, const char *name, lua_CFunction lib_cb)
-{
-	if (engine != NULL) {
-#if LUA_VERSION_NUM >= 502
-		luaL_requiref(engine->L, name, lib_cb, 1);
-		lua_pop(engine->L, 1);
-#else
-		lib_cb(engine->L);
-#endif
-	}
-}
-
 struct engine *engine_luaget(lua_State *L)
 {
 	lua_getglobal(L, "__engine");
