@@ -1,17 +1,5 @@
 /*  Copyright (C) 2015-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 #pragma once
@@ -37,7 +25,6 @@ struct engine {
     knot_mm_t *pool;
     char *hostname;
     struct lua_State *L;
-    char *moduledir;
 };
 
 int engine_init(struct engine *engine, knot_mm_t *pool);
@@ -58,15 +45,12 @@ int engine_ipc(struct engine *engine, const char *expr);
 
 int engine_load_sandbox(struct engine *engine);
 int engine_loadconf(struct engine *engine, const char *config_path);
-int engine_load_defaults(struct engine *engine);
 
 /** Start the lua engine and execute the config. */
 int engine_start(struct engine *engine);
 void engine_stop(struct engine *engine);
 int engine_register(struct engine *engine, const char *name, const char *precedence, const char* ref);
 int engine_unregister(struct engine *engine, const char *name);
-void engine_lualib(struct engine *engine, const char *name, int (*lib_cb) (struct lua_State *));
-
 
 /** Return engine light userdata. */
 struct engine *engine_luaget(struct lua_State *L);
@@ -75,10 +59,6 @@ struct engine *engine_luaget(struct lua_State *L);
 char *engine_get_hostname(struct engine *engine);
 int engine_set_hostname(struct engine *engine, const char *hostname);
 
-/** Set/get the per engine moduledir */
-char *engine_get_moduledir(struct engine *engine);
-int engine_set_moduledir(struct engine *engine, const char *moduledir);
-
 /** Load root hints from a zonefile (or config-time default if NULL).
  *
  * @return error message or NULL (statically allocated)
@@ -87,3 +67,28 @@ int engine_set_moduledir(struct engine *engine, const char *moduledir);
 KR_EXPORT
 const char* engine_hint_root_file(struct kr_context *ctx, const char *file);
 
+/* @internal Array of ip address shorthand. */
+typedef array_t(char*) addr_array_t;
+
+typedef array_t(const char*) config_array_t;
+
+typedef struct {
+	int fd;
+	endpoint_flags_t flags; /**< .sock_type isn't meaningful here */
+} flagged_fd_t;
+typedef array_t(flagged_fd_t) flagged_fd_array_t;
+
+struct args {
+	addr_array_t addrs, addrs_tls;
+	flagged_fd_array_t fds;
+	int control_fd;
+	int forks;
+	config_array_t config;
+	const char *rundir;
+	bool interactive;
+	bool quiet;
+	bool tty_binary_output;
+};
+
+/** Pointer to kresd arguments. */
+KR_EXPORT extern struct args *the_args;
