@@ -1,17 +1,5 @@
 /*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 #include <string.h>
@@ -27,10 +15,10 @@
 #include "daemon/session.h"
 
 #define negotiate_bufsize(func, handle, bufsize_want) do { \
-    int bufsize = 0; func(handle, &bufsize); \
-	if (bufsize < bufsize_want) { \
-		bufsize = bufsize_want; \
-		func(handle, &bufsize); \
+    int bufsize = 0; (func)((handle), &bufsize); \
+	if (bufsize < (bufsize_want)) { \
+		bufsize = (bufsize_want); \
+		(func)((handle), &bufsize); \
 	} \
 } while (0)
 
@@ -80,64 +68,125 @@ void udp_recv(uv_udp_t *handle, ssize_t nread, const uv_buf_t *buf,
 	}
 	if (nread <= 0) {
 		if (nread < 0) { /* Error response, notify resolver */
-			worker_submit(s, NULL);
+			worker_submit(s, NULL, NULL);
 		} /* nread == 0 is for freeing buffers, we don't need to do this */
 		return;
 	}
 	if (addr->sa_family == AF_UNSPEC) {
 		return;
 	}
-	struct sockaddr *peer = session_get_peer(s);
 	if (session_flags(s)->outgoing) {
+		const struct sockaddr *peer = session_get_peer(s);
 		assert(peer->sa_family != AF_UNSPEC);
 		if (kr_sockaddr_cmp(peer, addr) != 0) {
+			kr_log_verbose("[io] <= ignoring UDP from unexpected address '%s'\n",
+					kr_straddr(addr));
 			return;
 		}
-	} else {
-		memcpy(peer, addr, kr_sockaddr_len(addr));
 	}
 	ssize_t consumed = session_wirebuf_consume(s, (const uint8_t *)buf->base,
 						   nread);
 	assert(consumed == nread); (void)consumed;
-	session_wirebuf_process(s);
+	session_wirebuf_process(s, addr);
 	session_wirebuf_discard(s);
 	mp_flush(worker->pkt_pool.ctx);
 }
 
-static int udp_bind_finalize(uv_handle_t *handle)
+static int family_to_freebind_option(sa_family_t sa_family, int *level, int *name)
 {
-	check_bufsize(handle);
-	/* Handle is already created, just create context. */
-	struct session *s = session_new(handle, false);
-	assert(s);
-	session_flags(s)->outgoing = false;
-	return io_start_read(handle);
+	switch (sa_family) {
+	case AF_INET:
+		*level = IPPROTO_IP;
+#if defined(IP_FREEBIND)
+		*name = IP_FREEBIND;
+#elif defined(IP_BINDANY)
+		*name = IP_BINDANY;
+#else
+		return kr_error(ENOTSUP);
+#endif
+		break;
+	case AF_INET6:
+#if defined(IP_FREEBIND)
+		*level = IPPROTO_IP;
+		*name = IP_FREEBIND;
+#elif defined(IPV6_BINDANY)
+		*level = IPPROTO_IPV6;
+		*name = IPV6_BINDANY;
+#else
+		return kr_error(ENOTSUP);
+#endif
+		break;
+	default:
+		return kr_error(ENOTSUP);
+	}
+	return kr_ok();
 }
 
-int udp_bind(uv_udp_t *handle, struct sockaddr *addr)
+int io_bind(const struct sockaddr *addr, int type, const endpoint_flags_t *flags)
 {
-	unsigned flags = UV_UDP_REUSEADDR;
-	if (addr->sa_family == AF_INET6) {
-		flags |= UV_UDP_IPV6ONLY;
+	const int fd = socket(addr->sa_family, type, 0);
+	if (fd < 0) return kr_error(errno);
+
+	int yes = 1;
+	if (addr->sa_family == AF_INET || addr->sa_family == AF_INET6) {
+		if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes)))
+			return kr_error(errno);
+
+#ifdef SO_REUSEPORT_LB
+		if (setsockopt(fd, SOL_SOCKET, SO_REUSEPORT_LB, &yes, sizeof(yes)))
+			return kr_error(errno);
+#elif defined(SO_REUSEPORT) && defined(__linux__) /* different meaning on (Free)BSD */
+		if (setsockopt(fd, SOL_SOCKET, SO_REUSEPORT, &yes, sizeof(yes)))
+			return kr_error(errno);
+#endif
+
+#ifdef IPV6_V6ONLY
+		if (addr->sa_family == AF_INET6
+		    && setsockopt(fd, IPPROTO_IPV6, IPV6_V6ONLY, &yes, sizeof(yes)))
+			return kr_error(errno);
+#endif
+		if (flags != NULL && flags->freebind) {
+			int optlevel;
+			int optname;
+			int ret = family_to_freebind_option(addr->sa_family, &optlevel, &optname);
+			if (ret) return kr_error(ret);
+			if (setsockopt(fd, optlevel, optname, &yes, sizeof(yes)))
+				return kr_error(errno);
+		}
 	}
-	int ret = uv_udp_bind(handle, addr, flags);
-	if (ret != 0) {
-		return ret;
-	}
-	return udp_bind_finalize((uv_handle_t *)handle);
+
+	if (bind(fd, addr, kr_sockaddr_len(addr)))
+		return kr_error(errno);
+
+	return fd;
 }
 
-int udp_bindfd(uv_udp_t *handle, int fd)
+int io_listen_udp(uv_loop_t *loop, uv_udp_t *handle, int fd)
 {
 	if (!handle) {
 		return kr_error(EINVAL);
 	}
+	int ret = uv_udp_init(loop, handle);
+	if (ret) return ret;
 
-	int ret = uv_udp_open(handle, (uv_os_sock_t) fd);
-	if (ret != 0) {
-		return ret;
+	ret = uv_udp_open(handle, fd);
+	if (ret) return ret;
+
+	uv_handle_t *h = (uv_handle_t *)handle;
+	check_bufsize(h);
+	/* Handle is already created, just create context. */
+	struct session *s = session_new(h, false);
+	assert(s);
+	session_flags(s)->outgoing = false;
+
+	int socklen = sizeof(union inaddr);
+	ret = uv_udp_getsockname(handle, session_get_sockname(s), &socklen);
+	if (ret) {
+		kr_log_error("ERROR: getsockname failed: %s\n", uv_strerror(ret));
+		abort(); /* It might be nontrivial not to leak something here. */
 	}
-	return udp_bind_finalize((uv_handle_t *)handle);
+
+	return io_start_read(h);
 }
 
 void tcp_timeout_trigger(uv_timer_t *timer)
@@ -204,7 +253,7 @@ static void tcp_recv(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf)
 {
 	struct session *s = handle->data;
 	assert(s && session_get_handle(s) == (uv_handle_t *)handle &&
-	       handle->type == UV_TCP);	
+	       handle->type == UV_TCP);
 
 	if (session_flags(s)->closing) {
 		return;
@@ -257,7 +306,7 @@ static void tcp_recv(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf)
 	consumed = session_wirebuf_consume(s, data, data_len);
 	assert(consumed == data_len);
 
-	int ret = session_wirebuf_process(s);
+	int ret = session_wirebuf_process(s, session_get_peer(s));
 	if (ret < 0) {
 		/* An error has occurred, close the session. */
 		worker_end_tcp(s);
@@ -273,7 +322,7 @@ static void _tcp_accept(uv_stream_t *master, int status, bool tls)
 		return;
 	}
 
-	struct worker_ctx *worker = (struct worker_ctx *)master->loop->data;
+	struct worker_ctx *worker = the_worker;
 	uv_tcp_t *client = malloc(sizeof(uv_tcp_t));
 	if (!client) {
 		return;
@@ -304,19 +353,28 @@ static void _tcp_accept(uv_stream_t *master, int status, bool tls)
 		return;
 	}
 
-	/* Set deadlines for TCP connection and start reading.
-	 * It will re-check every half of a request time limit if the connection
-	 * is idle and should be terminated, this is an educated guess. */
-	struct sockaddr *peer = session_get_peer(s);
-	int peer_len = sizeof(union inaddr);
-	int ret = uv_tcp_getpeername(client, peer, &peer_len);
-	if (ret || peer->sa_family == AF_UNSPEC) {
+	/* Get peer's and our address.  We apparently get specific sockname here
+	 * even if we listened on a wildcard address. */
+	struct sockaddr *sa = session_get_peer(s);
+	int sa_len = sizeof(struct sockaddr_in6);
+	int ret = uv_tcp_getpeername(client, sa, &sa_len);
+	if (ret || sa->sa_family == AF_UNSPEC) {
+		session_close(s);
+		return;
+	}
+	sa = session_get_sockname(s);
+	sa_len = sizeof(struct sockaddr_in6);
+	ret = uv_tcp_getsockname(client, sa, &sa_len);
+	if (ret || sa->sa_family == AF_UNSPEC) {
 		session_close(s);
 		return;
 	}
 
-	const struct engine *engine = worker->engine;
-	const struct network *net = &engine->net;
+	/* Set deadlines for TCP connection and start reading.
+	 * It will re-check every half of a request time limit if the connection
+	 * is idle and should be terminated, this is an educated guess. */
+
+	const struct network *net = &worker->engine->net;
 	uint64_t idle_in_timeout = net->tcp.in_idle_timeout;
 
 	uint64_t timeout = KR_CONN_RTT_MAX / 2;
@@ -348,93 +406,215 @@ static void tls_accept(uv_stream_t *master, int status)
 	_tcp_accept(master, status, true);
 }
 
-static int set_tcp_option(uv_handle_t *handle, int option, int val)
+int io_listen_tcp(uv_loop_t *loop, uv_tcp_t *handle, int fd, int tcp_backlog, bool has_tls)
 {
-	uv_os_fd_t fd = 0;
-	if (uv_fileno(handle, &fd) == 0) {
-		return setsockopt(fd, IPPROTO_TCP, option, &val, sizeof(val));
+	const uv_connection_cb connection = has_tls ? tls_accept : tcp_accept;
+	if (!handle) {
+		return kr_error(EINVAL);
+	}
+	int ret = uv_tcp_init(loop, handle);
+	if (ret) return ret;
+
+	ret = uv_tcp_open(handle, (uv_os_sock_t) fd);
+	if (ret) return ret;
+
+	int val; (void)val;
+	/* TCP_DEFER_ACCEPT delays accepting connections until there is readable data. */
+#ifdef TCP_DEFER_ACCEPT
+	val = KR_CONN_RTT_MAX/1000;
+	if (setsockopt(fd, IPPROTO_TCP, TCP_DEFER_ACCEPT, &val, sizeof(val))) {
+		kr_log_error("[ io ] listen TCP (defer_accept): %s\n", strerror(errno));
+	}
+#endif
+
+	ret = uv_listen((uv_stream_t *)handle, tcp_backlog, connection);
+	if (ret != 0) {
+		return ret;
 	}
-	return 0; /* N/A */
-}
 
-static int tcp_bind_finalize(uv_handle_t *handle)
-{
 	/* TCP_FASTOPEN enables 1 RTT connection resumptions. */
 #ifdef TCP_FASTOPEN
-# ifdef __linux__
-	(void) set_tcp_option(handle, TCP_FASTOPEN, 16); /* Accepts queue length hint */
-# else
-	(void) set_tcp_option(handle, TCP_FASTOPEN, 1);  /* Accepts on/off */
-# endif
+	#ifdef __linux__
+	val = 16; /* Accepts queue length hint */
+	#else
+	val = 1; /* Accepts on/off */
+	#endif
+	if (setsockopt(fd, IPPROTO_TCP, TCP_FASTOPEN, &val, sizeof(val))) {
+		kr_log_error("[ io ] listen TCP (fastopen): %s%s\n", strerror(errno),
+			(errno != EPERM ? "" :
+			 ".  This may be caused by TCP Fast Open being disabled in the OS."));
+	}
 #endif
 
 	handle->data = NULL;
 	return 0;
 }
 
-static int _tcp_bind(uv_tcp_t *handle, struct sockaddr *addr, uv_connection_cb connection, int tcp_backlog)
+/**
+ * TTY control: process input and free() the buffer.
+ *
+ * For parameters see http://docs.libuv.org/en/v1.x/stream.html#c.uv_read_cb
+ *
+ * - This is just basic read-eval-print; libedit is supported through kresc;
+ */
+void io_tty_process_input(uv_stream_t *stream, ssize_t nread, const uv_buf_t *buf)
 {
-	unsigned flags = 0;
-	if (addr->sa_family == AF_INET6) {
-		flags |= UV_TCP_IPV6ONLY;
+	char *commands = buf ? buf->base : NULL; /* To be free()d on return. */
+
+	/* Set output streams */
+	FILE *out = stdout;
+	uv_os_fd_t stream_fd = 0;
+	struct args *args = the_args;
+	if (uv_fileno((uv_handle_t *)stream, &stream_fd)) {
+		uv_close((uv_handle_t *)stream, (uv_close_cb) free);
+		free(commands);
+		return;
+	}
+	if (stream_fd != STDIN_FILENO) {
+		if (nread < 0) { /* Close if disconnected */
+			uv_close((uv_handle_t *)stream, (uv_close_cb) free);
+		}
+		if (nread <= 0) {
+			free(commands);
+			return;
+		}
+		uv_os_fd_t dup_fd = dup(stream_fd);
+		if (dup_fd >= 0) {
+			out = fdopen(dup_fd, "w");
+		}
 	}
 
-	int ret = uv_tcp_bind(handle, addr, flags);
-	if (ret != 0) {
-		return ret;
-	}
+	char *cmd = NULL;
+	/* Execute */
+	if (stream && commands && nread > 0) {
+		/* Ensure commands is 0-terminated */
+		if (commands[nread - 1] == '\n') {
+			commands[nread - 1] = '\0';
+		} else {
+			if (nread >= buf->len) { /* only equality should be possible */
+				char *newbuf = realloc(commands, nread + 1);
+				if (!newbuf)
+					goto finish;
+				commands = newbuf;
+			}
+			commands[nread] = '\0';
+		}
 
-	/* TCP_DEFER_ACCEPT delays accepting connections until there is readable data. */
-#ifdef TCP_DEFER_ACCEPT
-	if (set_tcp_option((uv_handle_t *)handle, TCP_DEFER_ACCEPT, KR_CONN_RTT_MAX/1000) != 0) {
-		kr_log_info("[ io ] tcp_bind (defer_accept): %s\n", strerror(errno));
-	}
-#endif
+		const char *delim = args->quiet ? "" : "> ";
 
-	ret = uv_listen((uv_stream_t *)handle, tcp_backlog, connection);
-	if (ret != 0) {
-		return ret;
-	}
+		/* No command, just new line */
+		if (nread == 1 && args->tty_binary_output == false && commands[nread-1] == '\0') {
+			if (stream_fd != STDIN_FILENO) {
+				fprintf(out, "%s", delim);
+			}
+			if (stream_fd == STDIN_FILENO || VERBOSE_STATUS) {
+				fprintf(stdout, "%s", delim);
+			}
+		}
 
-	return tcp_bind_finalize((uv_handle_t *)handle);
+		cmd = strtok(commands, "\n");
+		while (cmd != NULL) {
+			/* Pseudo-command for switching to "binary output"; */
+			if (strcmp(cmd, "__binary") == 0) {
+				args->tty_binary_output = true;
+				cmd = strtok(NULL, "\n");
+				continue;
+			}
+
+			lua_State *L = the_worker->engine->L;
+			int ret = engine_cmd(L, cmd, false);
+			const char *message = "";
+			if (lua_gettop(L) > 0) {
+				message = lua_tostring(L, -1);
+			}
+
+			/* Simpler output in binary mode */
+			if (args->tty_binary_output) {
+				size_t len_s = strlen(message);
+				if (len_s > UINT32_MAX) {
+					cmd = strtok(NULL, "\n");
+					continue;
+				}
+				uint32_t len_n = htonl(len_s);
+				fwrite(&len_n, sizeof(len_n), 1, out);
+				fwrite(message, len_s, 1, out);
+				lua_settop(L, 0);
+				cmd = strtok(NULL, "\n");
+				continue;
+			}
+			/* Log to remote socket if connected */
+			if (stream_fd != STDIN_FILENO) {
+				if (VERBOSE_STATUS)
+					fprintf(stdout, "%s\n", cmd); /* Duplicate command to logs */
+				if (message)
+					fprintf(out, "%s", message); /* Duplicate output to sender */
+				if (message || !args->quiet)
+					fprintf(out, "\n");
+				fprintf(out, "%s", delim);
+			}
+			if (stream_fd == STDIN_FILENO || VERBOSE_STATUS) {
+				/* Log to standard streams */
+				FILE *fp_out = ret ? stderr : stdout;
+				if (message)
+					fprintf(fp_out, "%s", message);
+				if (message || !args->quiet)
+					fprintf(fp_out, "\n");
+				fprintf(fp_out, "%s", delim);
+			}
+			lua_settop(L, 0);
+			cmd = strtok(NULL, "\n");
+		}
+	}
+finish:
+	free(commands);
+	/* Close if redirected */
+	if (stream_fd != STDIN_FILENO) {
+		fclose(out);
+	}
 }
 
-int tcp_bind(uv_tcp_t *handle, struct sockaddr *addr, int tcp_backlog)
+void io_tty_alloc(uv_handle_t *handle, size_t suggested, uv_buf_t *buf)
 {
-	return _tcp_bind(handle, addr, tcp_accept, tcp_backlog);
+	buf->len = suggested;
+	buf->base = malloc(suggested);
 }
 
-int tcp_bind_tls(uv_tcp_t *handle, struct sockaddr *addr, int tcp_backlog)
+void io_tty_accept(uv_stream_t *master, int status)
 {
-	return _tcp_bind(handle, addr, tls_accept, tcp_backlog);
+	uv_tcp_t *client = malloc(sizeof(*client));
+	struct args *args = the_args;
+	if (client) {
+		 uv_tcp_init(master->loop, client);
+		 if (uv_accept(master, (uv_stream_t *)client) != 0) {
+			free(client);
+			return;
+		 }
+		 uv_read_start((uv_stream_t *)client, io_tty_alloc, io_tty_process_input);
+		 /* Write command line */
+		 if (!args->quiet) {
+			uv_buf_t buf = { "> ", 2 };
+			uv_try_write((uv_stream_t *)client, &buf, 1);
+		 }
+	}
 }
 
-static int _tcp_bindfd(uv_tcp_t *handle, int fd, uv_connection_cb connection, int tcp_backlog)
+int io_listen_pipe(uv_loop_t *loop, uv_pipe_t *handle, int fd)
 {
 	if (!handle) {
 		return kr_error(EINVAL);
 	}
+	int ret = uv_pipe_init(loop, handle, 0);
+	if (ret) return ret;
 
-	int ret = uv_tcp_open(handle, (uv_os_sock_t) fd);
-	if (ret != 0) {
-		return ret;
-	}
+	ret = uv_pipe_open(handle, fd);
+	if (ret) return ret;
 
-	ret = uv_listen((uv_stream_t *)handle, tcp_backlog, connection);
-	if (ret != 0) {
-		return ret;
-	}
-	return tcp_bind_finalize((uv_handle_t *)handle);
-}
+	ret = uv_listen((uv_stream_t *)handle, 16, io_tty_accept);
+	if (ret) return ret;
 
-int tcp_bindfd(uv_tcp_t *handle, int fd, int tcp_backlog)
-{
-	return _tcp_bindfd(handle, fd, tcp_accept, tcp_backlog);
-}
+	handle->data = NULL;
 
-int tcp_bindfd_tls(uv_tcp_t *handle, int fd, int tcp_backlog)
-{
-	return _tcp_bindfd(handle, fd, tls_accept, tcp_backlog);
+	return 0;
 }
 
 int io_create(uv_loop_t *loop, uv_handle_t *handle, int type, unsigned family, bool has_tls)
@@ -461,8 +641,10 @@ void io_deinit(uv_handle_t *handle)
 	if (!handle) {
 		return;
 	}
-	session_free(handle->data);
-	handle->data = NULL;
+	if (handle->data) {
+		session_free(handle->data);
+		handle->data = NULL;
+	}
 }
 
 void io_free(uv_handle_t *handle)
