@@ -1,4 +1,6 @@
 -- LuaJIT ffi bindings for libkres, a DNS resolver library.
+-- SPDX-License-Identifier: GPL-3.0-or-later
+--
 -- @note Since it's statically compiled, it expects to find the symbols in the C namespace.
 
 local kres -- the module
@@ -57,6 +59,40 @@ int inet_pton(int af, const char *src, void *dst);
 int gettimeofday(struct timeval *tv, struct timezone *tz);
 ]]
 
+
+-- TMP: compatibility with both libknot 2.8 and 2.9
+local knot_rdataset_t_cdef
+local sover_pos = string.find(libknot_SONAME, '%d')
+if not sover_pos then
+	error('unexpected libknot soname: ' .. libknot_SONAME)
+end
+local sover = string.sub(libknot_SONAME, sover_pos , sover_pos)
+if sover == '9' then
+	knot_rdataset_t_cdef = [[
+		typedef struct {
+			uint16_t count;
+			knot_rdata_t *rdata;
+		} knot_rdataset_t;
+	]]
+elseif sover == '1' then -- it's 10 really, but this is simpler
+	knot_rdataset_t_cdef = [[
+		typedef struct {
+			uint16_t count;
+			uint32_t size;
+			knot_rdata_t *rdata;
+		} knot_rdataset_t;
+	]]
+else
+	error('unexpected libknot version: ' .. sover)
+end
+ffi.cdef([[
+		typedef struct {
+			uint16_t len;
+			uint8_t data[];
+		} knot_rdata_t;
+	]] .. knot_rdataset_t_cdef)
+
+
 require('kres-gen')
 
 -- Error code representation
@@ -239,7 +275,7 @@ local timeval_t = ffi.typeof('struct timeval')
 
 -- Metatype for sockaddr
 local addr_buf = ffi.new('char[16]')
-local str_addr_buf = ffi.new('char[46 + 1 + 6 + 1]') -- IPv6 + #port + \0
+local str_addr_buf = ffi.new('char[46 + 1 + 6 + 1]') -- INET6_ADDRSTRLEN + #port + \0
 local str_addr_buf_len = ffi.sizeof(str_addr_buf)
 local sockaddr_t = ffi.typeof('struct sockaddr')
 ffi.metatype( sockaddr_t, {
@@ -269,9 +305,9 @@ local lru_metatype = {
 	-- Create a new LRU with given value type
 	-- By default the LRU will have a capacity of 65536 elements
 	-- Note: At the point the parametrized type must be finalized
-	__new = function (ct, max_slots)
+	__new = function (ct, max_slots, alignment)
 		-- {0} will make sure that the value is coercible to a number
-		local o = ffi.new(ct, {0}, C.lru_create_impl(max_slots or 65536, nil, nil))
+		local o = ffi.new(ct, {0}, C.lru_create_impl(max_slots or 65536, alignment or 1, nil, nil))
 		if o.lru == nil then
 			return
 		end
@@ -492,51 +528,6 @@ local function knot_pkt_rr(section, i)
 	return ret
 end
 
--- Helpers for converting packet to text
-local function section_tostring(pkt, section_id)
-	local data = {}
-	local section = pkt.sections + section_id
-	if section.count > 0 then
-		table.insert(data, string.format('\n;; %s\n', const_section_str[section_id]))
-		for j = 0, section.count - 1 do
-			local rrset = knot_pkt_rr(section, j)
-			local rrtype = rrset.type
-			if rrtype ~= const_type.OPT and rrtype ~= const_type.TSIG then
-				table.insert(data, rrset:txt_dump())
-			end
-		end
-	end
-	return table.concat(data, '')
-end
-
-local function packet_tostring(pkt)
-	local hdr = string.format(';; ->>HEADER<<- opcode: %s; status: %s; id: %d\n',
-		const_opcode_str[pkt:opcode()], const_rcode_str[pkt:rcode()], pkt:id())
-	local flags = {}
-	for _,v in ipairs({'rd', 'tc', 'aa', 'qr', 'cd', 'ad', 'ra'}) do
-		if(pkt[v](pkt)) then table.insert(flags, v) end
-	end
-	local info = string.format(';; Flags: %s; QUERY: %d; ANSWER: %d; AUTHORITY: %d; ADDITIONAL: %d\n',
-		table.concat(flags, ' '), pkt:qdcount(), pkt:ancount(), pkt:nscount(), pkt:arcount())
-	local data = '\n'
-	if pkt.opt_rr ~= nil then
-		data = data..string.format(';; OPT PSEUDOSECTION:\n%s', pkt.opt_rr:tostring())
-	end
-	if pkt.tsig_rr ~= nil then
-		data = data..string.format(';; TSIG PSEUDOSECTION:\n%s', pkt.tsig_rr:tostring())
-	end
-	-- Zone transfer answers may omit question
-	if pkt:qdcount() > 0 then
-		data = data..string.format(';; QUESTION\n;; %s\t%s\t%s\n',
-			dname2str(pkt:qname()), const_type_str[pkt:qtype()], const_class_str[pkt:qclass()])
-	end
-	local data_sec = {}
-	for i = const_section.ANSWER, const_section.ADDITIONAL do
-		table.insert(data_sec, section_tostring(pkt, i))
-	end
-	return hdr..info..data..table.concat(data_sec, '')
-end
-
 -- Metatype for packet
 ffi.metatype( knot_pkt_t, {
 	__new = function (_, size, wire)
@@ -593,6 +584,12 @@ ffi.metatype( knot_pkt_t, {
 		cd = function (pkt, val) return pkt_bit(pkt, 3, 0x10, val) end,
 		ad = function (pkt, val) return pkt_bit(pkt, 3, 0x20, val) end,
 		ra = function (pkt, val) return pkt_bit(pkt, 3, 0x80, val) end,
+		-- "do" is a reserved word in Lua; only getter
+		dobit = function(pkt, val)
+			assert(val == nil, 'dobit is getter only')
+			assert(ffi.istype(knot_pkt_t, pkt))
+			return C.kr_pkt_has_dnssec(pkt)
+		end,
 		-- Question
 		qname = function(pkt)
 			assert(ffi.istype(knot_pkt_t, pkt))
@@ -679,7 +676,7 @@ ffi.metatype( knot_pkt_t, {
 		end,
 		tostring = function(pkt)
 			assert(ffi.istype(knot_pkt_t, pkt))
-			return packet_tostring(pkt)
+			return ffi.string(ffi.gc(C.kr_pkt_text(pkt), C.free))
 		end,
 		-- Return number of remaining empty bytes in the packet
 		-- This is generally useful to check if there's enough space
@@ -726,15 +723,56 @@ ffi.metatype( kr_query_t, {
 		end,
 	},
 })
+
+-- helper for trace_chain_callbacks
+-- ignores return values from successfull calls but logs tracebacks for throws
+local function void_xpcall_log_tb(func, req, msg)
+	local ok, err = xpcall(func, debug.traceback, req, msg)
+	if not ok then
+		log('error: callback %s req %s msg %s stack traceback:\n%s', func, req, msg, err)
+	end
+end
+
+local function void_xpcall_finish_tb(func, req)
+	local ok, err = xpcall(func, debug.traceback, req)
+	if not ok then
+		log('error: callback %s req %s stack traceback:\n%s', func, req, err)
+	end
+end
+
+
 -- Metatype for request
 local kr_request_t = ffi.typeof('struct kr_request')
 ffi.metatype( kr_request_t, {
 	__index = {
+		-- makes sense only when request is finished
+		all_from_cache = function(req)
+			assert(ffi.istype(kr_request_t, req))
+			local rplan = ffi.C.kr_resolve_plan(req)
+			if tonumber(rplan.pending.len) > 0 then
+				-- an unresolved query,
+				-- i.e. something is missing from the cache
+				return false
+			end
+			for idx=0, tonumber(rplan.resolved.len) - 1 do
+				if not rplan.resolved.at[idx].flags.CACHED then
+					return false
+				end
+			end
+			return true
+		end,
 		current = function(req)
 			assert(ffi.istype(kr_request_t, req))
 			if req.current_query == nil then return nil end
 			return req.current_query
 		end,
+		-- returns the initial query that started the request
+		initial = function(req)
+			assert(ffi.istype(kr_request_t, req))
+			local rplan = C.kr_resolve_plan(req)
+			if rplan.initial == nil then return nil end
+			return rplan.initial
+		end,
 		-- Return last query on the resolution plan
 		last = function(req)
 			assert(ffi.istype(kr_request_t, req))
@@ -769,6 +807,67 @@ ffi.metatype( kr_request_t, {
 			assert(ffi.istype(kr_request_t, req))
 			return C.kr_rplan_pop(C.kr_resolve_plan(req), qry)
 		end,
+		selected_tostring = function(req)
+			assert(ffi.istype(kr_request_t, req))
+			local buf = {}
+			if #req.answ_selected ~= 0 then
+				table.insert(buf, string.format('[%05d.00][dbg ] selected rrsets from answer sections:\n', req.uid))
+				table.insert(buf, tostring(req.answ_selected))
+			end
+			if #req.auth_selected ~= 0 then
+				table.insert(buf, string.format('[%05d.00][dbg ] selected rrsets from authority sections:\n', req.uid))
+				table.insert(buf, tostring(req.auth_selected))
+			end
+			if #req.add_selected ~= 0 then
+				table.insert(buf, string.format('[%05d.00][dbg ] selected rrsets from additional sections:\n', req.uid))
+				table.insert(buf, tostring(req.add_selected))
+			end
+			return table.concat(buf, '')
+		end,
+
+		-- chain new callbacks after the old ones
+		-- creates new wrapper functions as necessary
+		-- note: callbacks are FFI cdata pointers so tests must
+		--       use explicit "cb == nil", just "if cb" does not work
+		--
+		trace_chain_callbacks = function (req, new_log, new_finish)
+			local log_wrapper
+			if req.trace_log == nil then
+				req.trace_log = new_log
+			else
+				local old_log = req.trace_log
+				log_wrapper = ffi.cast('trace_log_f',
+				function(cbreq, msg)
+					jit.off(true, true) -- JIT for (C -> lua)^2 nesting isn't allowed
+					void_xpcall_log_tb(old_log, cbreq, msg)
+					void_xpcall_log_tb(new_log, cbreq, msg)
+				end)
+				req.trace_log = log_wrapper
+			end
+			local old_finish = req.trace_finish
+			if not (log_wrapper ~= nil or old_finish ~= nil) then
+				req.trace_finish = new_finish
+			else
+				local fin_wrapper
+				fin_wrapper = ffi.cast('trace_callback_f',
+				function(cbreq)
+					jit.off(true, true) -- JIT for (C -> lua)^2 nesting isn't allowed
+					if old_finish ~= nil then
+						void_xpcall_finish_tb(old_finish, cbreq)
+					end
+					if new_finish ~= nil then
+						void_xpcall_finish_tb(new_finish, cbreq)
+					end
+					-- beware: finish callbacks can call log callback
+					if log_wrapper ~= nil then
+						log_wrapper:free()
+					end
+					fin_wrapper:free()
+				end)
+				req.trace_finish = fin_wrapper
+			end
+		end,
+
 		-- Return per-request variable table
 		-- The request can store anything in this Lua table and it will be freed
 		-- when the request is closed, it doesn't have to worry about contents.
@@ -804,7 +903,17 @@ local function c_array_iter(t, i)
 	return i, t.at[i][0]
 end
 
--- Metatype for ranked record array
+-- Metatype for a single ranked record array entry (one RRset)
+local ranked_rr_array_entry_t = ffi.typeof('ranked_rr_array_entry_t')
+ffi.metatype(ranked_rr_array_entry_t, {
+	__tostring = function(self)
+		return string.format('; ranked rrset to_wire %s, rank 0%.2o, cached %s, qry_uid %s, revalidations %s\n%s',
+		self.to_wire, self.rank, self.cached, self.qry_uid,
+		self.revalidation_cnt, string.format('%s', self.rr))
+	end
+})
+
+-- Metatype for ranked record array (array of RRsets)
 local ranked_rr_array_t = ffi.typeof('ranked_rr_array_t')
 ffi.metatype(ranked_rr_array_t, {
 	__len = function(self)
@@ -818,7 +927,14 @@ ffi.metatype(ranked_rr_array_t, {
 			if i < 0 or i > self.len then return nil end
 			return self.at[i][0]
 		end,
-	}
+	},
+	__tostring = function(self)
+		local buf = {}
+		for _, rrset in ipairs(self) do
+			table.insert(buf, tostring(rrset))
+		end
+		return table.concat(buf, '')
+	end
 })
 
 -- Cache metatype
@@ -840,9 +956,9 @@ ffi.metatype( kr_cache_t, {
 			if ret ~= 0 then return nil, knot_error_t(ret) end
 			return true
 		end,
-		sync = function (self)
+		commit = function (self)
 			assert(ffi.istype(kr_cache_t, self))
-			local ret = C.kr_cache_sync(self)
+			local ret = C.kr_cache_commit(self)
 			if ret ~= 0 then return nil, knot_error_t(ret) end
 			return true
 		end,
@@ -916,8 +1032,9 @@ kres = {
 	rrset = knot_rrset_t,
 	packet = knot_pkt_t,
 	lru = function (max_size, value_type)
-	  local ct = ffi.typeof(typed_lru_t, value_type or ffi.typeof('uint64_t'))
-	  return ffi.metatype(ct, lru_metatype)(max_size)
+	  value_type = value_type or ffi.typeof('uint64_t')
+	  local ct = ffi.typeof(typed_lru_t, value_type)
+	  return ffi.metatype(ct, lru_metatype)(max_size, ffi.alignof(value_type))
 	end,
 
 	-- Metatypes.  Beware that any pointer will be cast silently...
