@@ -1,19 +1,23 @@
 /*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
+#include "kresconfig.h"
+
+#include "contrib/ccan/asprintf/asprintf.h"
+#include "contrib/cleanup.h"
+#include "contrib/ucw/mempool.h"
+#include "daemon/engine.h"
+#include "daemon/io.h"
+#include "daemon/network.h"
+#include "daemon/tls.h"
+#include "daemon/udp_queue.h"
+#include "daemon/worker.h"
+#include "lib/defines.h"
+#include "lib/dnssec.h"
+#include "lib/dnssec/ta.h"
+#include "lib/resolve.h"
+
 #include <arpa/inet.h>
 #include <assert.h>
 #include <getopt.h>
@@ -21,190 +25,23 @@
 #include <signal.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sys/resource.h>
 #include <unistd.h>
 
+#ifdef ENABLE_CAP_NG
+#include <cap-ng.h>
+#endif
+
+#include <lua.h>
 #include <uv.h>
-#ifdef HAS_SYSTEMD
+#if SYSTEMD_VERSION > 0
 #include <systemd/sd-daemon.h>
 #endif
 #include <libknot/error.h>
 
-#include <contrib/cleanup.h>
-#include <contrib/ucw/mempool.h>
-#include <contrib/ccan/asprintf/asprintf.h>
-#include "lib/defines.h"
-#include "lib/resolve.h"
-#include "lib/dnssec.h"
-#include "daemon/network.h"
-#include "daemon/worker.h"
-#include "daemon/engine.h"
-#include "daemon/bindings.h"
-#include "daemon/tls.h"
-#include "lib/dnssec/ta.h"
 
-/* We can fork early on Linux 3.9+ and do SO_REUSEPORT for better performance. */
-#if defined(UV_VERSION_HEX) && defined(SO_REUSEPORT) && defined(__linux__)
- #define CAN_FORK_EARLY 1
-#endif
+struct args the_args_value;  /** Static allocation for the_args singleton. */
 
-/* @internal Array of ip address shorthand. */
-typedef array_t(char*) addr_array_t;
-
-struct args {
-	int forks;
-	addr_array_t addr_set;
-	addr_array_t tls_set;
-	fd_array_t fd_set;
-	fd_array_t tls_fd_set;
-	char *keyfile;
-	int keyfile_unmanaged;
-	const char *moduledir;
-	const char *config;
-	int control_fd;
-	const char *rundir;
-	bool interactive;
-	bool quiet;
-	bool tty_binary_output;
-};
-
-/* lua_pcall helper function */
-static inline char *lua_strerror(int lua_err) {
-	switch (lua_err) {
-	case LUA_ERRRUN: return "a runtime error";
-	case LUA_ERRMEM: return "memory allocation error.";
-	case LUA_ERRERR: return "error while running the error handler function.";
-	default: return "a unknown error";
-	}
-}
-
-/**
- * TTY control: process input and free() the buffer.
- *
- * For parameters see http://docs.libuv.org/en/v1.x/stream.html#c.uv_read_cb
- *
- * - This is just basic read-eval-print; libedit is supported through kresc;
- * - stream->data contains program arguments (struct args);
- */
-static void tty_process_input(uv_stream_t *stream, ssize_t nread, const uv_buf_t *buf)
-{
-	char *cmd = buf ? buf->base : NULL; /* To be free()d on return. */
-
-	/* Set output streams */
-	FILE *out = stdout;
-	uv_os_fd_t stream_fd = 0;
-	struct args *args = stream->data;
-	if (uv_fileno((uv_handle_t *)stream, &stream_fd)) {
-		uv_close((uv_handle_t *)stream, (uv_close_cb) free);
-		free(cmd);
-		return;
-	}
-	if (stream_fd != STDIN_FILENO) {
-		if (nread < 0) { /* Close if disconnected */
-			uv_close((uv_handle_t *)stream, (uv_close_cb) free);
-		}
-		if (nread <= 0) {
-			free(cmd);
-			return;
-		}
-		uv_os_fd_t dup_fd = dup(stream_fd);
-		if (dup_fd >= 0) {
-			out = fdopen(dup_fd, "w");
-		}
-	}
-
-	/* Execute */
-	if (stream && cmd && nread > 0) {
-		/* Ensure cmd is 0-terminated */
-		if (cmd[nread - 1] == '\n') {
-			cmd[nread - 1] = '\0';
-		} else {
-			if (nread >= buf->len) { /* only equality should be possible */
-				char *newbuf = realloc(cmd, nread + 1);
-				if (!newbuf)
-					goto finish;
-				cmd = newbuf;
-			}
-			cmd[nread] = '\0';
-		}
-
-		/* Pseudo-command for switching to "binary output"; */
-		if (strcmp(cmd, "__binary") == 0) {
-			args->tty_binary_output = true;
-			goto finish;
-		}
-
-		struct engine *engine = ((struct worker_ctx *)stream->loop->data)->engine;
-		lua_State *L = engine->L;
-		int ret = engine_cmd(L, cmd, false);
-		const char *message = "";
-		if (lua_gettop(L) > 0) {
-			message = lua_tostring(L, -1);
-		}
-
-		/* Simpler output in binary mode */
-		if (args->tty_binary_output) {
-			size_t len_s = strlen(message);
-			if (len_s > UINT32_MAX)
-				goto finish;
-			uint32_t len_n = htonl(len_s);
-			fwrite(&len_n, sizeof(len_n), 1, out);
-			fwrite(message, len_s, 1, out);
-			lua_settop(L, 0);
-			goto finish;
-		}
-
-		/* Log to remote socket if connected */
-		const char *delim = args->quiet ? "" : "> ";
-		if (stream_fd != STDIN_FILENO) {
-			fprintf(stdout, "%s\n", cmd); /* Duplicate command to logs */
-			if (message)
-				fprintf(out, "%s", message); /* Duplicate output to sender */
-			if (message || !args->quiet)
-				fprintf(out, "\n");
-			fprintf(out, "%s", delim);
-		}
-		/* Log to standard streams */
-		FILE *fp_out = ret ? stderr : stdout;
-		if (message)
-			fprintf(fp_out, "%s", message);
-		if (message || !args->quiet)
-			fprintf(fp_out, "\n");
-		fprintf(fp_out, "%s", delim);
-		lua_settop(L, 0);
-	}
-finish:
-	fflush(out);
-	free(cmd);
-	/* Close if redirected */
-	if (stream_fd != STDIN_FILENO) {
-		fclose(out);
-	}
-}
-
-static void tty_alloc(uv_handle_t *handle, size_t suggested, uv_buf_t *buf) {
-	buf->len = suggested;
-	buf->base = malloc(suggested);
-}
-
-static void tty_accept(uv_stream_t *master, int status)
-{
-	uv_tcp_t *client = malloc(sizeof(*client));
-	struct args *args = master->data;
-	if (client) {
-		 uv_tcp_init(master->loop, client);
-		 if (uv_accept(master, (uv_stream_t *)client) != 0) {
-			free(client);
-			return;
-		 }
-		 client->data = args;
-		 uv_read_start((uv_stream_t *)client, tty_alloc, tty_process_input);
-		 /* Write command line */
-		 if (!args->quiet) {
-		 	uv_buf_t buf = { "> ", 2 };
-		 	uv_try_write((uv_stream_t *)client, &buf, 1);
-		 }
-	}
-}
 
 /* @internal AF_LOCAL reads may still be interrupted, loop it. */
 static bool ipc_readall(int fd, char *dst, size_t len)
@@ -338,20 +175,6 @@ end:
 	#pragma GCC diagnostic pop
 }
 
-/** Split away port from the address. */
-static const char *set_addr(char *addr, int *port)
-{
-	char *p = strchr(addr, '@');
-	if (!p) {
-		p = strchr(addr, '#');
-	}
-	if (p) {
-		*port = strtol(p + 1, NULL, 10);
-		*p = '\0';
-	}
-
-	return addr;
-}
 
 /*
  * Server operation.
@@ -395,13 +218,9 @@ static void help(int argc, char *argv[])
 	printf("\nParameters:\n"
 	       " -a, --addr=[addr]      Server address (default: localhost@53).\n"
 	       " -t, --tls=[addr]       Server address for TLS (default: off).\n"
-	       " -S, --fd=[fd]          Listen on given fd (handed out by supervisor).\n"
-	       " -T, --tlsfd=[fd]       Listen using TLS on given fd (handed out by supervisor).\n"
+	       " -S, --fd=[fd:kind]     Listen on given fd (handed out by supervisor, :kind is optional).\n"
 	       " -c, --config=[path]    Config file path (relative to [rundir]) (default: config).\n"
-	       " -k, --keyfile=[path]   File with root domain trust anchors (DS or DNSKEY), automatically updated.\n"
-	       " -K, --keyfile-ro=[path] File with read-only root domain trust anchors, for use with an external updater.\n"
-	       " -m, --moduledir=[path] Override the default module path (" MODULEDIR ").\n"
-	       " -f, --forks=N          Start N forks sharing the configuration.\n"
+	       " -n, --noninteractive   Don't start the read-eval-print loop for stdin+stdout.\n"
 	       " -q, --quiet            No command prompt in interactive mode.\n"
 	       " -v, --verbose          Run in verbose mode."
 #ifdef NOVERBOSELOG
@@ -429,35 +248,22 @@ static int run_worker(uv_loop_t *loop, struct engine *engine, fd_array_t *ipc_se
 		kr_log_error(
 			"[system] error: standard input is not a terminal or pipe; "
 			"use '-f 1' if you want non-interactive mode.  "
-			"Commands can be simply added to your configuration file or sent over the tty/$PID control socket.\n"
+			"Commands can be simply added to your configuration file or sent over the control socket.\n"
 			);
-		return 1;
+		return EXIT_FAILURE;
 	}
 
 	/* Control sockets or TTY */
-	auto_free char *sock_file = NULL;
 	uv_pipe_t pipe;
 	uv_pipe_init(loop, &pipe, 0);
 	pipe.data = args;
 	if (args->interactive) {
 		if (!args->quiet)
 			printf("[system] interactive mode\n> ");
-		fflush(stdout);
 		uv_pipe_open(&pipe, 0);
-		uv_read_start((uv_stream_t*) &pipe, tty_alloc, tty_process_input);
-	} else {
-		int pipe_ret = -1;
-		if (args->control_fd != -1) {
-			pipe_ret = uv_pipe_open(&pipe, args->control_fd);
-		} else {
-			(void) mkdir("tty", S_IRWXU|S_IRWXG);
-			sock_file = afmt("tty/%ld", (long)getpid());
-			if (sock_file) {
-				pipe_ret = uv_pipe_bind(&pipe, sock_file);
-			}
-		}
-		if (!pipe_ret)
-			uv_listen((uv_stream_t *) &pipe, 16, tty_accept);
+		uv_read_start((uv_stream_t*) &pipe, io_tty_alloc, io_tty_process_input);
+	} else if (args->control_fd != -1 && uv_pipe_open(&pipe, args->control_fd) == 0) {
+		uv_listen((uv_stream_t *) &pipe, 16, io_tty_accept);
 	}
 	/* Watch IPC pipes (or just assign them if leading the pgroup). */
 	if (!leader) {
@@ -471,67 +277,36 @@ static int run_worker(uv_loop_t *loop, struct engine *engine, fd_array_t *ipc_se
 	memcpy(&engine->ipc_set, ipc_set, sizeof(*ipc_set));
 
 	/* Notify supervisor. */
-#ifdef HAS_SYSTEMD
+#if SYSTEMD_VERSION > 0
 	sd_notify(0, "READY=1");
 #endif
 	/* Run event loop */
 	uv_run(loop, UV_RUN_DEFAULT);
-	if (sock_file) {
-		unlink(sock_file);
-	}
 	uv_close((uv_handle_t *)&pipe, NULL); /* Seems OK even on the stopped loop. */
-	return 0;
+	return EXIT_SUCCESS;
 }
 
-#ifdef HAS_SYSTEMD
-static void free_sd_socket_names(char **socket_names, int count)
-{
-	for (int i = 0; i < count; i++) {
-		free(socket_names[i]);
-	}
-	free(socket_names);
-}
-#endif
-
-static int set_keyfile(struct engine *engine, char *keyfile, bool unmanaged)
-{
-	assert(keyfile != NULL);
-	auto_free char *cmd = afmt("trust_anchors.config('%s',%s)",
-				   keyfile, unmanaged ? "true" : "nil");
-	if (!cmd) {
-		kr_log_error("[system] not enough memory\n");
-		return kr_error(ENOMEM);
-	}
-	int lua_ret = engine_cmd(engine->L, cmd, false);
-	if (lua_ret != 0) {
-		if (lua_gettop(engine->L) > 0) {
-			kr_log_error("%s\n", lua_tostring(engine->L, -1));
-		} else {
-			kr_log_error("[ ta ] keyfile '%s': failed to load (%s)\n",
-					keyfile, lua_strerror(lua_ret));
-		}
-		return lua_ret;
-	}
-
-	lua_settop(engine->L, 0);
-	return kr_ok();
-}
-
-
 static void args_init(struct args *args)
 {
 	memset(args, 0, sizeof(struct args));
+	/* Zeroed arrays are OK. */
 	args->forks = 1;
-	array_init(args->addr_set);
-	array_init(args->tls_set);
-	array_init(args->fd_set);
-	array_init(args->tls_fd_set);
-	args->moduledir = MODULEDIR;
 	args->control_fd = -1;
 	args->interactive = true;
 	args->quiet = false;
 }
 
+/* Free pointed-to resources. */
+static void args_deinit(struct args *args)
+{
+	array_clear(args->addrs);
+	array_clear(args->addrs_tls);
+	for (int i = 0; i < args->fds.len; ++i)
+		free_const(args->fds.at[i].flags.kind);
+	array_clear(args->fds);
+	array_clear(args->config);
+}
+
 static long strtol_10(const char *s)
 {
 	if (!s) abort();
@@ -550,57 +325,44 @@ static int parse_args(int argc, char **argv, struct args *args)
 	struct option opts[] = {
 		{"addr",       required_argument, 0, 'a'},
 		{"tls",        required_argument, 0, 't'},
-		{"fd",         required_argument, 0, 'S'},
-		{"tlsfd",      required_argument, 0, 'T'},
 		{"config",     required_argument, 0, 'c'},
-		{"keyfile",    required_argument, 0, 'k'},
-		{"keyfile-ro", required_argument, 0, 'K'},
 		{"forks",      required_argument, 0, 'f'},
-		{"moduledir",  required_argument, 0, 'm'},
+		{"noninteractive",   no_argument, 0, 'n'},
 		{"verbose",          no_argument, 0, 'v'},
 		{"quiet",            no_argument, 0, 'q'},
 		{"version",          no_argument, 0, 'V'},
 		{"help",             no_argument, 0, 'h'},
+		{"fd",         required_argument, 0, 'S'},
 		{0, 0, 0, 0}
 	};
-	while ((c = getopt_long(argc, argv, "a:t:S:T:c:f:m:K:k:vqVh", opts, &li)) != -1) {
+	while ((c = getopt_long(argc, argv, "a:t:c:f:nvqVhS:", opts, &li)) != -1) {
 		switch (c)
 		{
 		case 'a':
-			array_push(args->addr_set, optarg);
+			array_push(args->addrs, optarg);
 			break;
 		case 't':
-			array_push(args->tls_set, optarg);
-			break;
-		case 'S':
-			array_push(args->fd_set, strtol_10(optarg));
-			break;
-		case 'T':
-			array_push(args->tls_fd_set, strtol_10(optarg));
+			array_push(args->addrs_tls, optarg);
 			break;
 		case 'c':
-			args->config = optarg;
+			assert(optarg != NULL);
+			array_push(args->config, optarg);
 			break;
 		case 'f':
-			args->interactive = false;
 			args->forks = strtol_10(optarg);
+			if (args->forks == 1) {
+				kr_log_deprecate("use --noninteractive instead of --forks=1\n");
+			} else {
+				kr_log_deprecate("support for running multiple --forks will be removed\n");
+			}
 			if (args->forks <= 0) {
 				kr_log_error("[system] error '-f' requires a positive"
 						" number, not '%s'\n", optarg);
 				return EXIT_FAILURE;
 			}
-			break;
-		case 'K':
-			args->keyfile_unmanaged = 1;
-		case 'k':
-			if (args->keyfile != NULL) {
-				kr_log_error("[system] error only one of '--keyfile' and '--keyfile-ro' allowed\n");
-				return EXIT_FAILURE;
-			}
-			args->keyfile = optarg;
-			break;
-		case 'm':
-			args->moduledir = optarg;
+			/* fall through */
+		case 'n':
+			args->interactive = false;
 			break;
 		case 'v':
 			kr_verbose_set(true);
@@ -621,6 +383,28 @@ static int parse_args(int argc, char **argv, struct args *args)
 		default:
 			help(argc, argv);
 			return EXIT_FAILURE;
+		case 'S':
+			(void)0;
+			flagged_fd_t ffd = { 0 };
+			char *endptr;
+			ffd.fd = strtol(optarg, &endptr, 10);
+			if (endptr != optarg && endptr[0] == '\0') {
+				/* Plain DNS */
+				ffd.flags.tls = false;
+			} else if (endptr[0] == ':' && strcasecmp(endptr + 1, "tls") == 0) {
+				/* DoT */
+				ffd.flags.tls = true;
+				/* We know what .sock_type should be but it wouldn't help. */
+			} else if (endptr[0] == ':' && endptr[1] != '\0') {
+				/* Some other kind; no checks here. */
+				ffd.flags.kind = strdup(endptr + 1);
+			} else {
+				kr_log_error("[system] incorrect value passed to '-S/--fd': %s\n",
+						optarg);
+				return EXIT_FAILURE;
+			}
+			array_push(args->fds, ffd);
+			break;
 		}
 	}
 	if (optind < argc) {
@@ -629,107 +413,177 @@ static int parse_args(int argc, char **argv, struct args *args)
 	return -1;
 }
 
-static int bind_fds(struct network *net, fd_array_t *fd_set, bool tls) {
-	int ret = 0;
-	for (size_t i = 0; i < fd_set->len; ++i) {
-		ret = network_listen_fd(net, fd_set->at[i], tls);
+/** Just convert addresses to file-descriptors; clear *addrs on success.
+ * @note AF_UNIX is supported (starting with '/').
+ * @return zero or exit code for main()
+ */
+static int bind_sockets(addr_array_t *addrs, bool tls, flagged_fd_array_t *fds)
+{
+	bool has_error = false;
+	for (size_t i = 0; i < addrs->len; ++i) {
+		/* Get port and separate address string. */
+		uint16_t port = tls ? KR_DNS_TLS_PORT : KR_DNS_PORT;
+		char addr_buf[INET6_ADDRSTRLEN + 1];
+		int ret;
+		const char *addr_str;
+		const int family = kr_straddr_family(addrs->at[i]);
+		if (family == AF_UNIX) {
+			ret = 0;
+			addr_str = addrs->at[i];
+		} else { /* internet socket (or garbage) */
+			ret = kr_straddr_split(addrs->at[i], addr_buf, &port);
+			addr_str = addr_buf;
+		}
+		/* Get sockaddr. */
+		struct sockaddr *sa = NULL;
+		if (ret == 0) {
+			sa = kr_straddr_socket(addr_str, port, NULL);
+			if (!sa) ret = kr_error(EINVAL); /* could be ENOMEM but unlikely */
+		}
+		flagged_fd_t ffd = { .flags = { .tls = tls } };
+		if (ret == 0 && !tls && family != AF_UNIX) {
+			/* AF_UNIX can do SOCK_DGRAM, but let's not support that *here*. */
+			ffd.fd = io_bind(sa, SOCK_DGRAM, NULL);
+			if (ffd.fd < 0)
+				ret = ffd.fd;
+			else if (array_push(*fds, ffd) < 0)
+				ret = kr_error(ENOMEM);
+		}
+		if (ret == 0) { /* common for TCP and TLS, including AF_UNIX cases */
+			ffd.fd = io_bind(sa, SOCK_STREAM, NULL);
+			if (ffd.fd < 0)
+				ret = ffd.fd;
+			else if (array_push(*fds, ffd) < 0)
+				ret = kr_error(ENOMEM);
+		}
+		free(sa);
 		if (ret != 0) {
-			kr_log_error("[system] %slisten on fd=%d %s\n",
-				 tls ? "TLS " : "", fd_set->at[i], kr_strerror(ret));
-			break;
+			kr_log_error("[system] bind to '%s'%s: %s\n",
+				addrs->at[i], tls ? " (TLS)" : "", kr_strerror(ret));
+			has_error = true;
 		}
 	}
-	return ret;
+	array_clear(*addrs);
+	return has_error ? EXIT_FAILURE : kr_ok();
 }
 
-static int bind_sockets(struct network *net, addr_array_t *addr_set, bool tls) {	
-	uint32_t flags = tls ? NET_TCP|NET_TLS : NET_UDP|NET_TCP;
-	int ret = 0;
-	for (size_t i = 0; i < addr_set->len; ++i) {
-		int port = tls ? KR_DNS_TLS_PORT : KR_DNS_PORT;
-		const char *addr = set_addr(addr_set->at[i], &port);
-		ret = network_listen(net, addr, (uint16_t)port, flags);
+static int start_listening(struct network *net, flagged_fd_array_t *fds) {
+	int some_bad_ret = 0;
+	for (size_t i = 0; i < fds->len; ++i) {
+		flagged_fd_t *ffd = &fds->at[i];
+		int ret = network_listen_fd(net, ffd->fd, ffd->flags);
 		if (ret != 0) {
-			kr_log_error("[system] bind to '%s@%d' %s%s\n", 
-				addr, port, tls ? "(TLS) " : "", kr_strerror(ret));
-			break;
+			some_bad_ret = ret;
+			/* TODO: try logging address@port.  It's not too important,
+			 * because typical problems happen during binding already.
+			 * (invalid address, permission denied) */
+			kr_log_error("[system] listen on fd=%d: %s\n",
+					ffd->fd, kr_strerror(ret));
+			/* Continue printing all of these before exiting. */
+		} else {
+			ffd->flags.kind = NULL; /* ownership transferred */
 		}
 	}
-	return ret;
+	return some_bad_ret;
+}
+
+/* Drop POSIX 1003.1e capabilities. */
+static void drop_capabilities(void)
+{
+#ifdef ENABLE_CAP_NG
+	/* Drop all capabilities. */
+	if (capng_have_capability(CAPNG_EFFECTIVE, CAP_SETPCAP)) {
+		capng_clear(CAPNG_SELECT_BOTH);
+
+		/* Apply. */
+		if (capng_apply(CAPNG_SELECT_BOTH) < 0) {
+			kr_log_error("[system] failed to set process capabilities: %s\n",
+			          strerror(errno));
+		}
+	} else {
+		kr_log_info("[system] process not allowed to set capabilities, skipping\n");
+	}
+#endif /* ENABLE_CAP_NG */
 }
 
 int main(int argc, char **argv)
 {
-	int ret = 0;
-	struct args args;
-	args_init(&args);
-	if ((ret = parse_args(argc, argv, &args)) >= 0) {
-		return ret;
+	if (setvbuf(stdout, NULL, _IONBF, 0) || setvbuf(stderr, NULL, _IONBF, 0)) {
+		kr_log_error("[system] failed to to set output buffering (ignored): %s\n",
+				strerror(errno));
+		fflush(stderr);
 	}
+	if (strcmp("linux", OPERATING_SYSTEM) != 0)
+		kr_log_info("[warn] Knot Resolver is tested on Linux, other platforms might exhibit bugs.\n"
+				"Please report issues to https://gitlab.labs.nic.cz/knot/knot-resolver/issues/\n"
+				"Thank you for your time and interest!\n");
 
-#ifdef HAS_SYSTEMD
-	/* Accept passed sockets from systemd supervisor. */
-	char **socket_names = NULL;
-	int sd_nsocks = sd_listen_fds_with_names(0, &socket_names);
-	for (int i = 0; i < sd_nsocks; ++i) {
-		int fd = SD_LISTEN_FDS_START + i;
-		/* when run under systemd supervision, do not use interactive mode */
-		args.interactive = false;
-		if (args.forks != 1) {
-			kr_log_error("[system] when run under systemd-style supervision, "
-				     "use single-process only (bad: --forks=%d).\n", args.forks);
-			free_sd_socket_names(socket_names, sd_nsocks);
-			return EXIT_FAILURE;
-		}
-		if (!strcasecmp("control",socket_names[i])) {
-			args.control_fd = fd;
-		} else if (!strcasecmp("tls",socket_names[i])) {
-			array_push(args.tls_fd_set, fd);
-		} else {
-			array_push(args.fd_set, fd);
-		}
-	}
-	free_sd_socket_names(socket_names, sd_nsocks);
-#endif
+	the_args = &the_args_value;
+	args_init(the_args);
+	int ret = parse_args(argc, argv, the_args);
+	if (ret >= 0) goto cleanup_args;
+
+	ret = bind_sockets(&the_args->addrs, false, &the_args->fds);
+	if (ret) goto cleanup_args;
+	ret = bind_sockets(&the_args->addrs_tls, true, &the_args->fds);
+	if (ret) goto cleanup_args;
 
 	/* Switch to rundir. */
-	if (args.rundir != NULL) {
+	if (the_args->rundir != NULL) {
 		/* FIXME: access isn't a good way if we start as root and drop privileges later */
-		if (access(args.rundir, W_OK) != 0) {
-			kr_log_error("[system] rundir '%s': %s\n", args.rundir, strerror(errno));
-			return EXIT_FAILURE;
-		}
-		ret = chdir(args.rundir);
-		if (ret != 0) {
-			kr_log_error("[system] rundir '%s': %s\n", args.rundir, strerror(errno));
+		if (access(the_args->rundir, W_OK) != 0
+		    || chdir(the_args->rundir) != 0) {
+			kr_log_error("[system] rundir '%s': %s\n",
+					the_args->rundir, strerror(errno));
 			return EXIT_FAILURE;
 		}
 	}
 
-	if (args.config && strcmp(args.config, "-") != 0 && access(args.config, R_OK) != 0) {
-		kr_log_error("[system] config '%s': %s\n", args.config, strerror(errno));
-		return EXIT_FAILURE;
-	}
-	if (!args.config && access("config", R_OK) == 0) {
-		args.config = "config";
+	/* Select which config files to load and verify they are read-able. */
+	bool load_defaults = true;
+	size_t i = 0;
+	while (i < the_args->config.len) {
+		const char *config = the_args->config.at[i];
+		if (strcmp(config, "-") == 0) {
+			load_defaults = false;
+			array_del(the_args->config, i);
+			continue;  /* don't increment i */
+		} else if (access(config, R_OK) != 0) {
+			char cwd[PATH_MAX];
+			get_workdir(cwd, sizeof(cwd));
+			kr_log_error("[system] config '%s' (workdir '%s'): %s\n",
+				config, cwd, strerror(errno));
+			return EXIT_FAILURE;
+		}
+		i++;
 	}
+	if (the_args->config.len == 0 && access("config", R_OK) == 0)
+		array_push(the_args->config, "config");
+	if (load_defaults)
+		array_push(the_args->config, LIBDIR "/postconfig.lua");
 
-#ifndef CAN_FORK_EARLY
-	/* Forking is currently broken with libuv. We need libuv to bind to
-	 * sockets etc. before forking, but at the same time can't touch it before
-	 * forking otherwise it crashes, so it's a chicken and egg problem.
-	 * Disabling until https://github.com/libuv/libuv/pull/846 is done. */
-	 if (args.forks > 1 && args.fd_set.len == 0 && args.tls_fd_set.len == 0) {
-	 	kr_log_error("[system] forking >1 workers supported only on Linux 3.9+ or with supervisor\n");
-	 	return EXIT_FAILURE;
-	 }
-#endif
+	/* File-descriptor count limit: soft->hard. */
+	struct rlimit rlim;
+	ret = getrlimit(RLIMIT_NOFILE, &rlim);
+	if (ret == 0 && rlim.rlim_cur != rlim.rlim_max) {
+		kr_log_verbose("[system] increasing file-descriptor limit: %ld -> %ld\n",
+				(long)rlim.rlim_cur, (long)rlim.rlim_max);
+		rlim.rlim_cur = rlim.rlim_max;
+		ret = setrlimit(RLIMIT_NOFILE, &rlim);
+	}
+	if (ret) {
+		kr_log_error("[system] failed to get or set file-descriptor limit: %s\n",
+				strerror(errno));
+	} else if (rlim.rlim_cur < 512*1024) {
+		kr_log_info("[system] warning: hard limit for number of file-descriptors is only %ld but recommended value is 524288\n",
+				(long)rlim.rlim_cur);
+	}
 
 	/* Connect forks with local socket */
 	fd_array_t ipc_set;
 	array_init(ipc_set);
 	/* Fork subprocesses if requested */
-	int fork_id = fork_workers(&ipc_set, args.forks);
+	int fork_id = fork_workers(&ipc_set, the_args->forks);
 	if (fork_id < 0) {
 		return EXIT_FAILURE;
 	}
@@ -741,23 +595,22 @@ int main(int argc, char **argv)
 		.ctx = mp_new (4096),
 		.alloc = (knot_mm_alloc_t) mp_alloc
 	};
-	struct engine engine;
+	/** Static to work around lua_pushlightuserdata() limitations.
+	 * TODO: convert to a proper singleton like worker, most likely. */
+	static struct engine engine;
 	ret = engine_init(&engine, &pool);
 	if (ret != 0) {
 		kr_log_error("[system] failed to initialize engine: %s\n", kr_strerror(ret));
 		return EXIT_FAILURE;
 	}
-	/* Create worker */
-	struct worker_ctx *worker = worker_create(&engine, &pool, fork_id, args.forks);
-	if (!worker) {
-		kr_log_error("[system] not enough memory\n");
+	/* Initialize the worker */
+	ret = worker_init(&engine, fork_id, the_args->forks);
+	if (ret != 0) {
+		kr_log_error("[system] failed to initialize worker: %s\n", kr_strerror(ret));
 		return EXIT_FAILURE;
 	}
 
 	uv_loop_t *loop = uv_default_loop();
-	worker->loop = loop;
-	loop->data = worker;
-
 	/* Catch some signals. */
 	uv_signal_t sigint, sigterm;
 	if (true) ret = uv_signal_init(loop, &sigint);
@@ -783,58 +636,67 @@ int main(int argc, char **argv)
 		ret = EXIT_FAILURE;
 		goto cleanup;
 	}
+	/* Profiling: avoid SIGPROF waking up the event loop.  Otherwise the profiles
+	 * (of the usual type) may skew results, e.g. epoll_pwait() taking lots of time. */
+	ret = uv_loop_configure(loop, UV_LOOP_BLOCK_SIGNAL, SIGPROF);
+	if (ret) {
+		kr_log_info("[system] failed to block SIGPROF in event loop, ignoring: %s\n",
+				uv_strerror(ret));
+	}
 
-	/* Bind to passed fds and sockets*/
-	if (bind_fds(&engine.net, &args.fd_set, false) != 0 ||
-	    bind_fds(&engine.net, &args.tls_fd_set, true) != 0 ||
-	    bind_sockets(&engine.net, &args.addr_set, false) != 0 ||
-	    bind_sockets(&engine.net, &args.tls_set, true) != 0
-	) {
+	/* Start listening, in the sense of network_listen_fd(). */
+	if (start_listening(&engine.net, &the_args->fds) != 0) {
+		ret = EXIT_FAILURE;
+		goto cleanup;
+	}
+
+	ret = udp_queue_init_global(loop);
+	if (ret) {
+		kr_log_error("[system] failed to initialize UDP queue: %s\n",
+				kr_strerror(ret));
 		ret = EXIT_FAILURE;
 		goto cleanup;
 	}
 
 	/* Start the scripting engine */
-	engine_set_moduledir(&engine, args.moduledir);
-
 	if (engine_load_sandbox(&engine) != 0) {
 		ret = EXIT_FAILURE;
 		goto cleanup;
 	}
-	if (args.config != NULL && strcmp(args.config, "-") != 0) {
-		if(engine_loadconf(&engine, args.config) != 0) {
+
+	for (i = 0; i < the_args->config.len; ++i) {
+		const char *config = the_args->config.at[i];
+		if (engine_loadconf(&engine, config) != 0) {
 			ret = EXIT_FAILURE;
 			goto cleanup;
 		}
 		lua_settop(engine.L, 0);
 	}
-	if (args.keyfile != NULL && set_keyfile(&engine, args.keyfile, args.keyfile_unmanaged) != 0) {
-		ret = EXIT_FAILURE;
-		goto cleanup;
-	}
-	if (args.config == NULL || strcmp(args.config, "-") !=0) {
-		if(engine_load_defaults(&engine) != 0) {
-			ret = EXIT_FAILURE;
-			goto cleanup;
-		}
-	}
+
+	drop_capabilities();
+
 	if (engine_start(&engine) != 0) {
 		ret = EXIT_FAILURE;
 		goto cleanup;
 	}
 
+	if (network_engage_endpoints(&engine.net)) {
+		ret = EXIT_FAILURE;
+		goto cleanup;
+	}
+
 	/* Run the event loop */
-	ret = run_worker(loop, &engine, &ipc_set, fork_id == 0, &args);
+	ret = run_worker(loop, &engine, &ipc_set, fork_id == 0, the_args);
 
 cleanup:/* Cleanup. */
 	engine_deinit(&engine);
-	worker_reclaim(worker);
+	worker_deinit();
 	if (loop != NULL) {
-		uv_loop_close(loop);	
+		uv_loop_close(loop);
 	}
 	mp_delete(pool.ctx);
-	array_clear(args.addr_set);
-	array_clear(args.tls_set);
+cleanup_args:
+	args_deinit(the_args);
 	kr_crypto_cleanup();
 	return ret;
 }
