@@ -1,21 +1,13 @@
 /*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
+#include "kresconfig.h"
+#include "daemon/worker.h"
+
 #include <uv.h>
 #include <lua.h>
+#include <lauxlib.h>
 #include <libknot/packet/pkt.h>
 #include <libknot/descriptor.h>
 #include <contrib/ucw/lib.h>
@@ -28,15 +20,16 @@
 #include <sys/types.h>
 #include <unistd.h>
 #include <gnutls/gnutls.h>
-#include "lib/utils.h"
-#include "lib/layer.h"
-#include "daemon/worker.h"
-#include "daemon/bindings.h"
+
+#include "daemon/bindings/api.h"
 #include "daemon/engine.h"
 #include "daemon/io.h"
-#include "daemon/tls.h"
-#include "daemon/zimport.h"
 #include "daemon/session.h"
+#include "daemon/tls.h"
+#include "daemon/udp_queue.h"
+#include "daemon/zimport.h"
+#include "lib/layer.h"
+#include "lib/utils.h"
 
 
 /* Magic defaults for the worker. */
@@ -60,14 +53,14 @@
 struct request_ctx
 {
 	struct kr_request req;
-	struct {
-		union inaddr addr;
-		union inaddr dst_addr;
-		/* uv_handle_t *handle; */
 
+	struct {
+		/** Requestor's address; separate because of UDP session "sharing". */
+		union inaddr addr;
 		/** NULL if the request didn't come over network. */
 		struct session *session;
 	} source;
+
 	struct worker_ctx *worker;
 	struct qr_task *task;
 };
@@ -96,7 +89,12 @@ struct qr_task
 #define qr_task_ref(task) \
 	do { ++(task)->refs; } while(0)
 #define qr_task_unref(task) \
-	do { if (task && --(task)->refs == 0) { qr_task_free(task); } } while (0)
+	do { \
+		if (task) \
+			assert((task)->refs > 0); \
+		if ((task) && --(task)->refs == 0) \
+			qr_task_free((task)); \
+	} while (0)
 
 /** @internal get key for tcp session
  *  @note kr_straddr() return pointer to static string
@@ -109,7 +107,7 @@ static int qr_task_step(struct qr_task *task,
 			const struct sockaddr *packet_source,
 			knot_pkt_t *packet);
 static int qr_task_send(struct qr_task *task, struct session *session,
-			struct sockaddr *addr, knot_pkt_t *pkt);
+			const struct sockaddr *addr, knot_pkt_t *pkt);
 static int qr_task_finalize(struct qr_task *task, int state);
 static void qr_task_complete(struct qr_task *task);
 static struct session* worker_find_tcp_connected(struct worker_ctx *worker,
@@ -121,11 +119,8 @@ static struct session* worker_find_tcp_waiting(struct worker_ctx *worker,
 					       const struct sockaddr *addr);
 static void on_tcp_connect_timeout(uv_timer_t *timer);
 
-/** @internal Get singleton worker. */
-static inline struct worker_ctx *get_worker(void)
-{
-	return uv_default_loop()->data;
-}
+struct worker_ctx the_worker_value; /**< Static allocation is suitable for the singleton. */
+struct worker_ctx *the_worker = NULL;
 
 /*! @internal Create a UDP/TCP handle for an outgoing AF_INET* connection.
  *  socktype is SOCK_* */
@@ -263,8 +258,8 @@ static int subreq_key(char *dst, knot_pkt_t *pkt)
  * in case the request didn't come from network.
  */
 static struct request_ctx *request_create(struct worker_ctx *worker,
-					  uv_handle_t *handle,
-					  const struct sockaddr *addr,
+					  struct session *session,
+					  const struct sockaddr *peer,
 					  uint32_t uid)
 {
 	knot_mm_t pool = {
@@ -283,50 +278,27 @@ static struct request_ctx *request_create(struct worker_ctx *worker,
 
 	/* TODO Relocate pool to struct request */
 	ctx->worker = worker;
-	struct session *s = handle ? handle->data : NULL;
-	if (s) {
-		assert(session_flags(s)->outgoing == false);
+	if (session) {
+		assert(session_flags(session)->outgoing == false);
 	}
-	ctx->source.session = s;
+	ctx->source.session = session;
 
 	struct kr_request *req = &ctx->req;
 	req->pool = pool;
 	req->vars_ref = LUA_NOREF;
 	req->uid = uid;
-	req->daemon_context = worker;
-
-	/* Remember query source addr */
-	if (!addr || (addr->sa_family != AF_INET && addr->sa_family != AF_INET6)) {
-		ctx->source.addr.ip.sa_family = AF_UNSPEC;
-	} else {
-		memcpy(&ctx->source.addr, addr, kr_sockaddr_len(addr));
-		ctx->req.qsource.addr = &ctx->source.addr.ip;
+	if (session) {
+		/* We assume the session will be alive during the whole life of the request. */
+		req->qsource.dst_addr = session_get_sockname(session);
+		req->qsource.flags.tcp = session_get_handle(session)->type == UV_TCP;
+		req->qsource.flags.tls = session_flags(session)->has_tls;
+		/* We need to store a copy of peer address. */
+		memcpy(&ctx->source.addr.ip, peer, kr_sockaddr_len(peer));
+		req->qsource.addr = &ctx->source.addr.ip;
 	}
 
 	worker->stats.rconcurrent += 1;
 
-	if (!handle) {
-		return ctx;
-	}
-
-	/* Remember the destination address. */
-	int addr_len = sizeof(ctx->source.dst_addr);
-	struct sockaddr *dst_addr = &ctx->source.dst_addr.ip;
-	ctx->source.dst_addr.ip.sa_family = AF_UNSPEC;
-	if (handle->type == UV_UDP) {
-		if (uv_udp_getsockname((uv_udp_t *)handle, dst_addr, &addr_len) == 0) {
-			req->qsource.dst_addr = dst_addr;
-		}
-		req->qsource.flags.tcp = false;
-		req->qsource.flags.tls = false;
-	} else if (handle->type == UV_TCP) {
-		if (uv_tcp_getsockname((uv_tcp_t *)handle, dst_addr, &addr_len) == 0) {
-			req->qsource.dst_addr = dst_addr;
-		}
-		req->qsource.flags.tcp = true;
-		req->qsource.flags.tls = s && session_flags(s)->has_tls;
-	}
-
 	return ctx;
 }
 
@@ -398,26 +370,18 @@ static void request_free(struct request_ctx *ctx)
 	/* Return mempool to ring or free it if it's full */
 	pool_release(worker, ctx->req.pool.ctx);
 	/* @note The 'task' is invalidated from now on. */
-	/* Decommit memory every once in a while */
-	static int mp_delete_count = 0;
-	if (++mp_delete_count == 100000) {
-		lua_gc(worker->engine->L, LUA_GCCOLLECT, 0);
-#if defined(__GLIBC__) && defined(_GNU_SOURCE)
-		malloc_trim(0);
-#endif
-		mp_delete_count = 0;
-	}
 	worker->stats.rconcurrent -= 1;
 }
 
 static struct qr_task *qr_task_create(struct request_ctx *ctx)
 {
-	/* How much can client handle? */
-	struct engine *engine = ctx->worker->engine;
-	size_t pktbuf_max = KR_EDNS_PAYLOAD;
-	if (engine->resolver.opt_rr) {
-		pktbuf_max = MAX(knot_edns_get_payload(engine->resolver.opt_rr),
-				 pktbuf_max);
+	/* Choose (initial) pktbuf size.  As it is now, pktbuf can be used
+	 * for UDP answers from upstream *and* from cache
+	 * and for sending non-UDP queries upstream (?) */
+	uint16_t pktbuf_max = KR_EDNS_PAYLOAD;
+	const knot_rrset_t *opt_our = ctx->worker->engine->resolver.opt_rr;
+	if (opt_our) {
+		pktbuf_max = MAX(pktbuf_max, knot_edns_get_payload(opt_our));
 	}
 
 	/* Create resolution task */
@@ -513,7 +477,7 @@ static void qr_task_complete(struct qr_task *task)
 }
 
 /* This is called when we send subrequest / answer */
-static int qr_task_on_send(struct qr_task *task, uv_handle_t *handle, int status)
+int qr_task_on_send(struct qr_task *task, uv_handle_t *handle, int status)
 {
 
 	if (task->finished) {
@@ -566,7 +530,7 @@ static void on_write(uv_write_t *req, int status)
 }
 
 static int qr_task_send(struct qr_task *task, struct session *session,
-			struct sockaddr *addr, knot_pkt_t *pkt)
+			const struct sockaddr *addr, knot_pkt_t *pkt)
 {
 	if (!session) {
 		return qr_task_on_send(task, NULL, kr_error(EIO));
@@ -625,13 +589,28 @@ static int qr_task_send(struct qr_task *task, struct session *session,
 		ret = uv_udp_send(send_req, (uv_udp_t *)handle, &buf, 1, addr, &on_send);
 	} else if (handle->type == UV_TCP) {
 		uv_write_t *write_req = (uv_write_t *)ioreq;
-		uint16_t pkt_size = htons(pkt->size);
-		uv_buf_t buf[2] = {
-			{ (char *)&pkt_size, sizeof(pkt_size) },
-			{ (char *)pkt->wire, pkt->size }
+		/* We need to write message length in native byte order,
+		 * but we don't have a convenient place to store those bytes.
+		 * The problem is that all memory referenced from buf[] MUST retain
+		 * its contents at least until on_write() is called, and I currently
+		 * can't see any convenient place outside the `pkt` structure.
+		 * So we use directly the *individual* bytes in pkt->size.
+		 * The call to htonl() and the condition will probably be inlinable. */
+		int lsbi, slsbi; /* (second) least significant byte index */
+		if (htonl(1) == 1) { /* big endian */
+			lsbi  = sizeof(pkt->size) - 1;
+			slsbi = sizeof(pkt->size) - 2;
+		} else {
+			lsbi  = 0;
+			slsbi = 1;
+		}
+		uv_buf_t buf[3] = {
+			{ (char *)&pkt->size + slsbi, 1 },
+			{ (char *)&pkt->size + lsbi,  1 },
+			{ (char *)pkt->wire, pkt->size },
 		};
 		write_req->data = task;
-		ret = uv_write(write_req, (uv_stream_t *)handle, buf, 2, &on_write);
+		ret = uv_write(write_req, (uv_stream_t *)handle, buf, 3, &on_write);
 	} else {
 		assert(false);
 	}
@@ -718,7 +697,7 @@ static int session_tls_hs_cb(struct session *session, int status)
 
 	/* handshake was completed successfully */
 	struct tls_client_ctx_t *tls_client_ctx = session_tls_get_client_ctx(session);
-	struct tls_client_paramlist_entry *tls_params = tls_client_ctx->params;
+	tls_client_param_t *tls_params = tls_client_ctx->params;
 	gnutls_session_t tls_session = tls_client_ctx->c.tls_session;
 	if (gnutls_session_is_resumed(tls_session) != 0) {
 		kr_log_verbose("[tls_client] TLS session has resumed\n");
@@ -813,7 +792,8 @@ static int send_waiting(struct session *session)
 
 static void on_connect(uv_connect_t *req, int status)
 {
-	struct worker_ctx *worker = get_worker();
+	struct worker_ctx *worker = the_worker;
+	assert(worker);
 	uv_stream_t *handle = req->handle;
 	struct session *session = handle->data;
 	struct sockaddr *peer = session_get_peer(session);
@@ -938,7 +918,8 @@ static void on_tcp_connect_timeout(uv_timer_t *timer)
 	struct session *session = timer->data;
 
 	uv_timer_stop(timer);
-	struct worker_ctx *worker = get_worker();
+	struct worker_ctx *worker = the_worker;
+	assert(worker);
 
 	assert (session_tasklist_is_empty(session));
 
@@ -1166,10 +1147,24 @@ static int qr_task_finalize(struct qr_task *task, int state)
 	/* Send back answer */
 	assert(!session_flags(source_session)->closing);
 	assert(ctx->source.addr.ip.sa_family != AF_UNSPEC);
-	int res = qr_task_send(task, source_session,
-			       (struct sockaddr *)&ctx->source.addr,
-			        ctx->req.answer);
-	if (res != kr_ok()) {
+
+	int ret;
+	const uv_handle_t *src_handle = session_get_handle(source_session);
+	if (src_handle->type != UV_UDP && src_handle->type != UV_TCP) {
+		assert(false);
+		ret = kr_error(EINVAL);
+	} else if (src_handle->type == UV_UDP && ENABLE_SENDMMSG) {
+		int fd;
+		ret = uv_fileno(src_handle, &fd);
+		assert(!ret);
+		if (ret == 0) {
+			udp_queue_push(fd, &ctx->req, task);
+		}
+	} else {
+		ret = qr_task_send(task, source_session, &ctx->source.addr.ip, ctx->req.answer);
+	}
+
+	if (ret != kr_ok()) {
 		(void) qr_task_on_send(task, NULL, kr_error(EIO));
 		/* Since source session is erroneous detach all tasks. */
 		while (!session_tasklist_is_empty(source_session)) {
@@ -1298,11 +1293,9 @@ static int tcp_task_make_connection(struct qr_task *task, const struct sockaddr
 	struct worker_ctx *worker = ctx->worker;
 
 	/* Check if there must be TLS */
-	struct engine *engine = worker->engine;
-	struct network *net = &engine->net;
-	const char *key = tcpsess_key(addr);
 	struct tls_client_ctx_t *tls_ctx = NULL;
-	struct tls_client_paramlist_entry *entry = map_get(&net->tls_client_params, key);
+	struct network *net = &worker->engine->net;
+	tls_client_param_t *entry = tls_client_param_get(net->tls_client_params, addr);
 	if (entry) {
 		/* Address is configured to be used with TLS.
 		 * We need to allocate auxiliary data structure. */
@@ -1333,7 +1326,7 @@ static int tcp_task_make_connection(struct qr_task *task, const struct sockaddr
 
 	/* Add address to the waiting list.
 	 * Now it "is waiting to be connected to." */
-	int ret = worker_add_tcp_waiting(ctx->worker, addr, session);
+	int ret = worker_add_tcp_waiting(worker, addr, session);
 	if (ret < 0) {
 		free(conn);
 		session_close(session);
@@ -1349,7 +1342,7 @@ static int tcp_task_make_connection(struct qr_task *task, const struct sockaddr
 	ret = session_timer_start(session, on_tcp_connect_timeout,
 				  KR_CONN_RTT_MAX, 0);
 	if (ret != 0) {
-		worker_del_tcp_waiting(ctx->worker, addr);
+		worker_del_tcp_waiting(worker, addr);
 		free(conn);
 		session_close(session);
 		return kr_error(EINVAL);
@@ -1365,7 +1358,7 @@ static int tcp_task_make_connection(struct qr_task *task, const struct sockaddr
 	ret = uv_tcp_connect(conn, (uv_tcp_t *)client, addr , on_connect);
 	if (ret != 0) {
 		session_timer_stop(session);
-		worker_del_tcp_waiting(ctx->worker, addr);
+		worker_del_tcp_waiting(worker, addr);
 		free(conn);
 		session_close(session);
 		unsigned score = qry->flags.FORWARD || qry->flags.STUB ? KR_NS_FWD_DEAD : KR_NS_DEAD;
@@ -1385,7 +1378,7 @@ static int tcp_task_make_connection(struct qr_task *task, const struct sockaddr
 	ret = session_waitinglist_push(session, task);
 	if (ret < 0) {
 		session_timer_stop(session);
-		worker_del_tcp_waiting(ctx->worker, addr);
+		worker_del_tcp_waiting(worker, addr);
 		free(conn);
 		session_close(session);
 		return kr_error(EINVAL);
@@ -1486,6 +1479,18 @@ static int qr_task_step(struct qr_task *task,
 					   &sock_type, task->pktbuf);
 		if (unlikely(++task->iter_count > KR_ITER_LIMIT ||
 			     task->timeouts >= KR_TIMEOUT_LIMIT)) {
+
+			#ifndef NOVERBOSELOG
+			struct kr_rplan *rplan = &req->rplan;
+			struct kr_query *last  = kr_rplan_last(rplan);
+			if (task->iter_count > KR_ITER_LIMIT) {
+				VERBOSE_MSG(last, "canceling query due to exceeded iteration count limit of %d\n", KR_ITER_LIMIT);
+			}
+			if (task->timeouts >= KR_TIMEOUT_LIMIT) {
+				VERBOSE_MSG(last, "canceling query due to exceeded timeout retries limit of %d\n", KR_TIMEOUT_LIMIT);
+			}
+			#endif
+
 			return qr_task_finalize(task, KR_STATE_FAIL);
 		}
 	}
@@ -1508,16 +1513,19 @@ static int qr_task_step(struct qr_task *task,
 	if (task->addrlist_count > 0 && kr_inaddr_port(task->addrlist) == KR_DNS_PORT) {
 		/* TODO if there are multiple addresses (task->addrlist_count > 1)
 		 * check all of them. */
-		struct engine *engine = worker->engine;
-		struct network *net = &engine->net;
-		struct tls_client_paramlist_entry *tls_entry =
-			tls_client_try_upgrade(&net->tls_client_params, task->addrlist);
-		if (tls_entry != NULL) {
-			kr_inaddr_set_port(task->addrlist, KR_DNS_TLS_PORT);
+		struct network *net = &worker->engine->net;
+		/* task->addrlist has to contain TLS port before tls_client_param_get() call */
+		kr_inaddr_set_port(task->addrlist, KR_DNS_TLS_PORT);
+		tls_client_param_t *tls_entry =
+			tls_client_param_get(net->tls_client_params, task->addrlist);
+		if (tls_entry) {
 			packet_source = NULL;
 			sock_type = SOCK_STREAM;
 			/* TODO in this case in tcp_task_make_connection() will be performed
 			 * redundant map_get() call. */
+		} else {
+			/* The function is fairly cheap, so we just change there and back. */
+			kr_inaddr_set_port(task->addrlist, KR_DNS_PORT);
 		}
 	}
 
@@ -1554,7 +1562,7 @@ static int parse_packet(knot_pkt_t *query)
 	return ret;
 }
 
-int worker_submit(struct session *session, knot_pkt_t *query)
+int worker_submit(struct session *session, const struct sockaddr *peer, knot_pkt_t *query)
 {
 	if (!session) {
 		assert(false);
@@ -1586,10 +1594,9 @@ int worker_submit(struct session *session, knot_pkt_t *query)
 	/* Start new task on listening sockets,
 	 * or resume if this is subrequest */
 	struct qr_task *task = NULL;
-	struct sockaddr *addr = NULL;
+	const struct sockaddr *addr = NULL;
 	if (!is_outgoing) { /* request from a client */
-		struct request_ctx *ctx = request_create(worker, handle,
-							 session_get_peer(session),
+		struct request_ctx *ctx = request_create(worker, session, peer,
 							 knot_wire_get_id(query->wire));
 		if (!ctx) {
 			return kr_error(ENOMEM);
@@ -1611,12 +1618,15 @@ int worker_submit(struct session *session, knot_pkt_t *query)
 			return kr_error(ENOMEM);
 		}
 	} else if (query) { /* response from upstream */
-		task = session_tasklist_del_msgid(session, knot_wire_get_id(query->wire));
+		const uint16_t id = knot_wire_get_id(query->wire);
+		task = session_tasklist_del_msgid(session, id);
 		if (task == NULL) {
+			VERBOSE_MSG(NULL, "=> ignoring packet with mismatching ID %d\n",
+					(int)id);
 			return kr_error(ENOENT);
 		}
 		assert(!session_flags(session)->closing);
-		addr = session_get_peer(session);
+		addr = peer;
 	}
 	assert(uv_is_closing(session_get_handle(session)) == false);
 
@@ -1716,7 +1726,7 @@ int worker_end_tcp(struct session *session)
 	}
 
 	session_timer_stop(session);
-	
+
 	uv_handle_t *handle = session_get_handle(session);
 	struct worker_ctx *worker = handle->loop->data;
 	struct sockaddr *peer = session_get_peer(session);
@@ -1779,8 +1789,49 @@ int worker_end_tcp(struct session *session)
 	return kr_ok();
 }
 
-struct qr_task *worker_resolve_start(struct worker_ctx *worker, knot_pkt_t *query, struct kr_qflags options)
+knot_pkt_t * worker_resolve_mk_pkt(const char *qname_str, uint16_t qtype, uint16_t qclass,
+				   const struct kr_qflags *options)
 {
+	uint8_t qname[KNOT_DNAME_MAXLEN];
+	if (!knot_dname_from_str(qname, qname_str, sizeof(qname)))
+		return NULL;
+	knot_pkt_t *pkt = knot_pkt_new(NULL, KNOT_EDNS_MAX_UDP_PAYLOAD, NULL);
+	if (!pkt)
+		return NULL;
+	knot_pkt_put_question(pkt, qname, qclass, qtype);
+	knot_wire_set_rd(pkt->wire);
+	knot_wire_set_ad(pkt->wire);
+
+	/* Add OPT RR, including wire format so modules can see both representations.
+	 * knot_pkt_put() copies the outside; we need to duplicate the inside manually. */
+	knot_rrset_t *opt = knot_rrset_copy(the_worker->engine->resolver.opt_rr, NULL);
+	if (!opt) {
+		knot_pkt_free(pkt);
+		return NULL;
+	}
+	if (options->DNSSEC_WANT) {
+		knot_edns_set_do(opt);
+	}
+	knot_pkt_begin(pkt, KNOT_ADDITIONAL);
+	int ret = knot_pkt_put(pkt, KNOT_COMPR_HINT_NONE, opt, KNOT_PF_FREE);
+	if (ret == KNOT_EOK) {
+		free(opt); /* inside is owned by pkt now */
+	} else {
+		knot_rrset_free(opt, NULL);
+		knot_pkt_free(pkt);
+		return NULL;
+	}
+
+	if (options->DNSSEC_CD) {
+		knot_wire_set_cd(pkt->wire);
+	}
+
+	return pkt;
+}
+
+struct qr_task *worker_resolve_start(knot_pkt_t *query, struct kr_qflags options)
+{
+	struct worker_ctx *worker = the_worker;
 	if (!worker || !query) {
 		assert(!EINVAL);
 		return NULL;
@@ -1922,21 +1973,22 @@ bool worker_task_finished(struct qr_task *task)
 {
 	return task->finished;
 }
-/** Reserve worker buffers */
+
+/** Reserve worker buffers.  We assume worker's been zeroed. */
 static int worker_reserve(struct worker_ctx *worker, size_t ring_maxlen)
 {
+	worker->tcp_connected = map_make(NULL);
+	worker->tcp_waiting = map_make(NULL);
+	worker->subreq_out = trie_create(NULL);
+
 	array_init(worker->pool_mp);
 	if (array_reserve(worker->pool_mp, ring_maxlen)) {
 		return kr_error(ENOMEM);
 	}
-	memset(&worker->pkt_pool, 0, sizeof(worker->pkt_pool));
+
 	worker->pkt_pool.ctx = mp_new (4 * sizeof(knot_pkt_t));
 	worker->pkt_pool.alloc = (knot_mm_alloc_t) mp_alloc;
-	worker->subreq_out = trie_create(NULL);
-	worker->tcp_connected = map_make(NULL);
-	worker->tcp_waiting = map_make(NULL);
-	worker->tcp_pipeline_max = MAX_PIPELINED;
-	memset(&worker->stats, 0, sizeof(worker->stats));
+
 	return kr_ok();
 }
 
@@ -1950,47 +2002,59 @@ static inline void reclaim_mp_freelist(mp_freelist_t *list)
 	array_clear(*list);
 }
 
-void worker_reclaim(struct worker_ctx *worker)
+void worker_deinit(void)
 {
-	reclaim_mp_freelist(&worker->pool_mp);
-	mp_delete(worker->pkt_pool.ctx);
-	worker->pkt_pool.ctx = NULL;
-	trie_free(worker->subreq_out);
-	worker->subreq_out = NULL;
-	map_clear(&worker->tcp_connected);
-	map_clear(&worker->tcp_waiting);
+	struct worker_ctx *worker = the_worker;
+	assert(worker);
 	if (worker->z_import != NULL) {
 		zi_free(worker->z_import);
 		worker->z_import = NULL;
 	}
+	map_clear(&worker->tcp_connected);
+	map_clear(&worker->tcp_waiting);
+	trie_free(worker->subreq_out);
+	worker->subreq_out = NULL;
+
+	reclaim_mp_freelist(&worker->pool_mp);
+	mp_delete(worker->pkt_pool.ctx);
+	worker->pkt_pool.ctx = NULL;
+
+	the_worker = NULL;
 }
 
-struct worker_ctx *worker_create(struct engine *engine, knot_mm_t *pool,
-		int worker_id, int worker_count)
+int worker_init(struct engine *engine, int worker_id, int worker_count)
 {
-	/* Load bindings */
-	engine_lualib(engine, "modules", lib_modules);
-	engine_lualib(engine, "net",     lib_net);
-	engine_lualib(engine, "cache",   lib_cache);
-	engine_lualib(engine, "event",   lib_event);
-	engine_lualib(engine, "worker",  lib_worker);
+	assert(engine && engine->L);
+	assert(the_worker == NULL);
+	kr_bindings_register(engine->L);
 
 	/* Create main worker. */
-	struct worker_ctx *worker = mm_alloc(pool, sizeof(*worker));
-	if (!worker) {
-		return NULL;
-	}
+	struct worker_ctx *worker = &the_worker_value;
 	memset(worker, 0, sizeof(*worker));
+	worker->engine = engine;
+
+	uv_loop_t *loop = uv_default_loop();
+	worker->loop = loop;
+
 	worker->id = worker_id;
 	worker->count = worker_count;
-	worker->engine = engine;
-	worker->next_request_uid = UINT16_MAX + 1;
-	worker_reserve(worker, MP_FREELIST_SIZE);
+
+	/* Register table for worker per-request variables */
+	lua_newtable(engine->L);
+	lua_setfield(engine->L, -2, "vars");
+	lua_getfield(engine->L, -1, "vars");
+	worker->vars_table_ref = luaL_ref(engine->L, LUA_REGISTRYINDEX);
+	lua_pop(engine->L, 1);
+
+	worker->tcp_pipeline_max = MAX_PIPELINED;
 	worker->out_addr4.sin_family = AF_UNSPEC;
 	worker->out_addr6.sin6_family = AF_UNSPEC;
-	/* Register worker in Lua thread */
-	lua_pushlightuserdata(engine->L, worker);
-	lua_setglobal(engine->L, "__worker");
+
+	int ret = worker_reserve(worker, MP_FREELIST_SIZE);
+	if (ret) return ret;
+	worker->next_request_uid = UINT16_MAX + 1;
+
+	/* Set some worker.* fields in Lua */
 	lua_getglobal(engine->L, "worker");
 	lua_pushnumber(engine->L, worker_id);
 	lua_setfield(engine->L, -2, "id");
@@ -1998,13 +2062,16 @@ struct worker_ctx *worker_create(struct engine *engine, knot_mm_t *pool,
 	lua_setfield(engine->L, -2, "pid");
 	lua_pushnumber(engine->L, worker_count);
 	lua_setfield(engine->L, -2, "count");
-	/* Register table for worker per-request variables */
-	lua_newtable(engine->L);
-	lua_setfield(engine->L, -2, "vars");
-	lua_getfield(engine->L, -1, "vars");
-	worker->vars_table_ref = luaL_ref(engine->L, LUA_REGISTRYINDEX);
-	lua_pop(engine->L, 1);
-	return worker;
+
+	char cwd[PATH_MAX];
+	get_workdir(cwd, sizeof(cwd));
+	lua_pushstring(engine->L, cwd);
+	lua_setfield(engine->L, -2, "cwd");
+
+	the_worker = worker;
+	loop->data = the_worker;
+	/* ^^^^ This shouldn't be used anymore, but it's hard to be 100% sure. */
+	return kr_ok();
 }
 
 #undef VERBOSE_MSG
