@@ -1,17 +1,5 @@
 /*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 /* Module is intended to import resource records from file into resolver's cache.
@@ -338,10 +326,6 @@ static knot_pkt_t *zi_query_create(zone_import_ctx_t *z_import, knot_rrset_t *rr
  * @return -1 if failed; 0 if success */
 static int zi_rrset_import(zone_import_ctx_t *z_import, knot_rrset_t *rr)
 {
-	struct worker_ctx *worker = z_import->worker;
-
-	assert(worker);
-
 	/* Create "pseudo query" which asks for given rrset. */
 	knot_pkt_t *query = zi_query_create(z_import, rr);
 	if (!query) {
@@ -369,7 +353,7 @@ static int zi_rrset_import(zone_import_ctx_t *z_import, knot_rrset_t *rr)
 
 	/* This call creates internal structures which necessary for
 	 * resolving - qr_task & request_ctx. */
-	struct qr_task *task = worker_resolve_start(worker, query, options);
+	struct qr_task *task = worker_resolve_start(query, options);
 	if (!task) {
 		knot_pkt_free(query);
 		knot_pkt_free(answer);
@@ -486,7 +470,7 @@ static void zi_zone_process(uv_timer_t* handle)
 	}
 
 	if (z_import->rrset_sorted.len <= 0) {
-		VERBOSE_MSG(NULL, "zone is empty\n");
+		kr_log_error("[zimport] zone `%s` is empty\n", zone_name_str);
 		goto finish;
 	}
 
@@ -510,6 +494,15 @@ static void zi_zone_process(uv_timer_t* handle)
 	}
 	z_import->key = rr_key;
 
+	map_t *trust_anchors = &z_import->worker->engine->resolver.trust_anchors;
+	knot_rrset_t *rr_ta = kr_ta_get(trust_anchors, z_import->origin);
+	if (!rr_ta) {
+		kr_log_error("[zimport] error: TA for zone `%s` vanished, fail", zone_name_str);
+		failed = 1;
+		goto finish;
+	}
+	z_import->ta = rr_ta;
+
 	VERBOSE_MSG(NULL, "started: zone: '%s'\n", zone_name_str);
 
 	z_import->start_timestamp = kr_now();
@@ -524,7 +517,7 @@ static void zi_zone_process(uv_timer_t* handle)
 
 	int res = zi_rrset_import(z_import, rr_key);
 	if (res != 0) {
-		VERBOSE_MSG(NULL, "import failed: qname: '%s' type: '%s'\n",
+		kr_log_error("[zimport] import failed: qname: '%s' type: '%s'\n",
 			    kname_str, ktype_str);
 		failed = 1;
 		goto finish;
@@ -601,7 +594,7 @@ finish:
 	if (failed != 0) {
 		if (ns_imported == 0 && other_imported == 0) {
 			import_state = -1;
-			VERBOSE_MSG(NULL, "import failed; zone `%s` \n", zone_name_str);
+			kr_log_error("[zimport] import failed; zone `%s` \n", zone_name_str);
 		} else {
 			import_state = 1;
 		}
@@ -682,6 +675,7 @@ static int zi_record_store(zs_scanner_t *s)
 /** @internal zscanner callback. */
 static int zi_state_parsing(zs_scanner_t *s)
 {
+	bool empty = true;
 	while (zs_parse_record(s) == 0) {
 		switch (s->state) {
 		case ZS_STATE_DATA:
@@ -689,14 +683,10 @@ static int zi_state_parsing(zs_scanner_t *s)
 				return -1;
 			}
 			zone_import_ctx_t *z_import = (zone_import_ctx_t *) s->process.data;
-			if (z_import->origin == 0) {
-				z_import->origin = knot_dname_copy(s->zone_origin,
-								  &z_import->pool);
-			} else if (!knot_dname_is_equal(z_import->origin, s->zone_origin)) {
-				kr_log_error("[zscanner] line: %"PRIu64
-					     ": zone origin changed unexpectedly\n",
-					     s->line_counter);
-				return -1;
+			empty = false;
+			if (s->r_type == 6) {
+				z_import->origin = knot_dname_copy(s->r_owner,
+                                                                 &z_import->pool);
 			}
 			break;
 		case ZS_STATE_ERROR:
@@ -712,6 +702,14 @@ static int zi_state_parsing(zs_scanner_t *s)
 			return -1;
 		case ZS_STATE_EOF:
 		case ZS_STATE_STOP:
+			if (empty) {
+				kr_log_error("[zimport] empty zone file\n");
+				return -1;
+			}
+			if (!((zone_import_ctx_t *) s->process.data)->origin) {
+				kr_log_error("[zimport] zone file doesn't contain SOA record\n");
+				return -1;
+			}
 			return (s->error.counter == 0) ? 0 : -1;
 		default:
 			kr_log_error("[zscanner] line: %"PRIu64
@@ -777,20 +775,20 @@ int zi_zone_import(struct zone_import_ctx *z_import,
 		VERBOSE_MSG(NULL, "[zscanner] started; zone file `%s`\n",
 			    zone_file);
 		ret = zi_state_parsing(s);
-		/* Try to find TA for worker->z_import.origin. */
-		map_t *trust_anchors = &z_import->worker->engine->resolver.trust_anchors;
-		knot_rrset_t *rr = kr_ta_get(trust_anchors, z_import->origin);
-		if (rr) {
-			z_import->ta = rr;
-		} else {
-			/* For now - fail.
-			 * TODO - query DS and continue after answer had been obtained. */
-			KR_DNAME_GET_STR(zone_name_str, z_import->origin);
-			kr_log_error("[zimport] no TA found for `%s`, fail\n", zone_name_str);
-			ret = 1;
+		if (ret == 0) {
+			/* Try to find TA for worker->z_import.origin. */
+			map_t *trust_anchors = &z_import->worker->engine->resolver.trust_anchors;
+			knot_rrset_t *rr = kr_ta_get(trust_anchors, z_import->origin);
+			if (!rr) {
+				/* For now - fail.
+				 * TODO - query DS and continue after answer had been obtained. */
+				KR_DNAME_GET_STR(zone_name_str, z_import->origin);
+				kr_log_error("[zimport] no TA found for `%s`, fail\n", zone_name_str);
+				ret = 1;
+			}
+			elapsed = kr_now() - z_import->start_timestamp;
+			elapsed = elapsed > UINT_MAX ? UINT_MAX : elapsed;
 		}
-		elapsed = kr_now() - z_import->start_timestamp;
-		elapsed = elapsed > UINT_MAX ? UINT_MAX : elapsed;
 	}
 	zs_deinit(s);
 	free(s);
