@@ -1,17 +1,5 @@
 /*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 #include <assert.h>
@@ -27,6 +15,7 @@
 #include <libknot/errcode.h>
 #include <libknot/rrtype/rrsig.h>
 
+#include "contrib/base32hex.h"
 #include "contrib/cleanup.h"
 #include "contrib/ucw/lib.h"
 #include "lib/cache/api.h"
@@ -56,20 +45,22 @@ static const uint16_t CACHE_VERSION = 5;
 
 
 /** @internal Forward declarations of the implementation details
- * \param optout[out] Set *optout = true; when encountering an opt-out NSEC3 (optional). */
+ * \param needs_pkt[out] optionally set *needs_pkt = true;
+ * 	We do that when some RRset wasn't stashed to aggressive cache,
+ * 	even though it might have taken part in a successful DNSSEC proof:
+ * 	1. any opt-out NSEC3, as they typically aren't much use aggressively anyway
+ * 	2. some kinds of minimal NSEC* ranges, as they'd seem more trouble than worth:
+ * 	    - extremely short range of covered names limits the benefits severly
+ * 	    - the type-set is often a lie, either a working lie, e.g. CloudFlare's
+ * 	      black lies, or even a non-working lie, e.g. DVE-2018-0003
+ * 	3. some kinds of "weird" RRsets, to get at least some caching on them
+ */
 static ssize_t stash_rrset(struct kr_cache *cache, const struct kr_query *qry,
 		const knot_rrset_t *rr, const knot_rrset_t *rr_sigs, uint32_t timestamp,
-		uint8_t rank, trie_t *nsec_pmap, bool *has_optout);
+		uint8_t rank, trie_t *nsec_pmap, bool *needs_pkt);
 /** Preliminary checks before stash_rrset().  Don't call if returns <= 0. */
 static int stash_rrset_precond(const knot_rrset_t *rr, const struct kr_query *qry/*logs*/);
 
-/** @internal Removes all records from cache. */
-static inline int cache_clear(struct kr_cache *cache)
-{
-	cache->stats.delete += 1;
-	return cache_op(cache, clear);
-}
-
 /** @internal Open cache db transaction and check internal data version. */
 static int assert_right_version(struct kr_cache *cache)
 {
@@ -97,7 +88,7 @@ static int assert_right_version(struct kr_cache *cache)
 				memcpy(&ver, val.data, sizeof(ver));
 				kr_log_verbose("bad version: %d\n", (int)ver);
 			}
-			ret = cache_clear(cache);
+			ret = cache_op(cache, clear);
 		}
 		/* Either purged or empty. */
 		if (ret == 0) {
@@ -107,7 +98,7 @@ static int assert_right_version(struct kr_cache *cache)
 			ret = cache_op(cache, write, &key, &val, 1);
 		}
 	}
-	kr_cache_sync(cache);
+	kr_cache_commit(cache);
 	return ret;
 }
 
@@ -121,11 +112,11 @@ int kr_cache_open(struct kr_cache *cache, const struct kr_cdb_api *api, struct k
 		api = kr_cdb_lmdb();
 	}
 	cache->api = api;
-	int ret = cache->api->open(&cache->db, opts, mm);
+	memset(&cache->stats, 0, sizeof(cache->stats));
+	int ret = cache->api->open(&cache->db, &cache->stats, opts, mm);
 	if (ret != 0) {
 		return ret;
 	}
-	memset(&cache->stats, 0, sizeof(cache->stats));
 	cache->ttl_min = KR_CACHE_DEFAULT_TTL_MIN;
 	cache->ttl_max = KR_CACHE_DEFAULT_TTL_MAX;
 	/* Check cache ABI version */
@@ -157,13 +148,13 @@ void kr_cache_close(struct kr_cache *cache)
 	kr_cache_emergency_file_to_remove = NULL;
 }
 
-int kr_cache_sync(struct kr_cache *cache)
+int kr_cache_commit(struct kr_cache *cache)
 {
 	if (!cache_isvalid(cache)) {
 		return kr_error(EINVAL);
 	}
-	if (cache->api->sync) {
-		return cache_op(cache, sync);
+	if (cache->api->commit) {
+		return cache_op(cache, commit);
 	}
 	return kr_ok();
 }
@@ -175,7 +166,7 @@ int kr_cache_insert_rr(struct kr_cache *cache, const knot_rrset_t *rr, const kno
 		return kr_ok();
 	}
 	ssize_t written = stash_rrset(cache, NULL, rr, rrsig, timestamp, rank, NULL, NULL);
-		/* Zone's NSEC* parames aren't updated, but that's probably OK
+		/* Zone's NSEC* params aren't updated, but that's probably OK
 		 * for kr_cache_insert_rr() */
 	if (written >= 0) {
 		return kr_ok();
@@ -189,7 +180,7 @@ int kr_cache_clear(struct kr_cache *cache)
 	if (!cache_isvalid(cache)) {
 		return kr_error(EINVAL);
 	}
-	int ret = cache_clear(cache);
+	int ret = cache_op(cache, clear);
 	if (ret == 0) {
 		kr_cache_make_checkpoint(cache);
 		ret = assert_right_version(cache);
@@ -198,7 +189,7 @@ int kr_cache_clear(struct kr_cache *cache)
 }
 
 /* When going stricter, BEWARE of breaking entry_h_consistent_NSEC() */
-struct entry_h * entry_h_consistent(knot_db_val_t data, uint16_t type)
+struct entry_h * entry_h_consistent_E(knot_db_val_t data, uint16_t type)
 {
 	(void) type; /* unused, for now */
 	if (!data.data) return NULL;
@@ -336,7 +327,7 @@ int cache_peek(kr_layer_t *ctx, knot_pkt_t *pkt)
 	}
 
 	int ret = peek_nosync(ctx, pkt);
-	kr_cache_sync(&req->ctx->cache);
+	kr_cache_commit(&req->ctx->cache);
 	return ret;
 }
 
@@ -345,7 +336,7 @@ int cache_peek(kr_layer_t *ctx, knot_pkt_t *pkt)
 /** It's simply inside of cycle taken out to decrease indentation.  \return error code. */
 static int stash_rrarray_entry(ranked_rr_array_t *arr, int arr_i,
 			const struct kr_query *qry, struct kr_cache *cache,
-			int *unauth_cnt, trie_t *nsec_pmap, bool *has_optout);
+			int *unauth_cnt, trie_t *nsec_pmap, bool *needs_pkt);
 /** Stash a single nsec_p.  \return 0 (errors are ignored). */
 static int stash_nsec_p(const knot_dname_t *dname, const char *nsec_p_v,
 			struct kr_request *req);
@@ -368,28 +359,34 @@ int cache_stash(kr_layer_t *ctx, knot_pkt_t *pkt)
 	if (knot_wire_get_tc(pkt->wire)) {
 		return ctx->state;
 	}
+	int unauth_cnt = 0;
+	bool needs_pkt = false;
+	if (qry->flags.STUB) {
+		needs_pkt = true;
+		goto stash_packet;
+	}
+
 	/* Stash individual records. */
 	ranked_rr_array_t *selected[] = kr_request_selected(req);
-	int unauth_cnt = 0;
 	trie_t *nsec_pmap = trie_create(&req->pool);
 	if (!nsec_pmap) {
 		assert(!ENOMEM);
 		goto finally;
 	}
-	bool has_optout = false;
-		/* ^^ DNSSEC_OPTOUT is not fired in cases like `com. A`,
-		 * but currently we don't stash separate NSEC3 proving that. */
 	for (int psec = KNOT_ANSWER; psec <= KNOT_ADDITIONAL; ++psec) {
 		ranked_rr_array_t *arr = selected[psec];
 		/* uncached entries are located at the end */
 		for (ssize_t i = arr->len - 1; i >= 0; --i) {
 			ranked_rr_array_entry_t *entry = arr->at[i];
-			if (entry->qry_uid != qry->uid) {
+			if (entry->qry_uid != qry->uid || entry->dont_cache) {
 				continue;
-				/* TODO: probably safe to break but maybe not worth it */
+				/* TODO: probably safe to break on uid mismatch but maybe not worth it */
 			}
-			int ret = stash_rrarray_entry(arr, i, qry, cache, &unauth_cnt,
-							nsec_pmap, &has_optout);
+			int ret = stash_rrarray_entry(
+					arr, i, qry, cache, &unauth_cnt, nsec_pmap,
+					/* ADDITIONAL RRs are considered non-essential
+					 * in our (resolver) answers */
+					(psec == KNOT_ADDITIONAL ? NULL : &needs_pkt));
 			if (ret) {
 				VERBOSE_MSG(qry, "=> stashing RRs errored out\n");
 				goto finally;
@@ -407,14 +404,16 @@ int cache_stash(kr_layer_t *ctx, knot_pkt_t *pkt)
 	trie_it_free(it);
 	/* LATER(optim.): typically we also have corresponding NS record in the list,
 	 * so we might save a cache operation. */
-
-	stash_pkt(pkt, qry, req, has_optout);
+stash_packet:
+	if (qry->flags.PKT_IS_SANE && check_dname_for_lf(knot_pkt_qname(pkt), qry)) {
+		stash_pkt(pkt, qry, req, needs_pkt);
+	}
 
 finally:
 	if (unauth_cnt) {
 		VERBOSE_MSG(qry, "=> stashed also %d nonauth RRsets\n", unauth_cnt);
 	};
-	kr_cache_sync(cache);
+	kr_cache_commit(cache);
 	return ctx->state; /* we ignore cache-stashing errors */
 }
 
@@ -434,26 +433,77 @@ static int stash_rrset_precond(const knot_rrset_t *rr, const struct kr_query *qr
 	return 1/*proceed*/;
 }
 
+/** Return true on some cases of NSEC* RRsets covering minimal ranges.
+ * Also include some abnormal RR cases; qry is just for logging. */
+static bool rrset_has_min_range_or_weird(const knot_rrset_t *rr, const struct kr_query *qry)
+{
+	if (!rr->rrs.count) {
+		assert(!EINVAL);
+		return true; /*< weird */
+	}
+	bool ret; /**< NOT used for the weird cases */
+	if (rr->type == KNOT_RRTYPE_NSEC) {
+		/* NSEC: name -> \000.name
+		 * Note: we have to lower-case it; the reasons are better explained
+		 * on other knot_nsec_next() call sites. */
+		knot_dname_t next[KNOT_DNAME_MAXLEN];
+		if (knot_dname_to_wire(next, knot_nsec_next(rr->rrs.rdata), sizeof(next)) < 0)
+			return true; /*< weird */
+		knot_dname_to_lower(next);
+		ret = next[0] == '\1' && next[1] == '\0'
+			&& knot_dname_is_equal(next + 2, rr->owner);
+	} else if (rr->type == KNOT_RRTYPE_NSEC3) {
+		if (knot_nsec3_next_len(rr->rrs.rdata) != NSEC3_HASH_LEN
+		    || *rr->owner != NSEC3_HASH_TXT_LEN) {
+			return true; /*< weird */
+		}
+		/* Let's work on the binary hashes.  Find if they "differ by one",
+		 * by constructing the owner hash incremented by one and comparing. */
+		uint8_t owner_hash[NSEC3_HASH_LEN];
+		if (base32hex_decode(rr->owner + 1, NSEC3_HASH_TXT_LEN,
+					owner_hash, NSEC3_HASH_LEN) != NSEC3_HASH_LEN) {
+			return true; /*< weird */
+		}
+		for (int i = NSEC3_HASH_LEN - 1; i >= 0; --i) {
+			if (++owner_hash[i] != 0) break;
+		}
+		const uint8_t *next_hash = knot_nsec3_next(rr->rrs.rdata);
+		ret = memcmp(owner_hash, next_hash, NSEC3_HASH_LEN) == 0;
+	} else {
+		return false;
+	}
+	if (ret) VERBOSE_MSG(qry, "=> minimized NSEC* range detected\n");
+	return ret;
+}
+
 static ssize_t stash_rrset(struct kr_cache *cache, const struct kr_query *qry,
 		const knot_rrset_t *rr, const knot_rrset_t *rr_sigs, uint32_t timestamp,
-		uint8_t rank, trie_t *nsec_pmap, bool *has_optout)
+		uint8_t rank, trie_t *nsec_pmap, bool *needs_pkt)
 {
+	if (kr_rank_test(rank, KR_RANK_BOGUS)) {
+		WITH_VERBOSE(qry) {
+			auto_free char *type_str = kr_rrtype_text(rr->type);
+			VERBOSE_MSG(qry, "=> skipping bogus RR set %s\n", type_str);
+		}
+		return kr_ok();
+	}
 	assert(stash_rrset_precond(rr, qry) > 0);
 	if (!cache) {
 		assert(!EINVAL);
 		return kr_error(EINVAL);
 	}
 
+	int ret = kr_ok();
+	if (rrset_has_min_range_or_weird(rr, qry))
+		goto return_needs_pkt;
 	const int wild_labels = rr_sigs == NULL ? 0 :
 	       knot_dname_labels(rr->owner, NULL) - knot_rrsig_labels(rr_sigs->rrs.rdata);
-	if (wild_labels < 0) {
-		return kr_ok();
-	}
+	if (wild_labels < 0)
+		goto return_needs_pkt;
 	const knot_dname_t *encloser = rr->owner; /**< the closest encloser name */
 	for (int i = 0; i < wild_labels; ++i) {
 		encloser = knot_wire_next_label(encloser, NULL);
 	}
-	int ret = 0;
 
 	/* Construct the key under which RRs will be stored,
 	 * and add corresponding nsec_pmap item (if necessary). */
@@ -461,21 +511,18 @@ static ssize_t stash_rrset(struct kr_cache *cache, const struct kr_query *qry,
 	knot_db_val_t key;
 	switch (rr->type) {
 	case KNOT_RRTYPE_NSEC3:
-		/* Skip "suspicious" or opt-out NSEC3 sets. */
-		if (rr->rrs.count != 1) return kr_ok();
-		if (KNOT_NSEC3_FLAG_OPT_OUT & knot_nsec3_flags(rr->rrs.rdata)) {
-			if (has_optout) *has_optout = true;
-			return kr_ok();
-		}
+		/* Skip opt-out NSEC3 sets. */
+		if (KNOT_NSEC3_FLAG_OPT_OUT & knot_nsec3_flags(rr->rrs.rdata))
+			goto return_needs_pkt;
 		/* fall through */
 	case KNOT_RRTYPE_NSEC:
-		if (!kr_rank_test(rank, KR_RANK_SECURE)) {
-			/* Skip any NSEC*s that aren't validated. */
-			return kr_ok();
-		}
+		/* Skip any NSEC*s that aren't validated or are suspicious. */
+		if (!kr_rank_test(rank, KR_RANK_SECURE) || rr->rrs.count != 1)
+			goto return_needs_pkt;
 		if (!rr_sigs || !rr_sigs->rrs.count || !rr_sigs->rrs.rdata) {
 			assert(!EINVAL);
-			return kr_error(EINVAL);
+			ret = kr_error(EINVAL);
+			goto return_needs_pkt;
 		}
 		const knot_dname_t *signer = knot_rrsig_signer_name(rr_sigs->rrs.rdata);
 		const int signer_size = knot_dname_size(signer);
@@ -491,9 +538,15 @@ static ssize_t stash_rrset(struct kr_cache *cache, const struct kr_query *qry,
 
 		assert(rr->type == KNOT_RRTYPE_NSEC3);
 		const knot_rdata_t * const rdata = rr->rrs.rdata;
-		if (rdata->len <= 4) return kr_error(EILSEQ); /*< data from outside; less trust */
+		if (rdata->len <= 4) {
+			ret = kr_error(EILSEQ); /*< data from outside; less trust */
+			goto return_needs_pkt;
+		}
 		const int np_dlen = nsec_p_rdlen(rdata->data);
-		if (np_dlen > rdata->len) return kr_error(EILSEQ);
+		if (np_dlen > rdata->len) {
+			ret = kr_error(EILSEQ);
+			goto return_needs_pkt;
+		}
 		key = key_NSEC3(k, encloser, nsec_p_mkHash(rdata->data));
 		if (npp && !*npp) {
 			*npp = mm_alloc(&qry->request->pool, np_dlen);
@@ -508,12 +561,12 @@ static ssize_t stash_rrset(struct kr_cache *cache, const struct kr_query *qry,
 		ret = kr_dname_lf(k->buf, encloser, wild_labels);
 		if (ret) {
 			assert(!ret);
-			return kr_error(ret);
+			goto return_needs_pkt;
 		}
 		key = key_exact_type(k, rr->type);
 	}
 
-	/* Compute materialized sizes of the new data. */
+	/* Compute in-cache size for the new data. */
 	const knot_rdataset_t *rds_sigs = rr_sigs ? &rr_sigs->rrs : NULL;
 	const int rr_ssize = rdataset_dematerialize_size(&rr->rrs);
 	assert(rr_ssize == to_even(rr_ssize));
@@ -546,11 +599,11 @@ static ssize_t stash_rrset(struct kr_cache *cache, const struct kr_query *qry,
 		eh->rank = 0;
 		assert(false);
 	}
-	assert(entry_h_consistent(val_new_entry, rr->type));
+	assert(entry_h_consistent_E(val_new_entry, rr->type));
 
 	#if 0 /* Occasionally useful when debugging some kinds of changes. */
 	{
-	kr_cache_sync(cache);
+	kr_cache_commit(cache);
 	knot_db_val_t val = { NULL, 0 };
 	ret = cache_op(cache, read, &key, &val, 1);
 	if (ret != kr_error(ENOENT)) { // ENOENT might happen in some edge case, I guess
@@ -561,9 +614,6 @@ static ssize_t stash_rrset(struct kr_cache *cache, const struct kr_query *qry,
 	}
 	#endif
 
-	/* Update metrics */
-	cache->stats.insert += 1;
-
 	/* Verbose-log some not-too-common cases. */
 	WITH_VERBOSE(qry) { if (kr_rank_test(rank, KR_RANK_AUTH)
 				|| rr->type == KNOT_RRTYPE_NS) {
@@ -577,11 +627,14 @@ static ssize_t stash_rrset(struct kr_cache *cache, const struct kr_query *qry,
 	} }
 
 	return (ssize_t) val_new_entry.len;
+return_needs_pkt:
+	if (needs_pkt) *needs_pkt = true;
+	return ret;
 }
 
 static int stash_rrarray_entry(ranked_rr_array_t *arr, int arr_i,
 			const struct kr_query *qry, struct kr_cache *cache,
-			int *unauth_cnt, trie_t *nsec_pmap, bool *has_optout)
+			int *unauth_cnt, trie_t *nsec_pmap, bool *needs_pkt)
 {
 	ranked_rr_array_entry_t *entry = arr->at[arr_i];
 	if (entry->cached) {
@@ -603,6 +656,7 @@ static int stash_rrarray_entry(ranked_rr_array_t *arr, int arr_i,
 		/* TODO: ATM we assume that some properties are the same
 		 * for all RRSIGs in the set (esp. label count). */
 		ranked_rr_array_entry_t *e = arr->at[j];
+		assert(!e->in_progress);
 		bool ok = e->qry_uid == qry->uid && !e->cached
 			&& e->rr->type == KNOT_RRTYPE_RRSIG
 			&& knot_rrsig_type_covered(e->rr->rrs.rdata) == rr->type
@@ -614,7 +668,7 @@ static int stash_rrarray_entry(ranked_rr_array_t *arr, int arr_i,
 	}
 
 	ssize_t written = stash_rrset(cache, qry, rr, rr_sigs, qry->timestamp.tv_sec,
-					entry->rank, nsec_pmap, has_optout);
+					entry->rank, nsec_pmap, needs_pkt);
 	if (written < 0) {
 		kr_log_error("[%05u.%02u][cach] stash failed, ret = %d\n", qry->request->uid,
 			     qry->uid, ret);
@@ -749,7 +803,7 @@ static int peek_exact_real(struct kr_cache *cache, const knot_dname_t *name, uin
 	if (!ret) ret = entry_h_seek(&val, type);
 	if (ret) return kr_error(ret);
 
-	const struct entry_h *eh = entry_h_consistent(val, type);
+	const struct entry_h *eh = entry_h_consistent_E(val, type);
 	if (!eh || eh->is_packet) {
 		// TODO: no packets, but better get rid of whole kr_cache_peek_exact().
 		return kr_error(ENOENT);
@@ -877,9 +931,9 @@ int kr_cache_remove_subtree(struct kr_cache *cache, const knot_dname_t *name,
 		}
 		memcpy(keys[i].data, keyval[i][0].data, keys[i].len);
 	}
-	ret = cache->api->remove(cache->db, keys, count);
+	ret = cache_op(cache, remove, keys, count);
 cleanup:
-	kr_cache_sync(cache); /* Sync even after just kr_cache_match(). */
+	kr_cache_commit(cache); /* Sync even after just kr_cache_match(). */
 	/* Free keys */
 	while (--i >= 0) {
 		free(keys[i].data);
