@@ -1,17 +1,5 @@
 /*  Copyright (C) 2016-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ *  SPDX-License-Identifier: GPL-3.0-or-later
 */
 
 #include <assert.h>
@@ -25,7 +13,9 @@
 #include <lmdb.h>
 
 #include "contrib/cleanup.h"
+#include "contrib/ucw/lib.h"
 #include "lib/cache/cdb_lmdb.h"
+#include "lib/cache/cdb_api.h"
 #include "lib/cache/api.h"
 #include "lib/utils.h"
 
@@ -109,35 +99,44 @@ static int set_mapsize(MDB_env *env, size_t map_size)
 }
 
 #define FLAG_RENEW (2*MDB_RDONLY)
-/** mdb_txn_begin or _renew + handle MDB_MAP_RESIZED.
+/** mdb_txn_begin or _renew + handle retries in some situations
  *
- * The retrying logic for MDB_MAP_RESIZED is so ugly that it has its own function.
+ * The retrying logic is so ugly that it has its own function.
  * \note this assumes no transactions are active
  * \return MDB_ errcode, not usual kr_error(...)
  */
 static int txn_get_noresize(struct lmdb_env *env, unsigned int flag, MDB_txn **txn)
 {
 	assert(!env->txn.rw && (!env->txn.ro || !env->txn.ro_active));
+	int attempts = 0;
 	int ret;
+retry:
+	/* Do a few attempts in case we encounter multiple issues at once. */
+	if (++attempts > 2) {
+		return kr_error(1);
+	}
+
 	if (flag == FLAG_RENEW) {
 		ret = mdb_txn_renew(*txn);
 	} else {
 		ret = mdb_txn_begin(env->env, NULL, flag, txn);
 	}
-	if (ret != MDB_MAP_RESIZED) {
-		return ret;
-	}
-	//:unlikely
-	/* Another process increased the size; let's try to recover. */
-	kr_log_info("[cache] detected size increased by another process\n");
-	ret = mdb_env_set_mapsize(env->env, 0);
-	if (ret != MDB_SUCCESS) {
-		return ret;
-	}
-	if (flag == FLAG_RENEW) {
-		ret = mdb_txn_renew(*txn);
-	} else {
-		ret = mdb_txn_begin(env->env, NULL, flag, txn);
+
+	if (unlikely(ret == MDB_MAP_RESIZED)) {
+		kr_log_info("[cache] detected size increased by another process\n");
+		ret = mdb_env_set_mapsize(env->env, 0);
+		if (ret == MDB_SUCCESS) {
+			goto retry;
+		}
+	} else if (unlikely(ret == MDB_READERS_FULL)) {
+		int cleared;
+		ret = mdb_reader_check(env->env, &cleared);
+		if (ret == MDB_SUCCESS)
+			kr_log_info("[cache] cleared %d stale reader locks\n", cleared);
+		else
+			kr_log_error("[cache] failed to clear stale reader locks: "
+					"LMDB error %d %s\n", ret, mdb_strerror(ret));
+		goto retry;
 	}
 	return ret;
 }
@@ -148,7 +147,7 @@ static int txn_get(struct lmdb_env *env, MDB_txn **txn, bool rdonly)
 	assert(env && txn);
 	if (env->txn.rw) {
 		/* Reuse the *open* RW txn even if only reading is requested.
-		 * We leave the management of this to the cdb_sync command.
+		 * We leave the management of this to the cdb_commit command.
 		 * The user may e.g. want to do some reads between the writes. */
 		*txn = env->txn.rw;
 		return kr_ok();
@@ -185,11 +184,12 @@ static int txn_get(struct lmdb_env *env, MDB_txn **txn, bool rdonly)
 	return kr_ok();
 }
 
-static int cdb_sync(knot_db_t *db)
+static int cdb_commit(knot_db_t *db, struct kr_cdb_stats *stats)
 {
 	struct lmdb_env *env = db;
 	int ret = kr_ok();
 	if (env->txn.rw) {
+		stats->commit++;
 		ret = lmdb_error(mdb_txn_commit(env->txn.rw));
 		env->txn.rw = NULL; /* the transaction got freed even in case of errors */
 	} else if (env->txn.ro && env->txn.ro_active) {
@@ -201,7 +201,7 @@ static int cdb_sync(knot_db_t *db)
 }
 
 /** Obtain a read-only cursor (and a read-only transaction). */
-static int txn_curs_get(struct lmdb_env *env, MDB_cursor **curs)
+static int txn_curs_get(struct lmdb_env *env, MDB_cursor **curs, struct kr_cdb_stats *stats)
 {
 	assert(env && curs);
 	if (env->txn.ro_curs_active) {
@@ -209,7 +209,7 @@ static int txn_curs_get(struct lmdb_env *env, MDB_cursor **curs)
 	}
 	/* Only in a read-only txn; TODO: it's a bit messy/coupled */
 	if (env->txn.rw) {
-		int ret = cdb_sync(env);
+		int ret = cdb_commit(env, stats);
 		if (ret) return ret;
 	}
 	MDB_txn *txn = NULL;
@@ -244,22 +244,23 @@ static void free_txn_ro(struct lmdb_env *env)
 }
 
 /*! \brief Close the database. */
-static void cdb_close_env(struct lmdb_env *env)
+static void cdb_close_env(struct lmdb_env *env, struct kr_cdb_stats *stats)
 {
 	assert(env && env->env);
 
 	/* Get rid of any transactions. */
-	cdb_sync(env);
+	cdb_commit(env, stats);
 	free_txn_ro(env);
 
 	mdb_env_sync(env->env, 1);
+	stats->close++;
 	mdb_dbi_close(env->env, env->dbi);
 	mdb_env_close(env->env);
 	memset(env, 0, sizeof(*env));
 }
 
 /*! \brief Open database environment. */
-static int cdb_open_env(struct lmdb_env *env, unsigned flags, const char *path, size_t mapsize)
+static int cdb_open_env(struct lmdb_env *env, unsigned flags, const char *path, size_t mapsize, struct kr_cdb_stats *stats)
 {
 	int ret = mkdir(path, LMDB_DIR_MODE);
 	if (ret == -1 && errno != EEXIST) {
@@ -267,6 +268,7 @@ static int cdb_open_env(struct lmdb_env *env, unsigned flags, const char *path,
 	}
 
 	MDB_env *mdb_env = NULL;
+	stats->open++;
 	ret = mdb_env_create(&mdb_env);
 	if (ret != MDB_SUCCESS) {
 		return lmdb_error(ret);
@@ -274,12 +276,14 @@ static int cdb_open_env(struct lmdb_env *env, unsigned flags, const char *path,
 
 	ret = set_mapsize(mdb_env, mapsize);
 	if (ret != 0) {
+		stats->close++;
 		mdb_env_close(mdb_env);
 		return ret;
 	}
 
 	ret = mdb_env_open(mdb_env, path, flags, LMDB_FILE_MODE);
 	if (ret != MDB_SUCCESS) {
+		stats->close++;
 		mdb_env_close(mdb_env);
 		return lmdb_error(ret);
 	}
@@ -290,12 +294,13 @@ static int cdb_open_env(struct lmdb_env *env, unsigned flags, const char *path,
 	return 0;
 }
 
-static int cdb_open(struct lmdb_env *env, const char *path, size_t mapsize)
+static int cdb_open(struct lmdb_env *env, const char *path, size_t mapsize,
+		struct kr_cdb_stats *stats)
 {
 	/* Cache doesn't require durability, we can be
 	 * loose with the requirements as a tradeoff for speed. */
 	const unsigned flags = MDB_WRITEMAP | MDB_MAPASYNC | MDB_NOTLS;
-	int ret = cdb_open_env(env, flags, path, mapsize);
+	int ret = cdb_open_env(env, flags, path, mapsize, stats);
 	if (ret != 0) {
 		return ret;
 	}
@@ -304,6 +309,7 @@ static int cdb_open(struct lmdb_env *env, const char *path, size_t mapsize)
 	MDB_txn *txn = NULL;
 	ret = mdb_txn_begin(env->env, NULL, 0, &txn);
 	if (ret != MDB_SUCCESS) {
+		stats->close++;
 		mdb_env_close(env->env);
 		return lmdb_error(ret);
 	}
@@ -311,12 +317,41 @@ static int cdb_open(struct lmdb_env *env, const char *path, size_t mapsize)
 	ret = mdb_dbi_open(txn, NULL, 0, &env->dbi);
 	if (ret != MDB_SUCCESS) {
 		mdb_txn_abort(txn);
+		stats->close++;
 		mdb_env_close(env->env);
 		return lmdb_error(ret);
 	}
 
+#if !defined(__MACOSX__) && !(defined(__APPLE__) && defined(__MACH__))
+	auto_free char *mdb_datafile = kr_strcatdup(2, path, "/data.mdb");
+	int fd = open(mdb_datafile, O_RDWR);
+	if (fd == -1) {
+		mdb_txn_abort(txn);
+		stats->close++;
+		mdb_env_close(env->env);
+		return kr_error(errno);
+	}
+
+	ret = posix_fallocate(fd, 0, mapsize);
+	if (ret == EINVAL) {
+		/* POSIX says this can happen when the feature isn't supported by the FS.
+		 * We haven't seen this happen on Linux+glibc but it was reported on FreeBSD.*/
+		kr_log_info("[cache] space pre-allocation failed and ignored; "
+				"your (file)system probably doesn't support it.\n");
+	} else if (ret != 0) {
+		mdb_txn_abort(txn);
+		stats->close++;
+		mdb_env_close(env->env);
+		close(fd);
+		return kr_error(ret);
+	}
+	close(fd);
+#endif
+
+	stats->commit++;
 	ret = mdb_txn_commit(txn);
 	if (ret != MDB_SUCCESS) {
+		stats->close++;
 		mdb_env_close(env->env);
 		return lmdb_error(ret);
 	}
@@ -324,9 +359,10 @@ static int cdb_open(struct lmdb_env *env, const char *path, size_t mapsize)
 	return 0;
 }
 
-static int cdb_init(knot_db_t **db, struct kr_cdb_opts *opts, knot_mm_t *pool)
+static int cdb_init(knot_db_t **db, struct kr_cdb_stats *stats,
+		struct kr_cdb_opts *opts, knot_mm_t *pool)
 {
-	if (!db || !opts) {
+	if (!db || !stats || !opts) {
 		return kr_error(EINVAL);
 	}
 
@@ -348,7 +384,7 @@ static int cdb_init(knot_db_t **db, struct kr_cdb_opts *opts, knot_mm_t *pool)
 	}
 
 	/* Open the database. */
-	int ret = cdb_open(env, opts->path, opts->maxsize);
+	int ret = cdb_open(env, opts->path, opts->maxsize, stats);
 	if (ret != 0) {
 		free(env);
 		return ret;
@@ -358,14 +394,14 @@ static int cdb_init(knot_db_t **db, struct kr_cdb_opts *opts, knot_mm_t *pool)
 	return 0;
 }
 
-static void cdb_deinit(knot_db_t *db)
+static void cdb_deinit(knot_db_t *db, struct kr_cdb_stats *stats)
 {
 	struct lmdb_env *env = db;
-	cdb_close_env(env);
+	cdb_close_env(env, stats);
 	free(env);
 }
 
-static int cdb_count(knot_db_t *db)
+static int cdb_count(knot_db_t *db, struct kr_cdb_stats *stats)
 {
 	struct lmdb_env *env = db;
 	MDB_txn *txn = NULL;
@@ -375,14 +411,16 @@ static int cdb_count(knot_db_t *db)
 	}
 
 	MDB_stat stat;
+	stats->count++;
 	ret = mdb_stat(txn, env->dbi, &stat);
 
 	return (ret == MDB_SUCCESS) ? stat.ms_entries : lmdb_error(ret);
 }
 
-static int cdb_clear(knot_db_t *db)
+static int cdb_clear(knot_db_t *db, struct kr_cdb_stats *stats)
 {
 	struct lmdb_env *env = db;
+	stats->clear++;
 	/* First try mdb_drop() to clear the DB; this may fail with ENOSPC. */
 	/* If we didn't do this, explicit cache.clear() ran on an instance
 	 * would lead to the instance detaching from the cache of others,
@@ -394,7 +432,7 @@ static int cdb_clear(knot_db_t *db)
 		if (ret == kr_ok()) {
 			ret = lmdb_error(mdb_drop(txn, env->dbi, 0));
 			if (ret == kr_ok()) {
-				ret = cdb_sync(db);
+				ret = cdb_commit(db, stats);
 			}
 			if (ret == kr_ok()) {
 				return ret;
@@ -404,7 +442,7 @@ static int cdb_clear(knot_db_t *db)
 	}
 
 	/* We are about to switch to a different file, so end all txns, to be sure. */
-	(void) cdb_sync(db);
+	(void) cdb_commit(db, stats);
 	free_txn_ro(db);
 
 	/* Since there is no guarantee that there will be free
@@ -460,15 +498,15 @@ static int cdb_clear(knot_db_t *db)
 	/* Keep copy as it points to current handle internals. */
 	auto_free char *path_copy = strdup(path);
 	size_t mapsize = env->mapsize;
-	cdb_close_env(env);
-	ret = cdb_open(env, path_copy, mapsize);
+	cdb_close_env(env, stats);
+	ret = cdb_open(env, path_copy, mapsize, stats);
 	/* Environment updated, release lockfile. */
 	unlink(lockfile);
 	return ret;
 }
 
-static int cdb_readv(knot_db_t *db, const knot_db_val_t *key, knot_db_val_t *val,
-		     int maxcount)
+static int cdb_readv(knot_db_t *db, struct kr_cdb_stats *stats,
+		const knot_db_val_t *key, knot_db_val_t *val, int maxcount)
 {
 	struct lmdb_env *env = db;
 	MDB_txn *txn = NULL;
@@ -481,8 +519,11 @@ static int cdb_readv(knot_db_t *db, const knot_db_val_t *key, knot_db_val_t *val
 		/* Convert key structs */
 		MDB_val _key = val_knot2mdb(key[i]);
 		MDB_val _val = val_knot2mdb(val[i]);
+		stats->read++;
 		ret = mdb_get(txn, env->dbi, &_key, &_val);
 		if (ret != MDB_SUCCESS) {
+			if (ret == MDB_NOTFOUND)
+				stats->read_miss++;
 			ret = lmdb_error(ret);
 			if (ret == kr_error(ENOSPC)) {
 				/* we're likely to be forced to cache clear anyway */
@@ -497,20 +538,23 @@ static int cdb_readv(knot_db_t *db, const knot_db_val_t *key, knot_db_val_t *val
 }
 
 static int cdb_write(struct lmdb_env *env, MDB_txn **txn, const knot_db_val_t *key,
-			knot_db_val_t *val, unsigned flags)
+			knot_db_val_t *val, unsigned flags,
+			struct kr_cdb_stats *stats)
 {
 	/* Convert key structs and write */
 	MDB_val _key = val_knot2mdb(*key);
 	MDB_val _val = val_knot2mdb(*val);
+	stats->write++;
 	int ret = mdb_put(*txn, env->dbi, &_key, &_val, flags);
 
 	/* Try to recover from doing too much writing in a single transaction. */
 	if (ret == MDB_TXN_FULL) {
-		ret = cdb_sync(env);
+		ret = cdb_commit(env, stats);
 		if (ret) {
 			ret = txn_get(env, txn, false);
 		}
 		if (ret) {
+			stats->write++;
 			ret = mdb_put(*txn, env->dbi, &_key, &_val, flags);
 		}
 	}
@@ -524,8 +568,8 @@ static int cdb_write(struct lmdb_env *env, MDB_txn **txn, const knot_db_val_t *k
 	return kr_ok();
 }
 
-static int cdb_writev(knot_db_t *db, const knot_db_val_t *key, knot_db_val_t *val,
-			int maxcount)
+static int cdb_writev(knot_db_t *db, struct kr_cdb_stats *stats,
+		const knot_db_val_t *key, knot_db_val_t *val, int maxcount)
 {
 	struct lmdb_env *env = db;
 	MDB_txn *txn = NULL;
@@ -541,13 +585,14 @@ static int cdb_writev(knot_db_t *db, const knot_db_val_t *key, knot_db_val_t *va
 		if (val[i].len > 0 && val[i].data == NULL) {
 			mdb_flags |= MDB_RESERVE;
 		}
-		ret = cdb_write(env, &txn, &key[i], &val[i], mdb_flags);
+		ret = cdb_write(env, &txn, &key[i], &val[i], mdb_flags, stats);
 	}
 
 	return ret;
 }
 
-static int cdb_remove(knot_db_t *db, knot_db_val_t keys[], int maxcount)
+static int cdb_remove(knot_db_t *db, struct kr_cdb_stats *stats,
+		knot_db_val_t keys[], int maxcount)
 {
 	struct lmdb_env *env = db;
 	MDB_txn *txn = NULL;
@@ -557,17 +602,21 @@ static int cdb_remove(knot_db_t *db, knot_db_val_t keys[], int maxcount)
 	for (int i = 0; ret == kr_ok() && i < maxcount; ++i) {
 		MDB_val _key = val_knot2mdb(keys[i]);
 		MDB_val val = { 0, NULL };
+		stats->remove++;
 		ret = lmdb_error(mdb_del(txn, env->dbi, &_key, &val));
 		if (ret == kr_ok())
 			deleted++;
-		else if (ret == KNOT_ENOENT)
+		else if (ret == KNOT_ENOENT) {
+			stats->remove_miss++;
 			ret = kr_ok();  /* skip over non-existing entries */
+		}
 	}
 
 	return ret < 0 ? ret : deleted;
 }
 
-static int cdb_match(knot_db_t *db, knot_db_val_t *key, knot_db_val_t keyval[][2], int maxcount)
+static int cdb_match(knot_db_t *db, struct kr_cdb_stats *stats,
+		knot_db_val_t *key, knot_db_val_t keyval[][2], int maxcount)
 {
 	struct lmdb_env *env = db;
 	MDB_txn *txn = NULL;
@@ -585,6 +634,7 @@ static int cdb_match(knot_db_t *db, knot_db_val_t *key, knot_db_val_t keyval[][2
 
 	MDB_val cur_key = val_knot2mdb(*key);
 	MDB_val cur_val = { 0, NULL };
+	stats->match++;
 	ret = mdb_cursor_get(cur, &cur_key, &cur_val, MDB_SET_RANGE);
 	if (ret != MDB_SUCCESS) {
 		mdb_cursor_close(cur);
@@ -605,88 +655,49 @@ static int cdb_match(knot_db_t *db, knot_db_val_t *key, knot_db_val_t keyval[][2
 		} else {
 			break;
 		}
+		stats->match++;
 		ret = mdb_cursor_get(cur, &cur_key, &cur_val, MDB_NEXT);
 	}
+	if (results == 0)
+		stats->match_miss++;
+
 
 	mdb_cursor_close(cur);
 	return results;
 }
 
 
-static int cdb_prune(knot_db_t *db, int limit)
-{
-	return -1;
-#if 0
-	/* Sync in-flight transactions */
-	cdb_sync(db);
-
-	/* Prune old records */
-	struct lmdb_env *env = db;
-	MDB_txn *txn = NULL;
-	int ret = txn_get(env, &txn, false);
-	if (ret != 0) {
-		return ret;
-	}
-
-	MDB_cursor *cur = NULL;
-	ret = mdb_cursor_open(txn, env->dbi, &cur);
-	if (ret != 0) {
-		return lmdb_error(ret);
-	}
-
-	MDB_val cur_key, cur_val;
-	ret = mdb_cursor_get(cur, &cur_key, &cur_val, MDB_FIRST);
-	if (ret != 0) {
-		mdb_cursor_close(cur);
-		return lmdb_error(ret);
-	}
-
-	int results = 0;
-	struct timeval now;
-	gettimeofday(&now, NULL);
-	while (ret == 0 && results < limit) {
-		/* Ignore special namespaces. */
-		if (cur_key.mv_size < 2 || ((const char *)cur_key.mv_data)[0] == 'V') {
-			ret = mdb_cursor_get(cur, &cur_key, &cur_val, MDB_NEXT);
-			continue;
-		}
-		/* Check entry age. */
-		struct kr_cache_entry *entry = cur_val.mv_data;
-		if (entry->timestamp > now.tv_sec ||
-			(now.tv_sec - entry->timestamp) < entry->ttl) {
-			ret = mdb_cursor_get(cur, &cur_key, &cur_val, MDB_NEXT);
-			continue;
-		}
-		/* Remove entry */
-		mdb_cursor_del(cur, 0);
-		++results;
-		ret = mdb_cursor_get(cur, &cur_key, &cur_val, MDB_NEXT);
-	}
-	mdb_cursor_close(cur);
-	return ret < 0 ? ret : results;
-#endif
-}
-
-static int cdb_read_leq(knot_db_t *env, knot_db_val_t *key, knot_db_val_t *val)
+static int cdb_read_leq(knot_db_t *env, struct kr_cdb_stats *stats,
+		knot_db_val_t *key, knot_db_val_t *val)
 {
 	assert(env && key && key->data && val);
 	MDB_cursor *curs = NULL;
-	int ret = txn_curs_get(env, &curs);
+	int ret = txn_curs_get(env, &curs, stats);
 	if (ret) return ret;
 
 	MDB_val key2_m = val_knot2mdb(*key);
 	MDB_val val2_m = { 0, NULL };
+	stats->read_leq++;
 	ret = mdb_cursor_get(curs, &key2_m, &val2_m, MDB_SET_RANGE);
-	if (ret) return lmdb_error(ret);
+	if (ret) {
+		stats->read_leq_miss++;
+		return lmdb_error(ret);
+	}
 	/* test for equality //:unlikely */
 	if (key2_m.mv_size == key->len
 	    && memcmp(key2_m.mv_data, key->data, key->len) == 0) {
 		ret = 0; /* equality */
 		goto success;
 	}
+	stats->read_leq_miss++;
+
 	/* we must be greater than key; do one step to smaller */
+	stats->read_leq++;
 	ret = mdb_cursor_get(curs, &key2_m, &val2_m, MDB_PREV);
-	if (ret) return lmdb_error(ret);
+	if (ret) {
+		stats->read_leq_miss++;
+		return lmdb_error(ret);
+	}
 	ret = 1;
 success:
 	/* finalize the output */
@@ -700,9 +711,9 @@ const struct kr_cdb_api *kr_cdb_lmdb(void)
 {
 	static const struct kr_cdb_api api = {
 		"lmdb",
-		cdb_init, cdb_deinit, cdb_count, cdb_clear, cdb_sync,
+		cdb_init, cdb_deinit, cdb_count, cdb_clear, cdb_commit,
 		cdb_readv, cdb_writev, cdb_remove,
-		cdb_match, cdb_prune,
+		cdb_match,
 		cdb_read_leq
 	};
 
