@@ -1,17 +1,5 @@
 /*  Copyright (C) 2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 /** @file
@@ -132,7 +120,7 @@ static int entry_h_len(const knot_db_val_t val)
 	if (!eh->is_packet) { /* Positive RRset + its RRsig set (may be empty). */
 		int sets = 2;
 		while (sets-- > 0) {
-			d += rdataset_dematerialized_size(d);
+			d += KR_CACHE_RR_COUNT_SIZE + rdataset_dematerialized_size(d, NULL);
 			if (d > data_bound) {
 				assert(!EILSEQ);
 				return kr_error(EILSEQ);
@@ -234,7 +222,7 @@ int entry_h_splice(
 		}
 		/* val is on the entry, in either case (or error) */
 		if (!ret) {
-			eh_orig = entry_h_consistent(val, type);
+			eh_orig = entry_h_consistent_E(val, type);
 		}
 	} else {
 		/* We want to fully overwrite the entry, so don't even read it. */
@@ -268,7 +256,6 @@ int entry_h_splice(
 	/* Now we're in trouble.  In some cases, parts of data to be written
 	 * is an lmdb entry that may be invalidated by our write request.
 	 * (lmdb does even in-place updates!) Therefore we copy all into a buffer.
-	 * (We don't bother deallocating from the mempool.)
 	 * LATER(optim.): do this only when neccessary, or perhaps another approach.
 	 * This is also complicated by the fact that the val_new_entry part
 	 * is to be written *afterwards* by the caller.
@@ -281,13 +268,13 @@ int entry_h_splice(
 		.len = entry_list_serial_size(el),
 		.data = NULL,
 	};
-	void *buf = mm_alloc(&qry->request->pool, val.len);
-	entry_list_memcpy(buf, el);
+	uint8_t buf[val.len];
+	entry_list_memcpy((struct entry_apex *)buf, el);
 	ret = cache_write_or_clear(cache, &key, &val, qry);
 	if (ret) return kr_error(ret);
 	memcpy(val.data, buf, val.len); /* we also copy the "empty" space, but well... */
 	val_new_entry->data = (uint8_t *)val.data
-			    + ((uint8_t *)el[i_type].data - (uint8_t *)buf);
+			    + ((uint8_t *)el[i_type].data - buf);
 	return kr_ok();
 }
 
