@@ -1,17 +1,5 @@
 /*  Copyright (C) 2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 /** @file
@@ -63,7 +51,8 @@ struct entry_apex;
 /** Check basic consistency of entry_h for 'E' entries, not looking into ->data.
  * (for is_packet the length of data is checked)
  */
-struct entry_h * entry_h_consistent(knot_db_val_t data, uint16_t type);
+KR_EXPORT
+struct entry_h * entry_h_consistent_E(knot_db_val_t data, uint16_t type);
 
 struct entry_apex * entry_apex_consistent(knot_db_val_t val);
 
@@ -71,12 +60,22 @@ struct entry_apex * entry_apex_consistent(knot_db_val_t val);
 static inline struct entry_h * entry_h_consistent_NSEC(knot_db_val_t data)
 {
 	/* ATM it's enough to just extend the checks for exact entries. */
-	const struct entry_h *eh = entry_h_consistent(data, KNOT_RRTYPE_NSEC);
+	const struct entry_h *eh = entry_h_consistent_E(data, KNOT_RRTYPE_NSEC);
 	bool ok = eh != NULL;
 	ok = ok && !eh->is_packet && !eh->has_optout;
 	return ok ? /*const-cast*/(struct entry_h *)eh : NULL;
 }
 
+static inline struct entry_h * entry_h_consistent(knot_db_val_t data, uint16_t type)
+{
+	switch (type) {
+	case KNOT_RRTYPE_NSEC:
+	case KNOT_RRTYPE_NSEC3:
+		return entry_h_consistent_NSEC(data);
+	default:
+		return entry_h_consistent_E(data, type);
+	}
+}
 
 /* nsec_p* - NSEC* chain parameters */
 
@@ -226,7 +225,7 @@ int entry_h_splice(
 	const struct kr_query *qry, struct kr_cache *cache, uint32_t timestamp);
 
 /** Parse an entry_apex into individual items.  @return error code. */
-int entry_list_parse(const knot_db_val_t val, entry_list_t list);
+KR_EXPORT int entry_list_parse(const knot_db_val_t val, entry_list_t list);
 
 static inline size_t to_even(size_t n)
 {
@@ -253,9 +252,11 @@ void entry_list_memcpy(struct entry_apex *ea, entry_list_t list);
 /* Packet caching; implementation in ./entry_pkt.c */
 
 /** Stash the packet into cache (if suitable, etc.)
- * \param has_optout whether the packet contains an opt-out NSEC3 */
+ * \param needs_pkt we need the packet due to not stashing some RRs;
+ * 		see stash_rrset() for details
+ * It assumes check_dname_for_lf(). */
 void stash_pkt(const knot_pkt_t *pkt, const struct kr_query *qry,
-		const struct kr_request *req, bool has_optout);
+		const struct kr_request *req, bool needs_pkt);
 
 /** Try answering from packet cache, given an entry_h.
  *
@@ -294,12 +295,19 @@ static inline int rdataset_dematerialize_size(const knot_rdataset_t *rds)
 	return KR_CACHE_RR_COUNT_SIZE + (rds == NULL ? 0 : knot_rdataset_size(rds));
 }
 
-static inline int rdataset_dematerialized_size(const uint8_t *data)
+/** Analyze the length of a dematerialized rdataset.
+ * Note that in the data it's KR_CACHE_RR_COUNT_SIZE and then this returned size. */
+static inline int rdataset_dematerialized_size(const uint8_t *data, uint16_t *rdataset_count)
 {
-	knot_rdataset_t rds;
-	memcpy(&rds.count, data, sizeof(rds.count));
-	rds.rdata = (knot_rdata_t *)(data + sizeof(rds.count));
-	return sizeof(rds.count) + knot_rdataset_size(&rds);
+	uint16_t count;
+	assert(sizeof(count) == KR_CACHE_RR_COUNT_SIZE);
+	memcpy(&count, data, sizeof(count));
+	const uint8_t *rdata = data + sizeof(count);
+	if (rdataset_count)
+		*rdataset_count = count;
+	for (int i = 0; i < count; ++i)
+		rdata += knot_rdata_size(((knot_rdata_t *)rdata)->len);
+	return rdata - (data + sizeof(count));
 }
 
 /** Serialize an rdataset. */
@@ -394,7 +402,7 @@ int nsec3_src_synth(struct key *k, struct answer *ans, const knot_dname_t *clenc
 #define VERBOSE_MSG(qry, ...) QRVERBOSE((qry), "cach",  ## __VA_ARGS__)
 
 /** Shorthand for operations on cache backend */
-#define cache_op(cache, op, ...) (cache)->api->op((cache)->db, ## __VA_ARGS__)
+#define cache_op(cache, op, ...) (cache)->api->op((cache)->db, &(cache)->stats, ## __VA_ARGS__)
 
 
 static inline uint16_t get_uint16(const void *address)
