@@ -1,17 +1,5 @@
 /*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 /** @file iterate.c
@@ -36,6 +24,7 @@
 #include <libknot/rrtype/rdname.h>
 #include <libknot/rrtype/rrsig.h>
 
+#include "kresconfig.h"
 #include "lib/layer/iterate.h"
 #include "lib/resolve.h"
 #include "lib/rplan.h"
@@ -81,10 +70,14 @@ static bool is_paired_to_query(const knot_pkt_t *answer, struct kr_query *query)
 	uint16_t qtype = query->stype;
 	const knot_dname_t *qname = minimized_qname(query, &qtype);
 
+	/* ID should already match, thanks to session_tasklist_del_msgid()
+	 * in worker_submit(), but it won't hurt to check again. */
 	return query->id      == knot_wire_get_id(answer->wire) &&
-	       knot_wire_get_qdcount(answer->wire) > 0 &&
+	       knot_wire_get_qdcount(answer->wire) == 1 &&
 	       query->sclass  == knot_pkt_qclass(answer) &&
 	       qtype          == knot_pkt_qtype(answer) &&
+	       /* qry->secret had been xor-applied to answer already,
+		* so this also checks for correctness of case randomization */
 	       knot_dname_is_equal(qname, knot_pkt_qname(answer));
 }
 
@@ -381,7 +374,7 @@ static int pick_authority(knot_pkt_t *pkt, struct kr_request *req, bool to_wire)
 						qry->flags.FORWARD || referral);
 		int ret = kr_ranked_rrarray_add(&req->auth_selected, rr,
 						rank, to_wire, qry->uid, &req->pool);
-		if (ret != kr_ok()) {
+		if (ret < 0) {
 			return ret;
 		}
 	}
@@ -496,14 +489,17 @@ static int unroll_cname(knot_pkt_t *pkt, struct kr_request *req, bool referral,
 	const knot_pktsection_t *an = knot_pkt_section(pkt, KNOT_ANSWER);
 	const knot_dname_t *cname = NULL;
 	const knot_dname_t *pending_cname = query->sname;
-	unsigned cname_chain_len = 0;
 	bool is_final = (query->parent == NULL);
-	uint32_t iter_count = 0;
 	bool strict_mode = (query->flags.STRICT);
+
+	query->cname_depth = query->cname_parent ? query->cname_parent->cname_depth : 1;
+
 	do {
 		/* CNAME was found at previous iteration, but records may not follow the correct order.
 		 * Try to find records for pending_cname owner from section start. */
 		cname = pending_cname;
+		size_t cname_answ_selected_i = -1;
+		bool cname_is_occluded = false; /* whether `cname` is in a DNAME's bailiwick */
 		pending_cname = NULL;
 		const int cname_labels = knot_dname_labels(cname, NULL);
 		for (unsigned i = 0; i < an->count; ++i) {
@@ -512,13 +508,33 @@ static int unroll_cname(knot_pkt_t *pkt, struct kr_request *req, bool referral,
 			/* Skip the RR if its owner+type doesn't interest us. */
 			const uint16_t type = kr_rrset_type_maysig(rr);
 			const bool type_OK = rr->type == query->stype || type == query->stype
-				|| type == KNOT_RRTYPE_CNAME || type == KNOT_RRTYPE_DNAME;
-				/* TODO: actually handle DNAMEs */
-			if (rr->rclass != KNOT_CLASS_IN || !type_OK
-			    || !knot_dname_is_equal(rr->owner, cname)
+						|| type == KNOT_RRTYPE_CNAME;
+			if (rr->rclass != KNOT_CLASS_IN
 			    || knot_dname_in_bailiwick(rr->owner, query->zone_cut.name) < 0) {
 				continue;
 			}
+			const bool all_OK = type_OK && knot_dname_is_equal(rr->owner, cname);
+
+			const bool to_wire = is_final && !referral;
+
+			if (!all_OK && type == KNOT_RRTYPE_DNAME
+					&& knot_dname_in_bailiwick(cname, rr->owner) >= 1) {
+				/* This DNAME (or RRSIGs) cover the current target (`cname`),
+				 * so it is interesting and will occlude its CNAME.
+				 * We rely on CNAME being sent along with DNAME
+				 * (mandatory unless YXDOMAIN). */
+				cname_is_occluded = true;
+				uint8_t rank = get_initial_rank(rr, query, true,
+						query->flags.FORWARD || referral);
+				int ret = kr_ranked_rrarray_add(&req->answ_selected, rr,
+						rank, to_wire, query->uid, &req->pool);
+				if (ret < 0) {
+					return KR_STATE_FAIL;
+				}
+			}
+			if (!all_OK) {
+				continue;
+			}
 
 			if (rr->type == KNOT_RRTYPE_RRSIG) {
 				int rrsig_labels = knot_rrsig_labels(rr->rrs.rdata);
@@ -534,46 +550,47 @@ static int unroll_cname(knot_pkt_t *pkt, struct kr_request *req, bool referral,
 			}
 
 			/* Process records matching current SNAME */
-			int state = KR_STATE_FAIL;
-			bool to_wire = false;
-			if (is_final) {
-				/* if not referral, mark record to be written to final answer */
-				to_wire = !referral;
-			} else {
+			if (!is_final) {
 				int cnt_ = 0;
-				state = update_nsaddr(rr, query->parent, &cnt_);
-				if (state == KR_STATE_FAIL) {
+				int state = update_nsaddr(rr, query->parent, &cnt_);
+				if (state & KR_STATE_FAIL) {
 					return state;
 				}
 			}
 			uint8_t rank = get_initial_rank(rr, query, true,
 					query->flags.FORWARD || referral);
-			state = kr_ranked_rrarray_add(&req->answ_selected, rr,
-						      rank, to_wire, query->uid, &req->pool);
-			if (state != kr_ok()) {
+			int ret = kr_ranked_rrarray_add(&req->answ_selected, rr,
+						rank, to_wire, query->uid, &req->pool);
+			if (ret < 0) {
 				return KR_STATE_FAIL;
 			}
-			/* Jump to next CNAME target */
-			if ((query->stype == KNOT_RRTYPE_CNAME) || (rr->type != KNOT_RRTYPE_CNAME)) {
-				continue;
+			cname_answ_selected_i = ret;
+
+			/* Select the next CNAME target, but don't jump immediately.
+			 * There can be records for "old" cname (RRSIGs are interesting);
+			 * more importantly there might be a DNAME for `cname_is_occluded`. */
+			if (query->stype != KNOT_RRTYPE_CNAME && rr->type == KNOT_RRTYPE_CNAME) {
+				pending_cname = knot_cname_name(rr->rrs.rdata);
+				if (!pending_cname) {
+					break;
+				}
 			}
-			cname_chain_len += 1;
-			pending_cname = knot_cname_name(rr->rrs.rdata);
-			if (!pending_cname) {
-				break;
-			}
-			if (cname_chain_len > an->count || cname_chain_len > KR_CNAME_CHAIN_LIMIT) {
-				VERBOSE_MSG("<= too long cname chain\n");
-				return KR_STATE_FAIL;
-			}
-			/* Don't use pending_cname immediately.
-			 * There are can be records for "old" cname. */
 		}
 		if (!pending_cname) {
 			break;
 		}
+		if (cname_is_occluded) {
+			req->answ_selected.at[cname_answ_selected_i]->dont_cache = true;
+		}
+		if (++(query->cname_depth) > KR_CNAME_CHAIN_LIMIT) {
+			VERBOSE_MSG("<= error: CNAME chain exceeded max length %d\n",
+					/* people count objects from 0, no CNAME = 0 */
+					(int)KR_CNAME_CHAIN_LIMIT - 1);
+			return KR_STATE_FAIL;
+		}
+
 		if (knot_dname_is_equal(cname, pending_cname)) {
-			VERBOSE_MSG("<= cname chain loop\n");
+			VERBOSE_MSG("<= error: CNAME chain loop detected\n");
 			return KR_STATE_FAIL;
 		}
 		/* In strict mode, explicitly fetch each CNAME target. */
@@ -600,11 +617,7 @@ static int unroll_cname(knot_pkt_t *pkt, struct kr_request *req, bool referral,
 			cname = pending_cname;
 			break;
 		}
-	} while (++iter_count < KR_CNAME_CHAIN_LIMIT);
-	if (iter_count >= KR_CNAME_CHAIN_LIMIT) {
-		VERBOSE_MSG("<= too long cname chain\n");
-		return KR_STATE_FAIL;
-	}
+	} while (true);
 	*cname_ret = cname;
 	return kr_ok();
 }
@@ -836,7 +849,7 @@ static int process_stub(knot_pkt_t *pkt, struct kr_request *req)
 		/* KR_RANK_AUTH: we don't have the records directly from
 		 * an authoritative source, but we do trust the server and it's
 		 * supposed to only send us authoritative records. */
-		if (err != kr_ok()) {
+		if (err < 0) {
 			return KR_STATE_FAIL;
 		}
 	}
@@ -856,7 +869,8 @@ static int process_stub(knot_pkt_t *pkt, struct kr_request *req)
 }
 
 
-/** Error handling, RFC1034 5.3.3, 4d. */
+/** Error handling, RFC1034 5.3.3, 4d.
+ * NOTE: returing this does not prevent further queries (by itself). */
 static int resolve_error(knot_pkt_t *pkt, struct kr_request *req)
 {
 	return KR_STATE_FAIL;
@@ -1014,6 +1028,7 @@ static int resolve(kr_layer_t *ctx, knot_pkt_t *pkt)
 	if (!query) {
 		return ctx->state;
 	}
+	query->flags.PKT_IS_SANE = false;
 
 	WITH_VERBOSE(query) {
 		if (query->flags.TRACE) {
@@ -1036,10 +1051,15 @@ static int resolve(kr_layer_t *ctx, knot_pkt_t *pkt)
 	} else
 #endif
 	if (pkt->parsed <= KNOT_WIRE_HEADER_SIZE) {
-		VERBOSE_MSG("<= malformed response\n");
+		VERBOSE_MSG("<= malformed response (parsed %d)\n", (int)pkt->parsed);
 		return resolve_badmsg(pkt, req, query);
 	} else if (!is_paired_to_query(pkt, query)) {
-		VERBOSE_MSG("<= ignoring mismatching response\n");
+		WITH_VERBOSE(query) {
+			const char *ns_str =
+				req->upstream.addr ? kr_straddr(req->upstream.addr) : "(internal)";
+			VERBOSE_MSG("<= ignoring mismatching response from %s\n",
+					ns_str ? ns_str : "(kr_straddr failed)");
+		}
 		/* Force TCP, to work around authoritatives messing up question
 		 * without yielding to spoofed responses. */
 		query->flags.TCP = true;
@@ -1057,6 +1077,10 @@ static int resolve(kr_layer_t *ctx, knot_pkt_t *pkt)
 		return KR_STATE_CONSUME;
 	}
 
+	/* If exiting above here, there's no sense to put it into packet cache.
+	 * The most important part is to check for spoofing: is_paired_to_query() */
+	query->flags.PKT_IS_SANE = true;
+
 #ifndef NOVERBOSELOG
 	const knot_lookup_t *rcode = knot_lookup_by_id(knot_rcode_names, knot_wire_get_rcode(pkt->wire));
 #endif
@@ -1066,24 +1090,16 @@ static int resolve(kr_layer_t *ctx, knot_pkt_t *pkt)
 	case KNOT_RCODE_NOERROR:
 	case KNOT_RCODE_NXDOMAIN:
 		break; /* OK */
+	case KNOT_RCODE_YXDOMAIN: /* Basically a successful answer; name just doesn't fit. */
+		knot_wire_set_rcode(req->answer->wire, KNOT_RCODE_YXDOMAIN);
+		break;
 	case KNOT_RCODE_REFUSED:
-	case KNOT_RCODE_SERVFAIL: {
+	case KNOT_RCODE_SERVFAIL:
 		if (query->flags.STUB) {
-			 /* Pass through in stub mode */
+			 /* just pass answer through if in stub mode */
 			break;
 		}
-		VERBOSE_MSG("<= rcode: %s\n", rcode ? rcode->name : "??");
-		query->fails += 1;
-		if (query->fails >= KR_QUERY_NSRETRY_LIMIT) {
-			query->fails = 0; /* Reset per-query counter. */
-			return resolve_error(pkt, req);
-		} else {
-			if (!query->flags.FORWARD) {
-				query->flags.NO_MINIMIZE = true; /* Drop minimisation as a safe-guard. */
-			}
-			return KR_STATE_CONSUME;
-		}
-	}
+		/* fall through */
 	case KNOT_RCODE_FORMERR:
 		VERBOSE_MSG("<= rcode: %s\n", rcode ? rcode->name : "??");
 		return resolve_badmsg(pkt, req, query);
@@ -1095,13 +1111,15 @@ static int resolve(kr_layer_t *ctx, knot_pkt_t *pkt)
 		return resolve_error(pkt, req);
 	}
 
+	int state;
 	/* Forwarding/stub mode is special. */
 	if (query->flags.STUB) {
-		return process_stub(pkt, req);
+		state = process_stub(pkt, req);
+		goto rrarray_finalize;
 	}
 
 	/* Resolve authority to see if it's referral or authoritative. */
-	int state = process_authority(pkt, req);
+	state = process_authority(pkt, req);
 	switch(state) {
 	case KR_STATE_CONSUME: /* Not referral, process answer. */
 		VERBOSE_MSG("<= rcode: %s\n", rcode ? rcode->name : "??");
@@ -1115,21 +1133,33 @@ static int resolve(kr_layer_t *ctx, knot_pkt_t *pkt)
 		break;
 	}
 
+rrarray_finalize:
+	/* Finish construction of libknot-format RRsets. */
+	(void)0;
+	ranked_rr_array_t *selected[] = kr_request_selected(req);
+	for (knot_section_t i = KNOT_ANSWER; i <= KNOT_ADDITIONAL; ++i) {
+		int ret = kr_ranked_rrarray_finalize(selected[i], query->uid, &req->pool);
+		if (unlikely(ret)) {
+			return KR_STATE_FAIL;
+		}
+	}
+
 	return state;
 }
 
 /** Module implementation. */
-const kr_layer_api_t *iterate_layer(struct kr_module *module)
+int iterate_init(struct kr_module *self)
 {
-	static const kr_layer_api_t _layer = {
+	static const kr_layer_api_t layer = {
 		.begin = &begin,
 		.reset = &reset,
 		.consume = &resolve,
 		.produce = &prepare_query
 	};
-	return &_layer;
+	self->layer = &layer;
+	return kr_ok();
 }
 
-KR_MODULE_EXPORT(iterate)
+KR_MODULE_EXPORT(iterate) /* useless for builtin module, but let's be consistent */
 
 #undef VERBOSE_MSG
