@@ -1,17 +1,5 @@
 /*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 #include <ctype.h>
@@ -23,6 +11,7 @@
 #include <libknot/rrtype/rdname.h>
 #include <libknot/descriptor.h>
 #include <ucw/mempool.h>
+#include "kresconfig.h"
 #include "lib/resolve.h"
 #include "lib/layer.h"
 #include "lib/rplan.h"
@@ -107,7 +96,7 @@ static int answer_finalize_yield(kr_layer_t *ctx) { return kr_ok(); }
 	for (size_t i = (from); i < (r)->ctx->modules->len; ++i) { \
 		struct kr_module *mod = (r)->ctx->modules->at[i]; \
 		if (mod->layer) { \
-			struct kr_layer layer = {.state = (r)->state, .api = mod->layer(mod), .req = (r)}; \
+			struct kr_layer layer = {.state = (r)->state, .api = mod->layer, .req = (r)}; \
 			if (layer.api && layer.api->func) { \
 				(r)->state = layer.api->func(&layer, ##__VA_ARGS__); \
 				if ((r)->state == KR_STATE_YIELD) { \
@@ -126,8 +115,7 @@ static int answer_finalize_yield(kr_layer_t *ctx) { return kr_ok(); }
 static inline size_t layer_id(struct kr_request *req, const struct kr_layer_api *api) {
 	module_array_t *modules = req->ctx->modules;
 	for (size_t i = 0; i < modules->len; ++i) {
-		struct kr_module *mod = modules->at[i];
-		if (mod->layer && mod->layer(mod) == api) {
+		if (modules->at[i]->layer == api) {
 			return i;
 		}
 	}
@@ -150,6 +138,8 @@ static void randomized_qname_case(knot_dname_t * restrict qname, uint32_t secret
 	assert(qname);
 	const int len = knot_dname_size(qname) - 2; /* Skip first, last label. */
 	for (int i = 0; i < len; ++i) {
+		/* Note: this relies on the fact that correct label lengths
+		 * can't pass the isletter() test (by "luck"). */
 		if (isletter(*++qname)) {
 				*qname ^= ((secret >> (i & 31)) & 1) * 0x20;
 		}
@@ -162,7 +152,12 @@ static int invalidate_ns(struct kr_rplan *rplan, struct kr_query *qry)
 	if (qry->ns.addr[0].ip.sa_family != AF_UNSPEC) {
 		const char *addr = kr_inaddr(&qry->ns.addr[0].ip);
 		int addr_len = kr_inaddr_len(&qry->ns.addr[0].ip);
-		return kr_zonecut_del(&qry->zone_cut, qry->ns.name, addr, addr_len);
+		int ret = kr_zonecut_del(&qry->zone_cut, qry->ns.name, addr, addr_len);
+		/* Also remove it from the qry->ns.addr array.
+		 * That's useful at least for STUB and FORWARD modes. */
+		memmove(qry->ns.addr, qry->ns.addr + 1,
+			sizeof(qry->ns.addr[0]) * (KR_NSREP_MAXADDR - 1));
+		return ret;
 	} else {
 		return kr_zonecut_del_all(&qry->zone_cut, qry->ns.name);
 	}
@@ -203,7 +198,7 @@ static void check_empty_nonterms(struct kr_query *qry, knot_pkt_t *pkt, struct k
 		assert(target[0]);
 		target = knot_wire_next_label(target, NULL);
 	}
-	kr_cache_sync(cache);
+	kr_cache_commit(cache);
 #endif
 }
 
@@ -304,10 +299,10 @@ static int ns_fetch_cut(struct kr_query *qry, const knot_dname_t *requested_name
 	return KR_STATE_PRODUCE;
 }
 
-static int ns_resolve_addr(struct kr_query *qry, struct kr_request *param)
+static int ns_resolve_addr(struct kr_query *qry, struct kr_request *req)
 {
-	struct kr_rplan *rplan = &param->rplan;
-	struct kr_context *ctx = param->ctx;
+	struct kr_rplan *rplan = &req->rplan;
+	struct kr_context *ctx = req->ctx;
 
 
 	/* Start NS queries from root, to avoid certain cases
@@ -338,7 +333,9 @@ static int ns_resolve_addr(struct kr_query *qry, struct kr_request *param)
 			return kr_error(EAGAIN);
 		}
 		/* No IPv4 nor IPv6, flag server as unusable. */
-		VERBOSE_MSG(qry, "=> unresolvable NS address, bailing out\n");
+		++req->count_no_nsaddr;
+		VERBOSE_MSG(qry, "=> unresolvable NS address, bailing out (counter: %u)\n",
+				req->count_no_nsaddr);
 		qry->ns.reputation |= KR_NS_NOIP4 | KR_NS_NOIP6;
 		kr_nsrep_update_rep(&qry->ns, qry->ns.reputation, ctx->cache_rep);
 		invalidate_ns(rplan, qry);
@@ -454,18 +451,6 @@ static int answer_prepare(struct kr_request *req, knot_pkt_t *query)
 	return kr_ok();
 }
 
-/** @return error code, ignoring if forced to truncate the packet. */
-static int write_extra_records(const rr_array_t *arr, uint16_t reorder, knot_pkt_t *answer)
-{
-	for (size_t i = 0; i < arr->len; ++i) {
-		int err = knot_pkt_put_rotate(answer, 0, arr->at[i], reorder, 0);
-		if (err != KNOT_EOK) {
-			return err == KNOT_ESPACE ? kr_ok() : kr_error(err);
-		}
-	}
-	return kr_ok();
-}
-
 /**
  * @param all_secure optionally &&-combine security of written RRs into its value.
  *		     (i.e. if you pass a pointer to false, it will always remain)
@@ -482,6 +467,7 @@ static int write_extra_ranked_records(const ranked_rr_array_t *arr, uint16_t reo
 
 	for (size_t i = 0; i < arr->len; ++i) {
 		ranked_rr_array_entry_t * entry = arr->at[i];
+		assert(!entry->in_progress);
 		if (!entry->to_wire) {
 			continue;
 		}
@@ -521,6 +507,10 @@ static int answer_padding(struct kr_request *request)
 		assert(false);
 		return kr_error(EINVAL);
 	}
+	if (!request->qsource.flags.tls) {
+		/* Not meaningful to pad without encryption. */
+		return kr_ok();
+	}
 	int32_t padding = request->ctx->tls_padding;
 	knot_pkt_t *answer = request->answer;
 	knot_rrset_t *opt_rr = answer->opt_rr;
@@ -548,42 +538,89 @@ static int answer_padding(struct kr_request *request)
 	return kr_ok();
 }
 
-static int answer_fail(struct kr_request *request)
+/* Make a clean SERVFAIL answer. */
+static void answer_fail(struct kr_request *request)
 {
+	/* Note: OPT in SERVFAIL response is still useful for cookies/additional info. */
 	knot_pkt_t *answer = request->answer;
+	knot_rrset_t *opt_rr = answer->opt_rr; /* it gets NULLed below */
 	int ret = kr_pkt_clear_payload(answer);
 	knot_wire_clear_ad(answer->wire);
 	knot_wire_clear_aa(answer->wire);
 	knot_wire_set_rcode(answer->wire, KNOT_RCODE_SERVFAIL);
-	if (ret == 0 && answer->opt_rr) {
-		/* OPT in SERVFAIL response is still useful for cookies/additional info. */
+	if (ret == 0 && opt_rr) {
 		knot_pkt_begin(answer, KNOT_ADDITIONAL);
+		answer->opt_rr = opt_rr;
 		answer_padding(request); /* Ignore failed padding in SERVFAIL answer. */
-		ret = edns_put(answer, false);
+		edns_put(answer, false);
 	}
+}
+
+/* Append EDNS records into the answer. */
+static int answer_append_edns(struct kr_request *request)
+{
+	knot_pkt_t *answer = request->answer;
+	if (!answer->opt_rr)
+		return kr_ok();
+	int ret = answer_padding(request);
+	if (!ret) ret = knot_pkt_begin(answer, KNOT_ADDITIONAL);
+	if (!ret) ret = knot_pkt_put(answer, KNOT_COMPR_HINT_NONE,
+				     answer->opt_rr, KNOT_PF_FREE);
 	return ret;
 }
 
-static int answer_finalize(struct kr_request *request, int state)
+static void answer_finalize(struct kr_request *request)
 {
 	struct kr_rplan *rplan = &request->rplan;
 	knot_pkt_t *answer = request->answer;
 
-	/* Always set SERVFAIL for bogus answers. */
-	if (state == KR_STATE_FAIL && rplan->pending.len > 0) {
-		struct kr_query *last = array_tail(rplan->pending);
-		if ((last->flags.DNSSEC_WANT) && (last->flags.DNSSEC_BOGUS)) {
-			return answer_fail(request);
+	if (answer->rrset_count != 0) {
+		/* Non-standard: we assume the answer had been constructed.
+		 * Let's check we don't have a "collision". */
+		const ranked_rr_array_t *selected[] = kr_request_selected(request);
+		for (int psec = KNOT_ANSWER; psec <= KNOT_ADDITIONAL; ++psec) {
+			const ranked_rr_array_t *arr = selected[psec];
+			for (ssize_t i = 0; i < arr->len; ++i) {
+				if (unlikely(arr->at[i]->to_wire)) {
+					assert(false);
+					answer_fail(request);
+					return;
+				}
+			}
 		}
+		/* We only add EDNS, and we even assume AD bit was correct. */
+		if (answer_append_edns(request)) {
+			answer_fail(request);
+			return;
+		}
+		return;
 	}
 
-	struct kr_query *last = rplan->resolved.len > 0 ? array_tail(rplan->resolved) : NULL;
+	struct kr_query *const last =
+		rplan->resolved.len > 0 ? array_tail(rplan->resolved) : NULL;
 		/* TODO  ^^^^ this is slightly fragile */
 
+	if (!last) {
+		/* Suspicious: no kr_query got resolved (not even from cache),
+		 * so let's (defensively) SERVFAIL the request.
+		 * ATM many checks below depend on `last` anyway,
+		 * so this helps to avoid surprises. */
+		answer_fail(request);
+		return;
+	}
+	/* TODO: clean this up in !660 or followup, and it isn't foolproof anyway. */
+	if (last->flags.DNSSEC_BOGUS
+	    || (rplan->pending.len > 0 && array_tail(rplan->pending)->flags.DNSSEC_BOGUS)) {
+		if (!knot_wire_get_cd(request->qsource.packet->wire)) {
+			answer_fail(request);
+			return;
+		}
+	}
+
 	/* AD flag.  We can only change `secure` from true to false.
 	 * Be conservative.  Primary approach: check ranks of all RRs in wire.
 	 * Only "negative answers" need special handling. */
-	bool secure = last != NULL && state == KR_STATE_DONE /*< suspicious otherwise */
+	bool secure = last != NULL && request->state == KR_STATE_DONE /*< suspicious otherwise */
 		&& knot_pkt_qtype(answer) != KNOT_RRTYPE_RRSIG;
 	if (last && (last->flags.STUB)) {
 		secure = false; /* don't trust forwarding for now */
@@ -593,47 +630,23 @@ static int answer_finalize(struct kr_request *request, int state)
 		secure = false; /* the last answer is insecure due to opt-out */
 	}
 
+	/* Write all RRsets meant for the answer. */
 	const uint16_t reorder = last ? last->reorder : 0;
 	bool answ_all_cnames = false/*arbitrary*/;
-	if (request->answ_selected.len > 0) {
-		assert(answer->current <= KNOT_ANSWER);
-		/* Write answer records. */
-		if (answer->current < KNOT_ANSWER) {
-			knot_pkt_begin(answer, KNOT_ANSWER);
-		}
-		if (write_extra_ranked_records(&request->answ_selected, reorder,
-						answer, &secure, &answ_all_cnames))
-		{
-			return answer_fail(request);
-		}
-	}
-
-	/* Write authority records. */
-	if (answer->current < KNOT_AUTHORITY) {
-		knot_pkt_begin(answer, KNOT_AUTHORITY);
-	}
-	if (write_extra_ranked_records(&request->auth_selected, reorder,
-	    answer, &secure, NULL)) {
-		return answer_fail(request);
-	}
-	/* Write additional records. */
-	knot_pkt_begin(answer, KNOT_ADDITIONAL);
-	if (write_extra_records(&request->additional, reorder, answer)) {
-		return answer_fail(request);
-	}
-	/* Write EDNS information */
-	if (answer->opt_rr) {
-		if (request->qsource.flags.tls) {
-			if (answer_padding(request) != kr_ok()) {
-				return answer_fail(request);
-			}
-		}
-		knot_pkt_begin(answer, KNOT_ADDITIONAL);
-		int ret = knot_pkt_put(answer, KNOT_COMPR_HINT_NONE,
-				       answer->opt_rr, KNOT_PF_FREE);
-		if (ret != KNOT_EOK) {
-			return answer_fail(request);
-		}
+	if (knot_pkt_begin(answer, KNOT_ANSWER)
+	    || write_extra_ranked_records(&request->answ_selected, reorder,
+					answer, &secure, &answ_all_cnames)
+	    || knot_pkt_begin(answer, KNOT_AUTHORITY)
+	    || write_extra_ranked_records(&request->auth_selected, reorder,
+					answer, &secure, NULL)
+	    || knot_pkt_begin(answer, KNOT_ADDITIONAL)
+	    || write_extra_ranked_records(&request->add_selected, reorder,
+					answer, NULL/*not relevant to AD*/, NULL)
+	    || answer_append_edns(request)
+	   )
+	{
+		answer_fail(request);
+		return;
 	}
 
 	if (!last) secure = false; /*< should be no-op, mostly documentation */
@@ -667,42 +680,36 @@ static int answer_finalize(struct kr_request *request, int state)
 	if (!secure) {
 		knot_wire_clear_ad(answer->wire);
 	}
-
-	return kr_ok();
 }
 
 static int query_finalize(struct kr_request *request, struct kr_query *qry, knot_pkt_t *pkt)
 {
-	int ret = 0;
 	knot_pkt_begin(pkt, KNOT_ADDITIONAL);
-	if (!(qry->flags.SAFEMODE)) {
-		/* Remove any EDNS records from any previous iteration. */
-		ret = edns_erase_and_reserve(pkt);
-		if (ret == 0) {
-			ret = edns_create(pkt, request->answer, request);
+	if (qry->flags.SAFEMODE)
+		return kr_ok();
+	/* Remove any EDNS records from any previous iteration. */
+	int ret = edns_erase_and_reserve(pkt);
+	if (ret) return ret;
+	ret = edns_create(pkt, request->answer, request);
+	if (ret) return ret;
+	if (qry->flags.STUB) {
+		/* Stub resolution (ask for +rd and +do) */
+		knot_wire_set_rd(pkt->wire);
+		if (knot_pkt_has_dnssec(request->qsource.packet)) {
+			knot_edns_set_do(pkt->opt_rr);
 		}
-		if (ret == 0) {
-			/* Stub resolution (ask for +rd and +do) */
-			if (qry->flags.STUB) {
-				knot_wire_set_rd(pkt->wire);
-				if (knot_pkt_has_dnssec(request->qsource.packet)) {
-					knot_edns_set_do(pkt->opt_rr);
-				}
-				if (knot_wire_get_cd(request->qsource.packet->wire)) {
-					knot_wire_set_cd(pkt->wire);
-				}
-			/* Full resolution (ask for +cd and +do) */
-			} else if (qry->flags.FORWARD) {
-				knot_wire_set_rd(pkt->wire);
-				knot_edns_set_do(pkt->opt_rr);
-				knot_wire_set_cd(pkt->wire);
-			} else if (qry->flags.DNSSEC_WANT) {
-				knot_edns_set_do(pkt->opt_rr);
-				knot_wire_set_cd(pkt->wire);
-			}
+		if (knot_wire_get_cd(request->qsource.packet->wire)) {
+			knot_wire_set_cd(pkt->wire);
+		}
+	} else {
+		/* Full resolution (ask for +cd and +do) */
+		knot_edns_set_do(pkt->opt_rr);
+		knot_wire_set_cd(pkt->wire);
+		if (qry->flags.FORWARD) {
+			knot_wire_set_rd(pkt->wire);
 		}
 	}
-	return ret;
+	return kr_ok();
 }
 
 int kr_resolve_begin(struct kr_request *request, struct kr_context *ctx, knot_pkt_t *answer)
@@ -713,7 +720,6 @@ int kr_resolve_begin(struct kr_request *request, struct kr_context *ctx, knot_pk
 	request->options = ctx->options;
 	request->state = KR_STATE_CONSUME;
 	request->current_query = NULL;
-	array_init(request->additional);
 	array_init(request->answ_selected);
 	array_init(request->auth_selected);
 	array_init(request->add_selected);
@@ -805,7 +811,7 @@ static void update_nslist_rtt(struct kr_context *ctx, struct kr_query *qry, cons
 	/* Calculate total resolution time from the time the query was generated. */
 	uint64_t elapsed = kr_now() - qry->timestamp_mono;
 	elapsed = elapsed > UINT_MAX ? UINT_MAX : elapsed;
- 
+
 	/* NSs in the preference list prior to the one who responded will be penalised
 	 * with the RETRY timer interval. This is because we know they didn't respond
 	 * for N retries, so their RTT must be at least N * RETRY.
@@ -852,7 +858,7 @@ static void update_nslist_score(struct kr_request *request, struct kr_query *qry
 {
 	struct kr_context *ctx = request->ctx;
 	/* On successful answer, update preference list RTT and penalise timer  */
-	if (request->state != KR_STATE_FAIL) {
+	if (!(request->state & KR_STATE_FAIL)) {
 		/* Update RTT information for preference list */
 		update_nslist_rtt(ctx, qry, src);
 		/* Do not complete NS address resolution on soft-fail. */
@@ -928,11 +934,28 @@ int kr_resolve_consume(struct kr_request *request, const struct sockaddr *src, k
 		update_nslist_score(request, qry, src, packet);
 	}
 	/* Resolution failed, invalidate current NS. */
-	if (request->state == KR_STATE_FAIL) {
+	if (request->state & KR_STATE_FAIL) {
 		invalidate_ns(rplan, qry);
 		qry->flags.RESOLVED = false;
 	}
 
+	/* For multiple errors in a row; invalidate_ns() is not enough. */
+	if (!qry->flags.CACHED) {
+		if (request->state & KR_STATE_FAIL) {
+			if (++request->count_fail_row > KR_CONSUME_FAIL_ROW_LIMIT) {
+				if (VERBOSE_STATUS || kr_log_rtrace_enabled(request)) {
+					kr_log_req(request, 0, 2, "resl",
+						"=> too many failures in a row, "
+						"bail out (mitigation for NXNSAttack "
+						"CVE-2020-12667)");
+				}
+				return KR_STATE_FAIL;
+			}
+		} else {
+			request->count_fail_row = 0;
+		}
+	}
+
 	/* Pop query if resolved. */
 	if (request->state == KR_STATE_YIELD) {
 		return KR_STATE_PRODUCE; /* Requery */
@@ -1283,7 +1306,7 @@ static int zone_cut_check(struct kr_request *request, struct kr_query *qry, knot
 	int state = KR_STATE_FAIL;
 	do {
 		state = ns_fetch_cut(qry, requested_name, request, packet);
-		if (state == KR_STATE_DONE || state == KR_STATE_FAIL) {
+		if (state == KR_STATE_DONE || (state & KR_STATE_FAIL)) {
 			return state;
 		} else if (state == KR_STATE_CONSUME) {
 			requested_name = knot_wire_next_label(requested_name, NULL);
@@ -1353,7 +1376,7 @@ int kr_resolve_produce(struct kr_request *request, struct sockaddr **dst, int *t
 		/* Resolve current query and produce dependent or finish */
 		request->state = KR_STATE_PRODUCE;
 		ITERATE_LAYERS(request, qry, produce, packet);
-		if (request->state != KR_STATE_FAIL && knot_wire_get_qr(packet->wire)) {
+		if (!(request->state & KR_STATE_FAIL) && knot_wire_get_qr(packet->wire)) {
 			/* Produced an answer from cache, consume it. */
 			qry->secret = 0;
 			request->state = KR_STATE_CONSUME;
@@ -1371,7 +1394,7 @@ int kr_resolve_produce(struct kr_request *request, struct sockaddr **dst, int *t
 		ITERATE_LAYERS(request, qry, reset);
 		return kr_rplan_empty(rplan) ? KR_STATE_DONE : KR_STATE_PRODUCE;
 	}
-	
+
 
 	/* This query has RD=0 or is ANY, stop here. */
 	if (qry->stype == KNOT_RRTYPE_ANY ||
@@ -1392,6 +1415,11 @@ int kr_resolve_produce(struct kr_request *request, struct sockaddr **dst, int *t
 
 ns_election:
 
+	if (unlikely(request->count_no_nsaddr >= KR_COUNT_NO_NSADDR_LIMIT)) {
+		VERBOSE_MSG(qry, "=> too many unresolvable NSs, bail out "
+				"(mitigation for NXNSAttack CVE-2020-12667)\n");
+		return KR_STATE_FAIL;
+	}
 	/* If the query has already selected a NS and is waiting for IPv4/IPv6 record,
 	 * elect best address only, otherwise elect a completely new NS.
 	 */
@@ -1541,13 +1569,13 @@ int kr_resolve_checkout(struct kr_request *request, const struct sockaddr *src,
 	 * don't affect the resolution or rest of the processing. */
 	int state = request->state;
 	ITERATE_LAYERS(request, qry, checkout, packet, dst, type);
-	if (request->state == KR_STATE_FAIL) {
+	if (request->state & KR_STATE_FAIL) {
 		request->state = state; /* Restore */
 		return kr_error(ECANCELED);
 	}
 
 	/* Randomize query case (if secret changed) */
-	knot_dname_t *qname = (knot_dname_t *)knot_pkt_qname(packet);
+	knot_dname_t *qname = knot_pkt_qname(packet);
 	if (qry->secret != old_minimization_secret) {
 		randomized_qname_case(qname, qry->secret);
 	}
@@ -1589,25 +1617,24 @@ int kr_resolve_checkout(struct kr_request *request, const struct sockaddr *src,
 
 int kr_resolve_finish(struct kr_request *request, int state)
 {
+	request->state = state;
 	/* Finalize answer and construct wire-buffer. */
 	ITERATE_LAYERS(request, NULL, answer_finalize);
-	if (request->state == KR_STATE_FAIL) {
-		state = KR_STATE_FAIL;
-	} else if (answer_finalize(request, state) != 0) {
-		state = KR_STATE_FAIL;
-	}
+	answer_finalize(request);
 
-	/* Error during processing, internal failure */
-	if (state != KR_STATE_DONE) {
-		knot_pkt_t *answer = request->answer;
-		if (knot_wire_get_rcode(answer->wire) == KNOT_RCODE_NOERROR) {
-			knot_wire_clear_ad(answer->wire);
-			knot_wire_clear_aa(answer->wire);
-			knot_wire_set_rcode(answer->wire, KNOT_RCODE_SERVFAIL);
+	/* Defensive style, in case someone has forgotten.
+	 * Beware: non-empty answers do make sense even with SERVFAIL case, etc. */
+	if (request->state != KR_STATE_DONE) {
+		uint8_t *wire = request->answer->wire;
+		switch (knot_wire_get_rcode(wire)) {
+		case KNOT_RCODE_NOERROR:
+		case KNOT_RCODE_NXDOMAIN:
+			knot_wire_clear_ad(wire);
+			knot_wire_clear_aa(wire);
+			knot_wire_set_rcode(wire, KNOT_RCODE_SERVFAIL);
 		}
 	}
 
-	request->state = state;
 	ITERATE_LAYERS(request, NULL, finish);
 
 #ifndef NOVERBOSELOG
