@@ -1,17 +1,5 @@
 /*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 #pragma once
@@ -157,7 +145,12 @@ typedef array_t(struct kr_module *) module_array_t;
 struct kr_context
 {
 	struct kr_qflags options;
+
+	/** Default EDNS towards *both* clients and upstream.
+	 * LATER: consider splitting the two, e.g. allow separately
+	 * configured limits for UDP packet size (say, LAN is under control). */
 	knot_rrset_t *opt_rr;
+
 	map_t trust_anchors;
 	map_t negative_anchors;
 	struct kr_zonecut root_hints;
@@ -177,8 +170,9 @@ struct kr_context
 /* Kept outside, because kres-gen.lua can't handle this depth
  * (and lines here were too long anyway). */
 struct kr_request_qsource_flags {
-	bool tcp:1; /**< true if the request is on TCP (or TLS); only meaningful if (dst_addr). */
-	bool tls:1; /**< true if the request is on TLS; only meaningful if (dst_addr). */
+	bool tcp:1; /**< true if the request is not on UDP; only meaningful if (dst_addr). */
+	bool tls:1; /**< true if the request is encrypted; only meaningful if (dst_addr). */
+	bool http:1; /**< true if the request is on HTTP; only meaningful if (dst_addr). */
 };
 
 /**
@@ -198,7 +192,9 @@ struct kr_request {
 	struct {
 		/** Address that originated the request. NULL for internal origin. */
 		const struct sockaddr *addr;
-		/** Address that accepted the request.  NULL for internal origin. */
+		/** Address that accepted the request.  NULL for internal origin.
+		 * Beware: in case of UDP on wildcard address it will be wildcard;
+		 * closely related: issue #173. */
 		const struct sockaddr *dst_addr;
 		const knot_pkt_t *packet;
 		struct kr_request_qsource_flags flags; /**< See definition above. */
@@ -213,7 +209,6 @@ struct kr_request {
 	ranked_rr_array_t answ_selected;
 	ranked_rr_array_t auth_selected;
 	ranked_rr_array_t add_selected;
-	rr_array_t additional;
 	bool answ_validated; /**< internal to validator; beware of caching, etc. */
 	bool auth_validated; /**< see answ_validated ^^ ; TODO */
 
@@ -232,7 +227,8 @@ struct kr_request {
 	int vars_ref; /**< Reference to per-request variable table. LUA_NOREF if not set. */
 	knot_mm_t pool;
 	unsigned int uid; /** for logging purposes only */
-	void *daemon_context; /** pointer to worker from daemon. Can be used in modules. */
+	unsigned int count_no_nsaddr;
+	unsigned int count_fail_row;
 };
 
 /** Initializer for an array of *_selected. */
@@ -308,7 +304,7 @@ int kr_resolve_checkout(struct kr_request *request, const struct sockaddr *src,
  *       be destroyed, as it's owned by caller.
  *
  * @param  request request state
- * @param  state   either DONE or FAIL state
+ * @param  state   either DONE or FAIL state (to be assigned to request->state)
  * @return         DONE
  */
 KR_EXPORT
