@@ -1,17 +1,5 @@
 /*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 #include <libknot/descriptor.h>
@@ -24,8 +12,6 @@
 #include "lib/layer.h"
 
 #define VERBOSE_MSG(qry, ...) QRVERBOSE(qry, "plan",  __VA_ARGS__)
-#define QUERY_PROVIDES(q, name, cls, type) \
-    ((q)->sclass == (cls) && (q)->stype == type && knot_dname_is_equal((q)->sname, name))
 
 inline static unsigned char chars_or(const unsigned char a, const unsigned char b)
 {
@@ -190,6 +176,12 @@ static struct kr_query *kr_rplan_push_query(struct kr_rplan *rplan,
 		}
 	}
 
+	assert((rplan->pending.len == 0 && rplan->resolved.len == 0)
+		== (rplan->initial == NULL));
+	if (rplan->initial == NULL) {
+		rplan->initial = qry;
+	}
+
 	array_push(rplan->pending, qry);
 
 	return qry;
@@ -253,7 +245,12 @@ int kr_rplan_pop(struct kr_rplan *rplan, struct kr_query *qry)
 	/* Find the query, it will likely be on top */
 	for (size_t i = rplan->pending.len; i > 0; i--) {
 		if (rplan->pending.at[i - 1] == qry) {
-			array_del(rplan->pending, i - 1);
+			/* Delete i-1 element by *sliding* the rest,
+			 * contrary to array_del() */
+			for (size_t j = i; j < rplan->pending.len; ++j)
+				rplan->pending.at[j - 1] = rplan->pending.at[j];
+			array_pop(rplan->pending);
+
 			array_push(rplan->resolved, qry);
 			break;
 		}
@@ -264,7 +261,8 @@ int kr_rplan_pop(struct kr_rplan *rplan, struct kr_query *qry)
 bool kr_rplan_satisfies(struct kr_query *closure, const knot_dname_t *name, uint16_t cls, uint16_t type)
 {
 	while (name && closure) {
-		if (QUERY_PROVIDES(closure, name, cls, type)) {
+		if (closure->sclass == cls && closure->stype == type
+		    && knot_dname_is_equal(closure->sname, name)) {
 			return true;
 		}
 		closure = closure->parent;
