@@ -1,48 +1,40 @@
 /*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
-#include <stdarg.h>
-#include <string.h>
-#include <stdlib.h>
-#include <stdio.h>
-#include <arpa/inet.h>
-#include <sys/time.h>
-#include <contrib/cleanup.h>
-#include <contrib/ccan/asprintf/asprintf.h>
-#include <ucw/mempool.h>
+#include "lib/utils.h"
+
+#include "contrib/ccan/asprintf/asprintf.h"
+#include "contrib/cleanup.h"
+#include "contrib/ucw/mempool.h"
+#include "kresconfig.h"
+#include "lib/defines.h"
+#include "lib/generic/array.h"
+#include "lib/module.h"
+#include "lib/nsrep.h"
+#include "lib/resolve.h"
+
 #include <gnutls/gnutls.h>
 #include <libknot/descriptor.h>
 #include <libknot/dname.h>
-#include <libknot/rrtype/rrsig.h>
 #include <libknot/rrset-dump.h>
+#include <libknot/rrtype/rrsig.h>
 #include <libknot/version.h>
 #include <uv.h>
 
-#include "lib/defines.h"
-#include "lib/utils.h"
-#include "lib/generic/array.h"
-#include "lib/nsrep.h"
-#include "lib/module.h"
-#include "lib/resolve.h"
-
+#include <arpa/inet.h>
+#include <stdarg.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+#include <sys/time.h>
+#include <sys/stat.h>
+#include <sys/statvfs.h>
+#include <sys/un.h>
 
 /* Always compile-in log symbols, even if disabled. */
 #undef kr_verbose_status
 #undef kr_verbose_set
-#undef kr_log_verbose
 
 /* Logging & debugging */
 bool kr_verbose_status = false;
@@ -71,6 +63,18 @@ void *mm_malloc(void *ctx, size_t n)
 	(void)ctx;
 	return malloc(n);
 }
+void *mm_malloc_aligned(void *ctx, size_t n)
+{
+	size_t alignment = (size_t)ctx;
+	void *res;
+	int err = posix_memalign(&res, alignment, n);
+	if (err == 0) {
+		return res;
+	} else {
+		assert(err == -1 && errno == ENOMEM);
+		return NULL;
+	}
+}
 
 /*
  * Macros.
@@ -116,54 +120,50 @@ bool kr_verbose_set(bool status)
 	return kr_verbose_status;
 }
 
-void kr_log_verbose(const char *fmt, ...)
+static void kr_vlog_req(
+	const struct kr_request * const req, uint32_t qry_uid,
+	const unsigned int indent, const char *source, const char *fmt,
+	va_list args)
 {
-	if (kr_verbose_status) {
-		va_list args;
-		va_start(args, fmt);
-		vprintf(fmt, args);
-		va_end(args);
-		fflush(stdout);
-	}
+	struct mempool *mp = mp_new(512);
+
+	const uint32_t req_uid = req ? req->uid : 0;
+	char *msg = mp_printf(mp, "[%05u.%02u][%-4s] %*s",
+				req_uid, qry_uid, source, indent, "");
+
+	msg = mp_vprintf_append(mp, msg, fmt, args);
+
+	if (kr_log_rtrace_enabled(req))
+		req->trace_log(req, msg);
+	else
+		/* caller is responsible for detecting verbose mode, use QRVERBOSE() macro */
+		printf("%s", msg);
+
+	mp_delete(mp);
 }
 
-void kr_log_qverbose_impl(const struct kr_query *qry, const char *cls, const char *fmt, ...)
+void kr_log_req(const struct kr_request * const req, uint32_t qry_uid,
+		const unsigned int indent, const char *source, const char *fmt, ...)
+{
+	va_list args;
+	va_start(args, fmt);
+	kr_vlog_req(req, qry_uid, indent, source, fmt, args);
+	va_end(args);
+}
+
+void kr_log_q(const struct kr_query * const qry,
+		const char *source, const char *fmt, ...)
 {
 	unsigned ind = 0;
 	for (const struct kr_query *q = qry; q; q = q->parent)
 		ind += 2;
-	uint32_t qry_uid = qry ? qry->uid : 0;
-	uint32_t req_uid = qry && qry->request ? qry->request->uid : 0;
-	/* Simplified kr_log_verbose() calls, first prefix then passed fmt...
-	 * Calling it would take about the same amount of code. */
-	printf("[%05u.%02u][%s] %*s", req_uid, qry_uid, cls, ind, "");
-	va_list args;
-	va_start(args, fmt);
-	vprintf(fmt, args);
-	va_end(args);
-	fflush(stdout);
-}
-
-bool kr_log_trace(const struct kr_query *query, const char *source, const char *fmt, ...)
-{
-	if (!kr_log_trace_enabled(query)) {
-		return false;
-	}
-
-	auto_free char *msg = NULL;
+	const uint32_t qry_uid = qry ? qry->uid : 0;
+	const struct kr_request *req = qry ? qry->request : NULL;
 
 	va_list args;
 	va_start(args, fmt);
-	int len = vasprintf(&msg, fmt, args);
+	kr_vlog_req(req, qry_uid, ind, source, fmt, args);
 	va_end(args);
-
-	/* Check formatting result before logging */
-	if (len < 0) {
-		return false;
-	}
-
-	query->request->trace_log(query, source, msg);
-	return true;
 }
 
 char* kr_strcatdup(unsigned n, ...)
@@ -210,7 +210,7 @@ char* kr_strcatdup(unsigned n, ...)
 	return result;
 }
 
-int kr_memreserve(void *baton, char **mem, size_t elm_size, size_t want, size_t *have)
+int kr_memreserve(void *baton, void **mem, size_t elm_size, size_t want, size_t *have)
 {
     if (*have >= want) {
         return 0;
@@ -219,8 +219,10 @@ int kr_memreserve(void *baton, char **mem, size_t elm_size, size_t want, size_t
         size_t next_size = array_next_count(want);
         void *mem_new = mm_alloc(pool, next_size * elm_size);
         if (mem_new != NULL) {
-            memcpy(mem_new, *mem, (*have)*(elm_size));
-            mm_free(pool, *mem);
+	    if (*mem) { /* 0-length memcpy from NULL isn't technically OK */
+		memcpy(mem_new, *mem, (*have)*(elm_size));
+		mm_free(pool, *mem);
+	    }
             *mem = mem_new;
             *have = next_size;
             return 0;
@@ -330,6 +332,7 @@ int kr_sockaddr_len(const struct sockaddr *addr)
 	switch (addr->sa_family) {
 	case AF_INET:  return sizeof(struct sockaddr_in);
 	case AF_INET6: return sizeof(struct sockaddr_in6);
+	case AF_UNIX:  return sizeof(struct sockaddr_un);
 	default:       return kr_error(EINVAL);
 	}
 }
@@ -393,11 +396,20 @@ void kr_inaddr_set_port(struct sockaddr *addr, uint16_t port)
 
 int kr_inaddr_str(const struct sockaddr *addr, char *buf, size_t *buflen)
 {
-	if (!addr || !buf || !buflen) {
+	if (!addr) {
+		return kr_error(EINVAL);
+	}
+	return kr_ntop_str(addr->sa_family, kr_inaddr(addr), kr_inaddr_port(addr),
+			   buf, buflen);
+}
+
+int kr_ntop_str(int family, const void *src, uint16_t port, char *buf, size_t *buflen)
+{
+	if (!src || !buf || !buflen) {
 		return kr_error(EINVAL);
 	}
 
-	if (!inet_ntop(addr->sa_family, kr_inaddr(addr), buf, *buflen)) {
+	if (!inet_ntop(family, src, buf, *buflen)) {
 		return kr_error(errno);
 	}
 	const int len = strlen(buf);
@@ -408,7 +420,7 @@ int kr_inaddr_str(const struct sockaddr *addr, char *buf, size_t *buflen)
 	}
 	*buflen = len_need;
 	buf[len] = '#';
-	u16tostr((uint8_t *)&buf[len + 1], kr_inaddr_port(addr));
+	u16tostr((uint8_t *)&buf[len + 1], port);
 	buf[len_need - 1] = 0;
 	return kr_ok();
 }
@@ -418,6 +430,9 @@ int kr_straddr_family(const char *addr)
 	if (!addr) {
 		return kr_error(EINVAL);
 	}
+	if (addr[0] == '/') {
+		return AF_UNIX;
+	}
 	if (strchr(addr, ':')) {
 		return AF_INET6;
 	}
@@ -433,28 +448,40 @@ int kr_family_len(int family)
 	}
 }
 
-struct sockaddr * kr_straddr_socket(const char *addr, int port)
+struct sockaddr * kr_straddr_socket(const char *addr, int port, knot_mm_t *pool)
 {
 	switch (kr_straddr_family(addr)) {
 	case AF_INET: {
-		struct sockaddr_in *res = malloc(sizeof(*res));
+		struct sockaddr_in *res = mm_alloc(pool, sizeof(*res));
 		if (uv_ip4_addr(addr, port, res) >= 0) {
 			return (struct sockaddr *)res;
 		} else {
-			free(res);
+			mm_free(pool, res);
 			return NULL;
 		}
 	}
 	case AF_INET6: {
-		struct sockaddr_in6 *res = malloc(sizeof(*res));
+		struct sockaddr_in6 *res = mm_alloc(pool, sizeof(*res));
 		if (uv_ip6_addr(addr, port, res) >= 0) {
 			return (struct sockaddr *)res;
 		} else {
-			free(res);
+			mm_free(pool, res);
 			return NULL;
 		}
 	}
+	case AF_UNIX: {
+		struct sockaddr_un *res;
+		const size_t alen = strlen(addr) + 1;
+		if (alen > sizeof(res->sun_path)) {
+			return NULL;
+		}
+		res = mm_alloc(pool, sizeof(*res));
+		res->sun_family = AF_UNIX;
+		memcpy(res->sun_path, addr, alen);
+		return (struct sockaddr *)res;
+	}
 	default:
+		assert(!EINVAL);
 		return NULL;
 	}
 }
@@ -491,54 +518,29 @@ int kr_straddr_subnet(void *dst, const char *addr)
 	return bit_len;
 }
 
-int kr_straddr_split(const char *addr, char *buf, size_t buflen, uint16_t *port)
+int kr_straddr_split(const char *instr, char ipaddr[static restrict (INET6_ADDRSTRLEN + 1)],
+		     uint16_t *port)
 {
-	const int base = 10;
-	long p = 0;
-	size_t addrlen = strlen(addr);
-	char *p_start = strchr(addr, '@');
-	char *p_end;
-
-	if (!p_start) {
-		p_start = strchr(addr, '#');
+	assert(instr && ipaddr && port);
+	/* Find where port number starts. */
+	const char *p_start = strchr(instr, '@');
+	if (!p_start)
+		p_start = strchr(instr, '#');
+	if (p_start) { /* Get and check the port number. */
+		if (p_start[1] == '\0') /* Don't accept empty port string. */
+			return kr_error(EILSEQ);
+		char *p_end;
+		long p = strtol(p_start + 1, &p_end, 10);
+		if (*p_end != '\0' || p <= 0 || p > UINT16_MAX)
+			return kr_error(EILSEQ);
+		*port = p;
 	}
-
-	if (p_start) {
-		if (p_start[1] != '\0'){
-			p = strtol(p_start + 1, &p_end, base);
-			if (*p_end != '\0' || p <= 0 || p > UINT16_MAX) {
-				return kr_error(EINVAL);
-			}
-		}
-		addrlen = p_start - addr;
-	}
-
-	/* Check if address is valid. */
-	if (addrlen >= INET6_ADDRSTRLEN) {
-		return kr_error(EINVAL);
-	}
-
-	char str[INET6_ADDRSTRLEN];
-	struct sockaddr_storage ss;
-
-	memcpy(str, addr, addrlen); str[addrlen] = '\0';
-
-	int family = kr_straddr_family(str);
-	if (family == kr_error(EINVAL) || !inet_pton(family, str, &ss)) {
-		return kr_error(EINVAL);
-	}
-
-	/* Address and port contains valid values, return it to caller */
-	if (buf) {
-		if (addrlen >= buflen) {
-			return kr_error(ENOSPC);
-		}
-		memcpy(buf, addr, addrlen); buf[addrlen] = '\0';
-	}
-	if (port) {
-		*port = (uint16_t)p;
-	}
-
+	/* Copy the address. */
+	const size_t addrlen = p_start ? p_start - instr : strlen(instr);
+	if (addrlen > INET6_ADDRSTRLEN)
+		return kr_error(EILSEQ);
+	memcpy(ipaddr, instr, addrlen);
+	ipaddr[addrlen] = '\0';
 	return kr_ok();
 }
 
@@ -677,6 +679,11 @@ static int to_wire_ensure_unique(ranked_rr_array_t *array, size_t index)
 	return kr_ok();
 }
 
+/* Implementation overview of _add() and _finalize():
+ * - for rdata we just maintain a list of pointers (in knot_rrset_t::additional)
+ * - we only construct the final rdataset at the end (and thus more efficiently)
+ */
+typedef array_t(knot_rdata_t *) rdata_array_t;
 int kr_ranked_rrarray_add(ranked_rr_array_t *array, const knot_rrset_t *rr,
 			  uint8_t rank, bool to_wire, uint32_t qry_uid, knot_mm_t *pool)
 {
@@ -691,47 +698,91 @@ int kr_ranked_rrarray_add(ranked_rr_array_t *array, const knot_rrset_t *rr,
 		}
 		if (stashed->qry_uid != qry_uid) {
 			break;
+			/* We do not guarantee merging RRs "across" any point that switched
+			 * to processing a different upstream packet (i.e. qry_uid).
+			 * In particular, iterator never returns KR_STATE_YIELD. */
 		}
 		if (!rrsets_match(stashed->rr, rr)) {
 			continue;
 		}
 		/* Found the entry to merge with.  Check consistency and merge. */
-		bool ok = stashed->rank == rank && !stashed->cached;
+		bool ok = stashed->rank == rank && !stashed->cached && stashed->in_progress;
 		if (!ok) {
 			assert(false);
 			return kr_error(EEXIST);
 		}
+		/* assert(rr->rrs.count == 1); */
+		/* ^^ shouldn't be a problem for this function, but it's probably a bug */
+
 		/* It may happen that an RRset is first considered useful
 		 * (to_wire = false, e.g. due to being part of glue),
 		 * and later we may find we also want it in the answer. */
 		stashed->to_wire = stashed->to_wire || to_wire;
 
-		return knot_rdataset_merge(&stashed->rr->rrs, &rr->rrs, pool);
+		/* We just add the reference into this in_progress RRset. */
+		rdata_array_t *ra = stashed->rr->additional;
+		if (ra == NULL) {
+			/* RRset not in array format yet -> convert it. */
+			ra = stashed->rr->additional = mm_alloc(pool, sizeof(*ra));
+			if (!ra) {
+				return kr_error(ENOMEM);
+			}
+			array_init(*ra);
+			int ret = array_reserve_mm(*ra, stashed->rr->rrs.count + rr->rrs.count,
+							kr_memreserve, pool);
+			if (ret) {
+				return kr_error(ret);
+			}
+			knot_rdata_t *r_it = stashed->rr->rrs.rdata;
+			for (int ri = 0; ri < stashed->rr->rrs.count;
+					++ri, r_it = knot_rdataset_next(r_it)) {
+				if (array_push(*ra, r_it) < 0) {
+					abort();
+				}
+			}
+		} else {
+			int ret = array_reserve_mm(*ra, ra->len + rr->rrs.count,
+							kr_memreserve, pool);
+			if (ret) {
+				return kr_error(ret);
+			}
+		}
+		/* Append to the array. */
+		knot_rdata_t *r_it = rr->rrs.rdata;
+		for (int ri = 0; ri < rr->rrs.count;
+				++ri, r_it = knot_rdataset_next(r_it)) {
+			if (array_push(*ra, r_it) < 0) {
+				abort();
+			}
+		}
+		return i;
 	}
 
 	/* No stashed rrset found, add */
 	int ret = array_reserve_mm(*array, array->len + 1, kr_memreserve, pool);
-	if (ret != 0) {
-		return kr_error(ENOMEM);
+	if (ret) {
+		return kr_error(ret);
 	}
 
 	ranked_rr_array_entry_t *entry = mm_alloc(pool, sizeof(ranked_rr_array_entry_t));
 	if (!entry) {
 		return kr_error(ENOMEM);
 	}
-	knot_rrset_t *copy = knot_rrset_copy(rr, pool);
-	if (!copy) {
+	memset(entry, 0, sizeof(*entry)); /* default all to zeros */
+
+	knot_rrset_t *rr_new = knot_rrset_new(rr->owner, rr->type, rr->rclass, rr->ttl, pool);
+	if (!rr_new) {
 		mm_free(pool, entry);
 		return kr_error(ENOMEM);
 	}
+	rr_new->rrs = rr->rrs;
+	assert(rr_new->additional == NULL);
 
 	entry->qry_uid = qry_uid;
-	entry->rr = copy;
+	entry->rr = rr_new;
 	entry->rank = rank;
-	entry->revalidation_cnt = 0;
-	entry->cached = false;
-	entry->yielded = false;
 	entry->to_wire = to_wire;
+	entry->in_progress = true;
 	if (array_push(*array, entry) < 0) {
 		/* Silence coverity.  It shouldn't be possible to happen,
 		 * due to the array_reserve_mm call above. */
@@ -739,9 +790,89 @@ int kr_ranked_rrarray_add(ranked_rr_array_t *array, const knot_rrset_t *rr,
 		return kr_error(ENOMEM);
 	}
 
-	return to_wire_ensure_unique(array, array->len - 1);
+	ret = to_wire_ensure_unique(array, array->len - 1);
+	if (ret < 0) return ret;
+	return array->len - 1;
 }
 
+/** Comparator for qsort() on an array of knot_data_t pointers. */
+static int rdata_p_cmp(const void *rp1, const void *rp2)
+{
+	/* Just correct types of the parameters and pass them dereferenced. */
+	const knot_rdata_t
+		*const *r1 = rp1,
+		*const *r2 = rp2;
+	return knot_rdata_cmp(*r1, *r2);
+}
+int kr_ranked_rrarray_finalize(ranked_rr_array_t *array, uint32_t qry_uid, knot_mm_t *pool)
+{
+	for (ssize_t array_i = array->len - 1; array_i >= 0; --array_i) {
+		ranked_rr_array_entry_t *stashed = array->at[array_i];
+		if (stashed->qry_uid != qry_uid) {
+			continue; /* We apparently can't always short-cut the cycle. */
+		}
+		if (!stashed->in_progress) {
+			continue;
+		}
+		rdata_array_t *ra = stashed->rr->additional;
+		if (!ra) {
+			/* No array, so we just need to copy the rdataset. */
+			knot_rdataset_t *rds = &stashed->rr->rrs;
+			knot_rdataset_t tmp = *rds;
+			int ret = knot_rdataset_copy(rds, &tmp, pool);
+			if (ret) {
+				return kr_error(ret);
+			}
+		} else {
+			/* Multiple RRs; first: sort the array. */
+			stashed->rr->additional = NULL;
+			qsort(ra->at, ra->len, sizeof(ra->at[0]), rdata_p_cmp);
+			/* Prune duplicates: NULL all except the last instance. */
+			int dup_count = 0;
+			for (int i = 0; i + 1 < ra->len; ++i) {
+				if (knot_rdata_cmp(ra->at[i], ra->at[i + 1]) == 0) {
+					ra->at[i] = NULL;
+					++dup_count;
+					QRVERBOSE(NULL, "iter", "deleted duplicate RR\n");
+				}
+			}
+			/* Prepare rdataset, except rdata contents. */
+			int size_sum = 0;
+			for (int i = 0; i < ra->len; ++i) {
+				if (ra->at[i]) {
+					size_sum += knot_rdata_size(ra->at[i]->len);
+				}
+			}
+			knot_rdataset_t *rds = &stashed->rr->rrs;
+			rds->count = ra->len - dup_count;
+			#if KNOT_VERSION_HEX >= 0x020900
+				rds->size = size_sum;
+			#endif
+			if (size_sum) {
+				rds->rdata = mm_alloc(pool, size_sum);
+				if (!rds->rdata) {
+					return kr_error(ENOMEM);
+				}
+			} else {
+				rds->rdata = NULL;
+			}
+			/* Everything is ready; now just copy all the rdata. */
+			uint8_t *raw_it = (uint8_t *)rds->rdata;
+			for (int i = 0; i < ra->len; ++i) {
+				if (ra->at[i] && size_sum/*linters*/) {
+					const int size = knot_rdata_size(ra->at[i]->len);
+					memcpy(raw_it, ra->at[i], size);
+					raw_it += size;
+				}
+			}
+			assert(raw_it == (uint8_t *)rds->rdata + size_sum);
+		}
+		stashed->in_progress = false;
+	}
+	return kr_ok();
+}
+
+
 int kr_ranked_rrarray_set_wire(ranked_rr_array_t *array, bool to_wire,
 			       uint32_t qry_uid, bool check_dups,
 			       bool (*extraCheck)(const ranked_rr_array_entry_t *))
@@ -769,7 +900,7 @@ static char *callprop(struct kr_module *module, const char *prop, const char *in
 	if (!module || !module->props || !prop) {
 		return NULL;
 	}
-	for (const struct kr_prop *p = module->props(); p && p->name; ++p) {
+	for (const struct kr_prop *p = module->props; p && p->name; ++p) {
 		if (p->cb != NULL && strcmp(p->name, prop) == 0) {
 			return p->cb(env, module, input);
 		}
@@ -849,6 +980,21 @@ static char *print_section_opt(struct mempool *mp, char *endp, const knot_rrset_
 
 }
 
+/**
+ * Detect if qname contains an uppercase letter.
+ */
+static bool qname_has_uppercase(const knot_dname_t *qname) {
+	const int len = knot_dname_size(qname) - 1;  /* skip root label at the end */
+	for (int i = 1; i < len; ++i) {  /* skip first length byte */
+		/* Note: this relies on the fact that correct label lengths
+		 * can't pass this test by "luck" and that correctness
+		 * is checked earlier by packet parser. */
+		if (qname[i] >= 'A' && qname[i] <= 'Z')
+			return true;
+	}
+	return false;
+}
+
 char *kr_pkt_text(const knot_pkt_t *pkt)
 {
 	if (!pkt) {
@@ -896,7 +1042,14 @@ char *kr_pkt_text(const knot_pkt_t *pkt)
 	if (qdcount == 1) {
 		KR_DNAME_GET_STR(qname, knot_pkt_qname(pkt));
 		KR_RRTYPE_GET_STR(rrtype, knot_pkt_qtype(pkt));
-		ptr = mp_printf_append(mp, ptr, ";; QUESTION SECTION\n%s\t\t%s\n", qname, rrtype);
+		const char *qnwarn;
+		if (qname_has_uppercase(knot_pkt_qname(pkt)))
+			qnwarn = \
+"; WARNING! Uppercase letters indicate positions with letter case mismatches!\n"
+";          Normally you should see all-lowercase qname here.\n";
+		else
+			qnwarn = "";
+		ptr = mp_printf_append(mp, ptr, ";; QUESTION SECTION\n%s%s\t\t%s\n", qnwarn, qname, rrtype);
 	} else if (qdcount > 1) {
 		ptr = mp_printf_append(mp, ptr, ";; Warning: unsupported QDCOUNT %hu\n", qdcount);
 	}
@@ -948,6 +1101,11 @@ uint64_t kr_now()
 	return uv_now(uv_default_loop());
 }
 
+void kr_uv_free_cb(uv_handle_t* handle)
+{
+	free(handle->data);
+}
+
 const char *kr_strptime_diff(const char *format, const char *time1_str,
 		             const char *time0_str, double *diff) {
 	assert(format != NULL);
@@ -1061,6 +1219,10 @@ void kr_rrset_init(knot_rrset_t *rrset, knot_dname_t *owner,
 	assert(rrset);
 	knot_rrset_init(rrset, owner, type, rclass, ttl);
 }
+uint16_t kr_pkt_has_dnssec(const knot_pkt_t *pkt)
+{
+	return knot_pkt_has_dnssec(pkt);
+}
 uint16_t kr_pkt_qclass(const knot_pkt_t *pkt)
 {
 	return knot_pkt_qclass(pkt);
@@ -1082,3 +1244,25 @@ uint16_t kr_rrsig_type_covered(const knot_rdata_t *rdata)
 	return knot_rrsig_type_covered(rdata);
 }
 
+time_t kr_file_mtime (const char* fname) {
+	struct stat fstat;
+
+	if (stat(fname, &fstat) != 0) {
+		return 0;
+	}
+
+	return fstat.st_mtime;
+}
+
+long long kr_fssize(const char *path)
+{
+	if (!path)
+		return kr_error(EINVAL);
+
+	struct statvfs buf;
+	if (statvfs(path, &buf) != 0)
+		return kr_error(errno);
+
+	return buf.f_frsize * buf.f_blocks;
+}
+
