@@ -1,17 +1,5 @@
 /*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 #pragma once
@@ -22,16 +10,18 @@
 #include <sys/socket.h>
 #include <sys/time.h>
 #include <netinet/in.h>
+#include <unistd.h>
 
 #include <gnutls/gnutls.h>
 #include <gnutls/crypto.h>
-#include <lua.h>
 
 #include <libknot/libknot.h>
 #include <libknot/packet/pkt.h>
 #include <libknot/rrset.h>
 #include <libknot/rrtype/rrsig.h>
+#include <uv.h>
 
+#include "kresconfig.h"
 #include "lib/generic/array.h"
 #include "lib/defines.h"
 
@@ -44,11 +34,14 @@ struct kr_request;
 
 /** @brief Callback for request events. */
 typedef void (*trace_callback_f)(struct kr_request *request);
-/** @brief Callback for request logging handler. */
-typedef void (*trace_log_f)(const struct kr_query *query, const char *source, const char *msg);
+/**
+ * @brief Callback for request logging handler.
+ * @param[in] msg Log message. Pointer is not valid after handler returns. */
+typedef void (*trace_log_f)(const struct kr_request *request, const char *msg);
 
-#define kr_log_info(...) do { printf(__VA_ARGS__); fflush(stdout); } while(0)
+#define kr_log_info printf
 #define kr_log_error(...) fprintf(stderr, ## __VA_ARGS__)
+#define kr_log_deprecate(...) fprintf(stderr, "deprecation WARNING: " __VA_ARGS__)
 
 /* Always export these, but override direct calls by macros conditionally. */
 /** Whether in --verbose mode.  Only use this for reading. */
@@ -57,30 +50,38 @@ KR_EXPORT extern bool kr_verbose_status;
 /** Set --verbose mode.  Not available if compiled with -DNOVERBOSELOG. */
 KR_EXPORT bool kr_verbose_set(bool status);
 
-/** Log a message if in --verbose mode. */
-KR_EXPORT KR_PRINTF(1)
-void kr_log_verbose(const char *fmt, ...);
-
-/** Utility for QRVERBOSE - use that instead. */
-KR_EXPORT KR_PRINTF(3)
-void kr_log_qverbose_impl(const struct kr_query *qry, const char *cls, const char *fmt, ...);
-
 /**
  * @brief Return true if the query has request log handler installed.
  */
-#define kr_log_trace_enabled(query) (__builtin_expect( \
-	(query) && (query)->request && (query)->request->trace_log, \
+#define kr_log_rtrace_enabled(req) (__builtin_expect( \
+	(req) && (req)->trace_log, \
+	false))
+
+#define kr_log_qtrace_enabled(qry) (__builtin_expect( \
+	(qry) && kr_log_rtrace_enabled(qry->request), \
 	false))
 
 /**
- * Log a message through the request log handler.
+ * Log a message through the request log handler or stdout.
+ * Caller is responsible for detecting verbose mode, use QRVERBOSE() macro.
+ * @param  qry_uid query ID to append to request ID, 0 means "no query"
+ * @param  indent level of indentation between [req.qry][source] and message
+ * @param  source message source
+ * @param  fmt message format
+ */
+KR_EXPORT KR_PRINTF(5)
+void kr_log_req(const struct kr_request * const req, uint32_t qry_uid,
+		const unsigned int indent, const char *source, const char *fmt, ...);
+
+/**
+ * Log a message through the request log handler or stdout.
+ * Caller is responsible for detecting verbose mode, use QRVERBOSE() macro.
  * @param  query current query
  * @param  source message source
  * @param  fmt message format
- * @return true if the message was logged
  */
 KR_EXPORT KR_PRINTF(3)
-bool kr_log_trace(const struct kr_query *query, const char *source, const char *fmt, ...);
+void kr_log_q(const struct kr_query *qry, const char *source, const char *fmt, ...);
 
 #ifdef NOVERBOSELOG
 /* Efficient compile-time disabling of verbose messages. */
@@ -90,8 +91,8 @@ bool kr_log_trace(const struct kr_query *query, const char *source, const char *
 
 /** Block run in --verbose mode; optimized when not run. */
 #define VERBOSE_STATUS __builtin_expect(kr_verbose_status, false)
-#define WITH_VERBOSE(query) if(__builtin_expect(kr_verbose_status || kr_log_trace_enabled(query), false))
-#define kr_log_verbose if(VERBOSE_STATUS) kr_log_verbose
+#define WITH_VERBOSE(query) if(__builtin_expect(kr_verbose_status || kr_log_qtrace_enabled(query), false))
+#define kr_log_verbose if(VERBOSE_STATUS) printf
 
 #define KR_DNAME_GET_STR(dname_str, dname) \
 	char dname_str[KR_DNAME_STR_MAXLEN]; \
@@ -109,6 +110,13 @@ bool kr_log_trace(const struct kr_query *query, const char *source, const char *
 #endif
 
 /** @cond Memory alloc routines */
+
+/** Readability: avoid const-casts in code. */
+static inline void free_const(const void *what)
+{
+	free((void *)what);
+}
+
 static inline void *mm_alloc(knot_mm_t *mm, size_t size)
 {
 	if (mm) return mm->alloc(mm->ctx, size);
@@ -120,7 +128,7 @@ static inline void mm_free(knot_mm_t *mm, const void *what)
 		if (mm->free)
 			mm->free((void *)what);
 	}
-	else free((void *)what);
+	else free_const(what);
 }
 
 /** Realloc implementation using memory context. */
@@ -129,6 +137,8 @@ void *mm_realloc(knot_mm_t *mm, void *what, size_t size, size_t prev_size);
 
 /** Trivial malloc() wrapper. */
 void *mm_malloc(void *ctx, size_t n);
+/** posix_memalign() wrapper. */
+void *mm_malloc_aligned(void *ctx, size_t n);
 
 /** Initialize mm with standard malloc+free. */
 static inline void mm_ctx_init(knot_mm_t *mm)
@@ -137,8 +147,29 @@ static inline void mm_ctx_init(knot_mm_t *mm)
 	mm->alloc = mm_malloc;
 	mm->free = free;
 }
+
+/** Initialize mm with malloc+free with higher alignment (a power of two). */
+static inline void mm_ctx_init_aligned(knot_mm_t *mm, size_t alignment)
+{
+	assert(__builtin_popcount(alignment) == 1);
+	mm->ctx = (uint8_t *)NULL + alignment; /*< roundabout to satisfy linters */
+	/* posix_memalign() doesn't allow alignment < sizeof(void*),
+	 * and there's no point in using it for small values anyway,
+	 * as plain malloc() guarantees at least max_align_t.
+	 * Nitpick: we might use that type when assuming C11. */
+	mm->alloc = alignment > sizeof(void*) ? mm_malloc_aligned : mm_malloc;
+	mm->free = free;
+}
+
 /* @endcond */
 
+/** A strcmp() variant directly usable for qsort() on an array of strings. */
+static inline int strcmp_p(const void *p1, const void *p2)
+{
+	return strcmp(*(char * const *)p1, *(char * const *)p2);
+}
+
+
 /** Return time difference in miliseconds.
   * @note based on the _BSD_SOURCE timersub() macro */
 static inline long time_diff(struct timeval *begin, struct timeval *end) {
@@ -153,10 +184,17 @@ static inline long time_diff(struct timeval *begin, struct timeval *end) {
     return res.tv_sec * 1000 + res.tv_usec / 1000;
 }
 
+/** Get current working directory with fallback value. */
+static inline void get_workdir(char *out, size_t len) {
+	if(getcwd(out, len) == NULL) {
+		static const char errprefix[] = "<invalid working directory>";
+		strncpy(out, errprefix, len);
+	}
+}
+
 /** @cond internal Array types */
 struct kr_context;
 
-typedef array_t(knot_rrset_t *) rr_array_t;
 struct ranked_rr_array_entry {
 	uint32_t qry_uid;
 	uint8_t rank; /**< enum kr_rank */
@@ -165,6 +203,8 @@ struct ranked_rr_array_entry {
 	bool yielded : 1;
 	bool to_wire : 1; /**< whether to be put into the answer */
 	bool expiring : 1; /**< low remaining TTL; see is_expiring; only used in cache ATM */
+	bool in_progress : 1; /**< build of RRset in progress, i.e. different format of RR data */
+	bool dont_cache : 1; /**< avoid caching; useful e.g. for generated data */
 	knot_rrset_t *rr;
 };
 typedef struct ranked_rr_array_entry ranked_rr_array_entry_t;
@@ -200,7 +240,7 @@ static inline uint64_t kr_rand_bytes(unsigned int size)
 	kr_rnd_buffered(data, size);
 	/* I would have liked to dump the random data into a size_t directly,
 	 * but that would work well only on little-endian machines,
-	 * so intsead I hope that the compiler will optimize this out.
+	 * so instead I hope that the compiler will optimize this out.
 	 * (Tested via reading assembly from usual gcc -O2 setup.)
 	 * Alternatively we could waste more rnd bytes, but that seemed worse. */
 	result = 0;
@@ -232,7 +272,7 @@ static inline bool kr_rand_coin(unsigned int nomin, unsigned int denomin)
 
 /** Memory reservation routine for knot_mm_t */
 KR_EXPORT
-int kr_memreserve(void *baton, char **mem, size_t elm_size, size_t want, size_t *have);
+int kr_memreserve(void *baton, void **mem, size_t elm_size, size_t want, size_t *have);
 
 /** @internal Fast packet reset. */
 KR_EXPORT
@@ -289,30 +329,56 @@ void kr_inaddr_set_port(struct sockaddr *addr, uint16_t port);
 KR_EXPORT
 int kr_inaddr_str(const struct sockaddr *addr, char *buf, size_t *buflen);
 
+/** Write string representation for given address as "<addr>#<port>".
+ * It's the same as kr_inaddr_str(), but the input address is input in native format
+ * like for inet_ntop() (4 or 16 bytes) and port must be separate parameter.  */
+KR_EXPORT
+int kr_ntop_str(int family, const void *src, uint16_t port, char *buf, size_t *buflen);
+
+/** @internal Create string representation addr#port.
+ *  @return pointer to static string
+ */
+static inline char *kr_straddr(const struct sockaddr *addr)
+{
+	assert(addr != NULL);
+	/* We are the sinle-threaded application */
+	static char str[INET6_ADDRSTRLEN + 1 + 5 + 1];
+	size_t len = sizeof(str);
+	int ret = kr_inaddr_str(addr, str, &len);
+	return ret != kr_ok() || len == 0 ? NULL : str;
+}
+
+
 /** Return address type for string. */
 KR_EXPORT KR_PURE
 int kr_straddr_family(const char *addr);
 /** Return address length in given family (struct in*_addr). */
 KR_EXPORT KR_CONST
 int kr_family_len(int family);
-/** Create a sockaddr* from string+port representation (also accepts IPv6 link-local). */
+
+/** Create a sockaddr* from string+port representation.
+ * Also accepts IPv6 link-local and AF_UNIX starting with "/" (ignoring port) */
 KR_EXPORT
-struct sockaddr * kr_straddr_socket(const char *addr, int port);
+struct sockaddr * kr_straddr_socket(const char *addr, int port, knot_mm_t *pool);
+
 /** Parse address and return subnet length (bits).
   * @warning 'dst' must be at least `sizeof(struct in6_addr)` long. */
 KR_EXPORT
 int kr_straddr_subnet(void *dst, const char *addr);
 
-/** Splits ip address specified as "addr@port" or "addr#port" into addr and port
-  * and performs validation.
-  * @note if #port part isn't present, then port will be set to 0.
-  *       buf and\or port can be set to NULL.
-  * @return kr_error(EINVAL) - addr part doesn't contains valid ip address or
-  *                            #port part is out-of-range (either < 0 either > UINT16_MAX)
-  *         kr_error(ENOSP)  - buflen is too small
-  */
+/** Splits ip address specified as "addr@port" or "addr#port" into addr and port.
+ * \param instr[in] zero-terminated input, e.g. "192.0.2.1#12345\0"
+ * \param ipaddr[out] working buffer for the port-less prefix of instr;
+ *                    length >= INET6_ADDRSTRLEN + 1.
+ * \param port[out] written in case it's specified in instr
+ * \return error code
+ * \note Typically you follow this by kr_straddr_socket().
+ * \note Only internet addresses are supported, i.e. no AF_UNIX sockets.
+ */
 KR_EXPORT
-int kr_straddr_split(const char *addr, char *buf, size_t buflen, uint16_t *port);
+int kr_straddr_split(const char *instr, char ipaddr[static restrict (INET6_ADDRSTRLEN + 1)],
+		     uint16_t *port);
+
 /** Formats ip address and port in "addr#port" format.
   * and performs validation.
   * @note Port always formatted as five-character string with leading zeros.
@@ -351,10 +417,19 @@ KR_EXPORT
 int kr_rrkey(char *key, uint16_t class, const knot_dname_t *owner,
 	     uint16_t type, uint16_t additional);
 
-/** @internal Add RRSet copy to ranked RR array. */
+/** Add RRSet copy to a ranked RR array.
+ *
+ * To convert to standard RRs inside, you need to call _finalize() afterwards,
+ * and the memory of rr->rrs.rdata has to remain until then.
+ *
+ * \return array index (>= 0) or error code (< 0)
+ */
 KR_EXPORT
 int kr_ranked_rrarray_add(ranked_rr_array_t *array, const knot_rrset_t *rr,
 			  uint8_t rank, bool to_wire, uint32_t qry_uid, knot_mm_t *pool);
+/** Finalize in_progress sets - all with matching qry_uid. */
+KR_EXPORT
+int kr_ranked_rrarray_finalize(ranked_rr_array_t *array, uint32_t qry_uid, knot_mm_t *pool);
 
 /** @internal Mark the RRSets from particular query as
  * "have (not) to be recorded in the final answer".
@@ -369,7 +444,11 @@ int kr_ranked_rrarray_set_wire(ranked_rr_array_t *array, bool to_wire,
 			       uint32_t qry_uid, bool check_dups,
 			       bool (*extraCheck)(const ranked_rr_array_entry_t *));
 
-KR_PURE
+/**
+ * @return Newly allocated string representation of packet.
+ * Caller has to free() returned string.
+ */
+KR_EXPORT
 char *kr_pkt_text(const knot_pkt_t *pkt);
 
 KR_PURE
@@ -411,30 +490,6 @@ static inline uint16_t kr_rrset_type_maysig(const knot_rrset_t *rr)
 	return type;
 }
 
-/** Printf onto the lua stack, avoiding additional copy (thin wrapper). */
-KR_PRINTF(2)
-static inline const char *lua_push_printf(lua_State *L, const char *fmt, ...)
-{
-	va_list args;
-	va_start(args, fmt);
-	const char *ret = lua_pushvfstring(L, fmt, args);
-	va_end(args);
-	return ret;
-}
-
-/** @internal Return string representation of addr.
- *  @note return pointer to static string
- */
-static inline char *kr_straddr(const struct sockaddr *addr)
-{
-	assert(addr != NULL);
-	/* We are the sinle-threaded application */
-	static char str[INET6_ADDRSTRLEN + 1 + 5 + 1];
-	size_t len = sizeof(str);
-	int ret = kr_inaddr_str(addr, str, &len);
-	return ret != kr_ok() || len == 0 ? NULL : str;
-}
-
 /** The current time in monotonic milliseconds.
  *
  * \note it may be outdated in case of long callbacks; see uv_now().
@@ -442,6 +497,9 @@ static inline char *kr_straddr(const struct sockaddr *addr)
 KR_EXPORT
 uint64_t kr_now();
 
+/** Call free(handle->data); it's useful e.g. as a callback in uv_close(). */
+KR_EXPORT void kr_uv_free_cb(uv_handle_t* handle);
+
 /** Convert name from lookup format to wire.  See knot_dname_lf
  *
  * \note len bytes are read and len+1 are written with *normal* LF,
@@ -488,8 +546,14 @@ const char *kr_strptime_diff(const char *format, const char *time1_str,
 /* Trivial non-inline wrappers, to be used in lua. */
 KR_EXPORT void kr_rrset_init(knot_rrset_t *rrset, knot_dname_t *owner,
 				uint16_t type, uint16_t rclass, uint32_t ttl);
+KR_EXPORT uint16_t kr_pkt_has_dnssec(const knot_pkt_t *pkt);
 KR_EXPORT uint16_t kr_pkt_qclass(const knot_pkt_t *pkt);
 KR_EXPORT uint16_t kr_pkt_qtype(const knot_pkt_t *pkt);
 KR_EXPORT uint32_t kr_rrsig_sig_inception(const knot_rdata_t *rdata);
 KR_EXPORT uint32_t kr_rrsig_sig_expiration(const knot_rdata_t *rdata);
 KR_EXPORT uint16_t kr_rrsig_type_covered(const knot_rdata_t *rdata);
+
+KR_EXPORT time_t kr_file_mtime (const char* fname);
+/** Return filesystem size in bytes. */
+KR_EXPORT long long kr_fssize(const char *path);
+
