@@ -1,5 +1,5 @@
-/* Copyright (C) Knot Resolver contributors. Licensed under GNU GPLv3 or
- * (at your option) any later version. See COPYING for text of the license.
+/* Copyright (C) Knot Resolver contributors
+ * SPDX-License-Identifier: GPL-3.0-or-later
  *
  * This module logs (query name, type) pairs which failed DNSSEC validation. */
 
@@ -12,11 +12,6 @@
 #include "lib/layer.h"
 #include "lib/generic/lru.h"
 
-/** @internal Compatibility wrapper for Lua < 5.2 */
-#if LUA_VERSION_NUM < 502
-#define lua_rawlen(L, obj) lua_objlen((L), (obj))
-#endif
-
 #ifdef LRU_REP_SIZE
  #define FREQUENT_COUNT LRU_REP_SIZE /* Size of frequent tables */
 #else
@@ -105,6 +100,18 @@ static char* dump_frequent(void *env, struct kr_module *module, const char *args
 KR_EXPORT
 int bogus_log_init(struct kr_module *module)
 {
+	static kr_layer_api_t layer = {
+		.consume = &consume,
+	};
+	layer.data = module;
+	module->layer = &layer;
+
+	static const struct kr_prop props[] = {
+		{ &dump_frequent, "frequent", "List most frequent queries.", },
+		{ NULL, NULL, NULL }
+	};
+	module->props = props;
+
 	struct stat_data *data = malloc(sizeof(*data));
 	if (!data) {
 		return kr_error(ENOMEM);
@@ -126,25 +133,4 @@ int bogus_log_deinit(struct kr_module *module)
 	return kr_ok();
 }
 
-
-KR_EXPORT
-const kr_layer_api_t *bogus_log_layer(struct kr_module *module)
-{
-	static kr_layer_api_t _layer = {
-		.consume = &consume,
-	};
-	_layer.data = module;
-	return &_layer;
-}
-
-KR_EXPORT
-struct kr_prop *bogus_log_props(void)
-{
-	static struct kr_prop prop_list[] = {
-		{ &dump_frequent, "frequent", "List most frequent queries.", },
-		{ NULL, NULL, NULL }
-	};
-	return prop_list;
-}
-
 KR_MODULE_EXPORT(bogus_log)
