@@ -1,17 +1,5 @@
 /*  Copyright (C) 2016-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 #include <assert.h>
@@ -423,7 +411,7 @@ int check_request(kr_layer_t *ctx)
 		/* Request has no server cookie. */
 		return_state = invalid_sc_status(return_state, false,
 		                                 ignore_badcookie, req, answer);
-		if (return_state == KR_STATE_FAIL) {
+		if (return_state & KR_STATE_FAIL) {
 			return return_state;
 		}
 		goto answer_add_cookies;
@@ -449,7 +437,7 @@ int check_request(kr_layer_t *ctx)
 		/* Invalid server cookie. */
 		return_state = invalid_sc_status(return_state, true,
 		                                 ignore_badcookie, req, answer);
-		if (return_state == KR_STATE_FAIL) {
+		if (return_state & KR_STATE_FAIL) {
 			return return_state;
 		}
 		goto answer_add_cookies;
