@@ -1,16 +1,5 @@
 /*
- * This program is free software: you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation, either version 3 of the License, or
- * (at your option) any later version.
-
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
-
- * You should have received a copy of the GNU General Public License
- * along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ * SPDX-License-Identifier: GPL-3.0-or-later
  *
  * @file dnstap.c
  * @brief dnstap based query logging support
@@ -211,6 +200,13 @@ static int dnstap_log(kr_layer_t *ctx) {
 
 KR_EXPORT
 int dnstap_init(struct kr_module *module) {
+	static kr_layer_api_t layer = {
+		.finish = &dnstap_log,
+	};
+	/* Store module reference */
+	layer.data = module;
+	module->layer = &layer;
+
 	/* allocated memory for internal data */
 	struct dnstap_data *data = malloc(sizeof(*data));
 	if (!data) {
@@ -368,15 +364,5 @@ int dnstap_config(struct kr_module *module, const char *conf) {
 	return kr_ok();
 }
 
-KR_EXPORT
-const kr_layer_api_t *dnstap_layer(struct kr_module *module) {
-	static kr_layer_api_t _layer = {
-		.finish = &dnstap_log,
-	};
-	/* Store module reference */
-	_layer.data = module;
-	return &_layer;
-}
-
 KR_MODULE_EXPORT(dnstap)
 
