@@ -1,17 +1,5 @@
 /*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 /**
@@ -36,14 +24,16 @@ static int edns_keepalive_finalize(kr_layer_t *ctx)
 	const bool ka_want =
 		req->qsource.flags.tcp &&
 		src_opt != NULL &&
-		knot_edns_get_option(src_opt, KNOT_EDNS_OPTION_TCP_KEEPALIVE) &&
+		knot_edns_get_option(src_opt, KNOT_EDNS_OPTION_TCP_KEEPALIVE
+					#if KNOT_VERSION_HEX >= 0x020900
+						, NULL
+					#endif
+				) &&
 		answ_opt != NULL;
 	if (!ka_want) {
 		return ctx->state;
 	}
-	const struct worker_ctx *worker = (const struct worker_ctx *)req->daemon_context;
-	assert(worker);
-	const struct network *net = &worker->engine->net;
+	const struct network *net = &the_worker->engine->net;
 	uint64_t timeout = net->tcp.in_idle_timeout / 100;
 	if (timeout > UINT16_MAX) {
 		timeout = UINT16_MAX;
@@ -62,15 +52,13 @@ static int edns_keepalive_finalize(kr_layer_t *ctx)
 	return ctx->state;
 }
 
-KR_EXPORT
-const kr_layer_api_t *edns_keepalive_layer(struct kr_module *module)
+KR_EXPORT int edns_keeapalive_init(struct kr_module *self)
 {
-	static kr_layer_api_t _layer = {
+	static const kr_layer_api_t layer = {
 		.answer_finalize = &edns_keepalive_finalize,
 	};
-	/* Store module reference */
-	_layer.data = module;
-	return &_layer;
+	self->layer = &layer;
+	return kr_ok();
 }
 
 KR_MODULE_EXPORT(edns_keepalive)
