@@ -1,17 +1,5 @@
 /*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 /**
@@ -33,6 +21,9 @@
 #include "lib/module.h"
 #include "lib/layer.h"
 
+#include <inttypes.h>
+#include <math.h>
+
 /* Defaults */
 #define VERBOSE_MSG(qry, ...) QRVERBOSE(qry, "hint",  __VA_ARGS__)
 #define ERR_MSG(...) kr_log_error("[     ][hint] " __VA_ARGS__)
@@ -41,7 +32,9 @@ struct hints_data {
 	struct kr_zonecut hints;
 	struct kr_zonecut reverse_hints;
 	bool use_nodata; /**< See hint_use_nodata() description, exposed via lua. */
+	uint32_t ttl;    /**< TTL used for the hints, exposed via lua. */
 };
+static const uint32_t HINTS_TTL_DEFAULT = 5;
 
 /** Useful for returning from module properties. */
 static char * bool2jsonstr(bool val)
@@ -79,16 +72,17 @@ static int put_answer(knot_pkt_t *pkt, struct kr_query *qry, knot_rrset_t *rr, b
 	return ret;
 }
 
-static int satisfy_reverse(struct kr_zonecut *hints, knot_pkt_t *pkt, struct kr_query *qry, bool use_nodata)
+static int satisfy_reverse(/*const*/ struct hints_data *data,
+			   knot_pkt_t *pkt, struct kr_query *qry)
 {
 	/* Find a matching name */
-	pack_t *addr_set = kr_zonecut_find(hints, qry->sname);
+	pack_t *addr_set = kr_zonecut_find(&data->reverse_hints, qry->sname);
 	if (!addr_set || addr_set->len == 0) {
 		return kr_error(ENOENT);
 	}
 	knot_dname_t *qname = knot_dname_copy(qry->sname, &pkt->mm);
 	knot_rrset_t rr;
-	knot_rrset_init(&rr, qname, KNOT_RRTYPE_PTR, KNOT_CLASS_IN, 0);
+	knot_rrset_init(&rr, qname, KNOT_RRTYPE_PTR, KNOT_CLASS_IN, data->ttl);
 
 	/* Append address records from hints */
 	uint8_t *addr = pack_last(*addr_set);
@@ -98,19 +92,20 @@ static int satisfy_reverse(struct kr_zonecut *hints, knot_pkt_t *pkt, struct kr_
 		knot_rrset_add_rdata(&rr, addr_val, len, &pkt->mm);
 	}
 
-	return put_answer(pkt, qry, &rr, use_nodata);
+	return put_answer(pkt, qry, &rr, data->use_nodata);
 }
 
-static int satisfy_forward(struct kr_zonecut *hints, knot_pkt_t *pkt, struct kr_query *qry, bool use_nodata)
+static int satisfy_forward(/*const*/ struct hints_data *data,
+			   knot_pkt_t *pkt, struct kr_query *qry)
 {
 	/* Find a matching name */
-	pack_t *addr_set = kr_zonecut_find(hints, qry->sname);
+	pack_t *addr_set = kr_zonecut_find(&data->hints, qry->sname);
 	if (!addr_set || addr_set->len == 0) {
 		return kr_error(ENOENT);
 	}
 	knot_dname_t *qname = knot_dname_copy(qry->sname, &pkt->mm);
 	knot_rrset_t rr;
-	knot_rrset_init(&rr, qname, qry->stype, qry->sclass, 0);
+	knot_rrset_init(&rr, qname, qry->stype, qry->sclass, data->ttl);
 	size_t family_len = sizeof(struct in_addr);
 	if (rr.type == KNOT_RRTYPE_AAAA) {
 		family_len = sizeof(struct in6_addr);
@@ -127,13 +122,13 @@ static int satisfy_forward(struct kr_zonecut *hints, knot_pkt_t *pkt, struct kr_
 		addr = pack_obj_next(addr);
 	}
 
-	return put_answer(pkt, qry, &rr, use_nodata);
+	return put_answer(pkt, qry, &rr, data->use_nodata);
 }
 
 static int query(kr_layer_t *ctx, knot_pkt_t *pkt)
 {
 	struct kr_query *qry = ctx->req->current_query;
-	if (!qry || ctx->state & (KR_STATE_FAIL)) {
+	if (!qry || (ctx->state & KR_STATE_FAIL)) {
 		return ctx->state;
 	}
 
@@ -147,11 +142,11 @@ static int query(kr_layer_t *ctx, knot_pkt_t *pkt)
 	switch(qry->stype) {
 	case KNOT_RRTYPE_A:
 	case KNOT_RRTYPE_AAAA: /* Find forward record hints */
-		if (satisfy_forward(&data->hints, pkt, qry, data->use_nodata) != 0)
+		if (satisfy_forward(data, pkt, qry) != 0)
 			return ctx->state;
 		break;
 	case KNOT_RRTYPE_PTR: /* Find PTR record */
-		if (satisfy_reverse(&data->reverse_hints, pkt, qry, data->use_nodata) != 0)
+		if (satisfy_reverse(data, pkt, qry) != 0)
 			return ctx->state;
 		break;
 	default:
@@ -570,26 +565,53 @@ static char* hint_use_nodata(void *env, struct kr_module *module, const char *ar
 	return bool2jsonstr(true);
 }
 
-/*
- * Module implementation.
- */
-
-KR_EXPORT
-const kr_layer_api_t *hints_layer(struct kr_module *module)
+static char* hint_ttl(void *env, struct kr_module *module, const char *args)
 {
-	static kr_layer_api_t _layer = {
-		.produce = &query,
-	};
-	/* Store module reference */
-	_layer.data = module;
-	return &_layer;
-}
+	struct hints_data *data = module->data;
 
+	/* Do no change on nonsense TTL values (incl. suspicious floats). */
+	JsonNode *root_node = args ? json_decode(args) : NULL;
+	if (root_node && root_node->tag == JSON_NUMBER) {
+		double ttl_d = root_node->number_;
+		uint32_t ttl = (uint32_t)round(ttl_d);
+		if (ttl_d >= 0 && fabs(ttl_d - ttl) * 64 < 1) {
+			data->ttl = ttl;
+		}
+	}
+	json_delete(root_node);
+
+	/* Always return the current TTL setting.  Plain number is valid JSON. */
+	char *result = NULL;
+	if (-1 == asprintf(&result, "%"PRIu32, data->ttl)) {
+		result = NULL;
+	}
+	return result;
+}
 
 /** Basic initialization: get a memory pool, etc. */
 KR_EXPORT
 int hints_init(struct kr_module *module)
 {
+	static kr_layer_api_t layer = {
+		.produce = &query,
+	};
+	/* Store module reference */
+	layer.data = module;
+	module->layer = &layer;
+
+	static const struct kr_prop props[] = {
+	    { &hint_set,    "set", "Set {name, address} hint.", },
+	    { &hint_del,    "del", "Delete one {name, address} hint or all addresses for the name.", },
+	    { &hint_get,    "get", "Retrieve hint for given name.", },
+	    { &hint_ttl,    "ttl", "Set/get TTL used for the hints.", },
+	    { &hint_add_hosts, "add_hosts", "Load a file with hosts-like formatting and add contents into hints.", },
+	    { &hint_root,   "root", "Replace root hints set (empty value to return current list).", },
+	    { &hint_root_file, "root_file", "Replace root hints set from a zonefile.", },
+	    { &hint_use_nodata, "use_nodata", "Synthesise NODATA if name matches, but type doesn't.  True by default.", },
+	    { NULL, NULL, NULL }
+	};
+	module->props = props;
+
 	/* Create pool and copy itself */
 	knot_mm_t _pool = {
 		.ctx = mp_new(4096),
@@ -609,6 +631,7 @@ int hints_init(struct kr_module *module)
 	kr_zonecut_init(&data->hints, (const uint8_t *)(""), pool);
 	kr_zonecut_init(&data->reverse_hints, (const uint8_t *)(""), pool);
 	data->use_nodata = true;
+	data->ttl = HINTS_TTL_DEFAULT;
 	module->data = data;
 
 	return kr_ok();
@@ -647,22 +670,6 @@ int hints_config(struct kr_module *module, const char *conf)
 	return kr_ok();
 }
 
-KR_EXPORT
-struct kr_prop *hints_props(void)
-{
-	static struct kr_prop prop_list[] = {
-	    { &hint_set,    "set", "Set {name, address} hint.", },
-	    { &hint_del,    "del", "Delete one {name, address} hint or all addresses for the name.", },
-	    { &hint_get,    "get", "Retrieve hint for given name.", },
-	    { &hint_add_hosts, "add_hosts", "Load a file with hosts-like formatting and add contents into hints.", },
-	    { &hint_root,   "root", "Replace root hints set (empty value to return current list).", },
-	    { &hint_root_file, "root_file", "Replace root hints set from a zonefile.", },
-	    { &hint_use_nodata, "use_nodata", "Synthesise NODATA if name matches, but type doesn't.  True by default.", },
-	    { NULL, NULL, NULL }
-	};
-	return prop_list;
-}
-
 KR_MODULE_EXPORT(hints)
 
 #undef VERBOSE_MSG
