@@ -1,3 +1,4 @@
+-- SPDX-License-Identifier: GPL-3.0-or-later
 local utils = require('test_utils')
 
 -- setup resolver
@@ -16,7 +17,7 @@ end
 -- test loading from config file
 local function test_custom()
 	-- load custom root hints file with fake ip address for a.root-server.net
-	local err_msg = hints.root_file(TEST_DIR .. 'hints_test.zone')
+	local err_msg = hints.root_file('hints_test.zone')
 	same(err_msg, '', 'load root hints from file')
 
 	-- get loaded root hints and change names to lowercase
