@@ -1,3 +1,4 @@
+-- SPDX-License-Identifier: GPL-3.0-or-later
 local kres = require('kres')
 local ffi = require('ffi')
 
@@ -12,45 +13,41 @@ local function getruleid()
 end
 
 -- Support for client sockets from inside policy actions
-local socket_client = function () return error("missing luasocket, can't create socket client") end
-local has_socket, socket = pcall(require, 'socket')
+local socket_client = function ()
+	return error("missing lua-cqueues library, can't create socket client")
+end
+local has_socket, socket = pcall(require, 'cqueues.socket')
 if has_socket then
 	socket_client = function (host, port)
 		local s, err, status
-		if host:find(':') then
-			s, err = socket.udp6()
-		else
-			s, err = socket.udp()
-		end
-		if not s then
-			return nil, err
-		end
-		status, err = s:setpeername(host, port)
+
+		s = socket.connect({ host = host, port = port, type = socket.SOCK_DGRAM })
+		s:setmode('bn', 'bn')
+		status, err = pcall(s.connect, s)
+
 		if not status then
-			return nil, err
+			return status, err
 		end
 		return s
 	end
 end
 
+-- Split address and port from a combined string.
 local function addr_split_port(target, default_port)
-	assert(default_port)
-	assert(type(default_port) == 'number')
-	local addr, port = target:match '([^@]*)@?(.*)'
-	local nport
-	if port ~= "" then
-		nport = tonumber(port)
-		if not nport or nport < 1 or nport > 65535 then
-			error('port "'.. port  ..'" is not valid')
-		end
+	assert(default_port and type(default_port) == 'number')
+	local port = ffi.new('uint16_t[1]', default_port)
+	local addr = ffi.new('char[47]') -- INET6_ADDRSTRLEN + 1
+	local ret = ffi.C.kr_straddr_split(target, addr, port)
+	if ret ~= 0 then
+		error('failed to parse address ' .. target)
 	end
-	return addr, nport or default_port
+	return addr, tonumber(port[0])
 end
 
 -- String address@port -> sockaddr.
 local function addr2sock(target, default_port)
 	local addr, port = addr_split_port(target, default_port)
-	local sock = ffi.gc(ffi.C.kr_straddr_socket(addr, port), ffi.C.free);
+	local sock = ffi.gc(ffi.C.kr_straddr_socket(addr, port, nil), ffi.C.free);
 	if sock == nil then
 		error("target '"..target..'" is not a valid IP address')
 	end
@@ -67,13 +64,13 @@ end
 -- Mirror request elsewhere, and continue solving
 function policy.MIRROR(target)
 	local addr, port = addr_split_port(target, 53)
-	local sink, err = socket_client(addr, port)
+	local sink, err = socket_client(ffi.string(addr), port)
 	if not sink then panic('MIRROR target %s is not a valid: %s', target, err) end
 	return function(state, req)
 		if state == kres.FAIL then return state end
 		local query = req.qsource.packet
 		if query ~= nil then
-			sink:send(ffi.string(query.wire, query.size))
+			sink:send(ffi.string(query.wire, query.size), 1, tonumber(query.size))
 		end
 		return -- Chain action to next
 	end
@@ -144,114 +141,35 @@ function policy.FORWARD(target)
 	end
 end
 
--- object must be non-empty string or non-empty table of non-empty strings
-local function is_nonempty_string_or_table(object)
-	if type(object) == 'string' then
-		return #object ~= 0
-	elseif type(object) ~= 'table' or not next(object) then
-		return false
-	end
-	for _, val in pairs(object) do
-		if type(val) ~= 'string' or #val == 0 then
-			return false
-		end
-	end
-	return true
-end
-
-local function insert_from_string_or_table(source, destination)
-	if type(source) == 'table' then
-		for _, v in pairs(source) do
-			table.insert(destination, v)
-		end
-	else
-		table.insert(destination, source)
-	end
-end
-
--- Check for allowed authentication types and return type for the current target
-local function tls_forward_target_authtype(idx, target)
-	if (target.pin_sha256 and not (target.ca_file or target.hostname or target.insecure)) then
-		if not is_nonempty_string_or_table(target.pin_sha256) then
-			error('TLS_FORWARD target authentication is invalid at position '
-			      .. idx .. '; pin_sha256 must be string or list of strings')
-		end
-		return 'pin_sha256'
-	elseif (target.insecure and not (target.ca_file or target.hostname or target.pin_sha256)) then
-		return 'insecure'
-	elseif (target.hostname and not (target.insecure or target.pin_sha256)) then
-		if not (is_nonempty_string_or_table(target.hostname)) then
-			error('TLS_FORWARD target authentication is invalid at position '
-			      .. idx .. '; hostname must be string or list of strings')
-		end
-		-- if target.ca_file is empty, system CA will be used
-		return 'cert'
-	else
-		error('TLS_FORWARD authentication options at position ' .. idx
-		      .. ' are invalid; specify one of: pin_sha256 / hostname [+ca_file] / insecure')
-	end
-end
-
-local function tls_forward_target_check_syntax(idx, list_entry)
-	if type(list_entry) ~= 'table' then
-		error('TLS_FORWARD target must be a non-empty table (found '
-		      .. type(list_entry) .. ' at position ' .. idx .. ')')
-	end
-	if type(list_entry[1]) ~= 'string' then
-		error('TLS_FORWARD target must start with an IP address (found '
-		      .. type(list_entry[1]) .. ' at the beginning of target position ' .. idx .. ')')
-	end
-end
-
 -- Forward request and all subrequests to upstream over TLS; validate answers
-function policy.TLS_FORWARD(target)
-	local sockaddr_c_list = {}
-	local sockaddr_config = {}  -- items: { string_addr=<addr string>, auth_type=<auth type> }
-	local ca_files = {}
-	local hostnames = {}
-	local pins = {}
-	if type(target) ~= 'table' or #target < 1 then
+function policy.TLS_FORWARD(targets)
+	if type(targets) ~= 'table' or #targets < 1 then
 		error('TLS_FORWARD argument must be a non-empty table')
-	end
-	for idx, upstream_list_entry in pairs(target) do
-		tls_forward_target_check_syntax(idx, upstream_list_entry)
-		local auth_type = tls_forward_target_authtype(idx, upstream_list_entry)
-		local string_addr = upstream_list_entry[1]
-		local sockaddr_c = addr2sock(string_addr, 853)
-		local sockaddr_lua = ffi.string(sockaddr_c, ffi.C.kr_sockaddr_len(sockaddr_c))
-		if sockaddr_config[sockaddr_lua] then
-			error('TLS_FORWARD configuration cannot declare two configs for IP address ' .. string_addr)
-		end
-		table.insert(sockaddr_c_list, sockaddr_c)
-		sockaddr_config[sockaddr_lua] = {string_addr=string_addr, auth_type=auth_type}
-		if auth_type == 'cert' then
-			ca_files[sockaddr_lua] = {}
-			hostnames[sockaddr_lua] = {}
-			insert_from_string_or_table(upstream_list_entry.ca_file, ca_files[sockaddr_lua])
-			insert_from_string_or_table(upstream_list_entry.hostname, hostnames[sockaddr_lua])
-		elseif auth_type == 'pin_sha256' then
-			pins[sockaddr_lua] = {}
-			insert_from_string_or_table(upstream_list_entry.pin_sha256, pins[sockaddr_lua])
-		elseif auth_type ~= 'insecure' then
-			-- insecure does nothing, user does not want authentication
-			assert(false, 'unsupported auth_type')
-		end
+	elseif #targets > 4 then
+		error('TLS_FORWARD supports at most four targets (in a single call)')
 	end
 
-	-- Update the global table of authentication data only if all checks above passed
-	for sockaddr_lua, config in pairs(sockaddr_config) do
-		assert(#config.string_addr > 0)
-		if config.auth_type == 'insecure' then
-			net.tls_client(config.string_addr)
-		elseif config.auth_type == 'pin_sha256' then
-			assert(#pins[sockaddr_lua] > 0)
-			net.tls_client(config.string_addr, pins[sockaddr_lua])
-		elseif config.auth_type == 'cert' then
-			assert(#hostnames[sockaddr_lua] > 0)
-			net.tls_client(config.string_addr, ca_files[sockaddr_lua], hostnames[sockaddr_lua])
-		else
-			assert(false, 'unsupported auth_type')
+	local sockaddr_c_set = {}
+	local nslist = {} -- to persist in closure of the returned function
+	for idx, target in pairs(targets) do
+		if type(target) ~= 'table' or type(target[1]) ~= 'string' then
+			error(string.format('TLS_FORWARD configuration at position ' ..
+			'%d must be a table starting with an IP address', idx))
 		end
+		-- Note: some functions have checks with error() calls inside.
+		local sockaddr_c = addr2sock(target[1], 853)
+
+		-- Refuse repeated addresses in the same set.
+		local sockaddr_lua = ffi.string(sockaddr_c, ffi.C.kr_sockaddr_len(sockaddr_c))
+		if sockaddr_c_set[sockaddr_lua] then
+			error('TLS_FORWARD configuration cannot declare two configs for IP address '
+					.. target[1])
+		else
+			sockaddr_c_set[sockaddr_lua] = true;
+		end
+
+		table.insert(nslist, sockaddr_c)
+		net.tls_client(target)
 	end
 
 	return function(state, req)
@@ -264,7 +182,7 @@ function policy.TLS_FORWARD(target)
 		qry.flags.AWAIT_CUT = true
 		req.options.TCP = true
 		qry.flags.TCP = true
-		set_nslist(qry, sockaddr_c_list)
+		set_nslist(qry, nslist)
 		return state
 	end
 end
@@ -272,7 +190,7 @@ end
 -- Rewrite records in packet
 function policy.REROUTE(tbl, names)
 	-- Import renumbering rules
-	local ren = require('renumber')
+	local ren = require('kres_modules.renumber')
 	local prefixes = {}
 	for from, to in pairs(tbl) do
 		table.insert(prefixes, names and ren.name(from, to) or ren.prefix(from, to))
@@ -291,6 +209,32 @@ function policy.FLAGS(opts_set, opts_clear)
 	end
 end
 
+-- Create answer with passed arguments
+function policy.ANSWER(rtable, nodata)
+	return function(_, req)
+		local qry = req:current()
+		local answer = req.answer
+		local data = rtable[qry.stype]
+
+		ffi.C.kr_pkt_make_auth_header(answer)
+
+		if data == nil then
+			if nodata == true then
+				answer:rcode(kres.rcode.NOERROR)
+				return kres.DONE
+			end
+		else
+			local ttl = data.ttl or 1
+
+			answer:rcode(kres.rcode.NOERROR)
+			answer:begin(kres.section.ANSWER)
+			answer:put(qry.sname, ttl, qry.sclass, qry.stype, data.rdata)
+
+			return kres.DONE
+		end
+	end
+end
+
 local function mkauth_soa(answer, dname, mname)
 	if mname == nil then
 		mname = dname
@@ -433,34 +377,143 @@ end
 
 local function rpz_parse(action, path)
 	local rules = {}
+	local new_actions = {}
+	local origin = '\0'
 	local action_map = {
 		-- RPZ Policy Actions
 		['\0'] = action,
-		['\1*\0'] = action, -- deviates from RPZ spec
+		['\1*\0'] = policy.ANSWER({}, true),
 		['\012rpz-passthru\0'] = policy.PASS, -- the grammar...
 		['\008rpz-drop\0'] = policy.DROP,
 		['\012rpz-tcp-only\0'] = policy.TC,
 		-- Policy triggers @NYI@
 	}
-	local parser = require('zonefile').new()
-	if not parser:open(path) then error(string.format('failed to parse "%s"', path)) end
-	while parser:parse() do
-		local name = ffi.string(parser.r_owner, parser.r_owner_length)
-		local name_action = ffi.string(parser.r_data, parser.r_data_length)
-		rules[name] = action_map[name_action]
-		-- Warn when NYI
-		if #name > 1 and not action_map[name_action] then
-			print(string.format('[ rpz ] %s:%d: unsupported policy action', path, tonumber(parser.line_counter)))
+	local unsupp_rrs = function (rtype)
+		local set = {
+			kres.type.DNAME,
+			kres.type.NS,
+			kres.type.SOA,
+			kres.type.DNSKEY,
+			kres.type.DS,
+			kres.type.RRSIG,
+			kres.type.NSEC,
+			kres.type.NSEC3,
+		}
+		for _, l in pairs(set) do
+			if rtype == l then return true end
 		end
+		return false
+	end
+	local parser = require('zonefile').new()
+	local ok, errstr = parser:open(path)
+	if not ok then
+		error(string.format('failed to parse "%s": %s', path, errstr or "unknown error"))
+	end
+	while true do
+		ok, errstr = parser:parse()
+		if errstr then
+			log('[poli] RPZ %s:%d: %s', path, tonumber(parser.line_counter), errstr)
+		end
+		if not ok then break end
+
+		local full_name = ffi.gc(ffi.C.knot_dname_copy(parser.r_owner, nil), ffi.C.free)
+		local rdata = ffi.string(parser.r_data, parser.r_data_length)
+		ffi.C.knot_dname_to_lower(full_name)
+
+		if (parser.r_type == kres.type.SOA) then
+			origin = ffi.gc(ffi.C.knot_dname_copy(full_name, nil), ffi.C.free)
+			goto continue
+		end
+
+		local prefix_labels = ffi.C.knot_dname_in_bailiwick(full_name, origin)
+		local name
+		if prefix_labels > 0 then
+			local bytes = 0
+			for _=1,prefix_labels do
+				bytes = bytes + 1 + full_name[bytes]
+			end
+			name = ffi.string(full_name, bytes)
+			name = name..'\0'
+		else
+			name = ffi.string(full_name, parser.r_owner_length)
+		end
+
+		if parser.r_type == kres.type.CNAME then
+			if action_map[rdata] then
+				rules[name] = action_map[rdata]
+			else
+				log('[poli] RPZ %s:%d: CNAME with custom target in RPZ is not supported', path, tonumber(parser.line_counter))
+			end
+		else
+			-- Warn when NYI
+			if #name then
+				if unsupp_rrs(parser.r_type) then
+					log('[poli] RPZ %s:%d: RR type %s is not allowed in RPZ', path, tonumber(parser.line_counter),
+					    kres.tostring.type[parser.r_type])
+				else
+					if new_actions[name] == nil then new_actions[name] = {} end
+					new_actions[name][parser.r_type] = { ttl=parser.r_ttl, rdata=rdata }
+				end
+			end
+		end
+
+		::continue::
+	end
+	collectgarbage()
+	for qname, rrsets in pairs(new_actions) do
+		rules[qname] = policy.ANSWER(rrsets, true)
 	end
 	return rules
 end
 
+-- Split path into dirname and basename (like the shell utilities)
+local function get_dir_and_file(path)
+	local dir, file = string.match(path, "(.*)/([^/]+)")
+
+	-- If regex doesn't match then path must be the file directly (i.e. doesn't contain '/')
+	-- This assumes that the file exists (rpz_parse() would fail if it doesn't)
+	if not dir and not file then
+		dir = '.'
+		file = path
+	end
+
+	return dir, file
+end
+
 -- RPZ policy set
--- Create RPZ from zone file
-function policy.rpz(action, path)
+-- Create RPZ from zone file and optionally watch the file for changes
+function policy.rpz(action, path, watch)
 	local rules = rpz_parse(action, path)
-	collectgarbage()
+
+	if watch ~= false then
+		local has_notify, notify  = pcall(require, 'cqueues.notify')
+		if has_notify then
+			local bit = require('bit')
+
+			local dir, file = get_dir_and_file(path)
+			local watcher = notify.opendir(dir)
+			watcher:add(file, bit.bxor(notify.CREATE, notify.MODIFY))
+
+			worker.coroutine(function ()
+				for _, name in watcher:changes() do
+					-- Limit to changes on file we're interested in
+					-- Watcher will also fire for changes to the directory itself
+					if name == file then
+						-- If the file changes then reparse and replace the existing ruleset
+						if verbose() then
+							log('[poli] RPZ reloading: ' .. name)
+						end
+						rules = rpz_parse(action, path)
+					end
+				end
+			end)
+		elseif watch then -- explicitly requested and failed
+			error('[poli] lua-cqueues required to watch and reload RPZ file')
+		elseif verbose() then
+			log('[poli] lua-cqueues required to watch and reload RPZ file, continuing without watching')
+		end
+	end
+
 	return function(_, query)
 		local label = query:name()
 		local rule = rules[label]
@@ -472,6 +525,84 @@ function policy.rpz(action, path)
 	end
 end
 
+-- Apply an action when query belongs to a slice (determined by slice_func())
+function policy.slice(slice_func, ...)
+	local actions = {...}
+	if #actions <= 0 then
+		error('[poli] at least one action must be provided to policy.slice()')
+	end
+
+	return function(_, query)
+		local index = slice_func(query, #actions)
+		return actions[index]
+	end
+end
+
+-- Initializes slicing function that randomly assigns queries to a slice based on their registrable domain
+function policy.slice_randomize_psl(seed)
+	local has_psl, psl_lib = pcall(require, 'psl')
+	if not has_psl then
+		error('[poli] lua-psl is required for policy.slice_randomize_psl()')
+	end
+	-- load psl
+	local has_latest, psl = pcall(psl_lib.latest)
+	if not has_latest then -- compatiblity with lua-psl < 0.15
+		psl = psl_lib.builtin()
+	end
+
+	if seed == nil then
+		seed = os.time() / (3600 * 24 * 7)
+	end
+	seed = math.floor(seed)  -- convert to int
+
+	return function(query, length)
+		assert(length > 0)
+
+		local domain = kres.dname2str(query:name())
+		if domain == nil then -- invalid data: auto-select first action
+			return 1
+		end
+		if domain:len() > 1 then  --remove trailing dot
+			domain = domain:sub(0, -2)
+		end
+
+		-- do psl lookup for registrable domain
+		local reg_domain = psl:registrable_domain(domain)
+		if reg_domain == nil then  -- fallback to unreg. domain
+			reg_domain = psl:unregistrable_domain(domain)
+			if reg_domain == nil then  -- shouldn't happen: safe fallback
+				return 1
+			end
+		end
+
+		local rand_seed = seed
+		-- create deterministic seed for pseudo-random slice assignment
+		for i = 1, #reg_domain do
+			rand_seed = rand_seed + reg_domain:byte(i)
+		end
+
+		-- use lineral congruential generator with values from ANSI C
+		rand_seed = rand_seed % 0x80000000  -- ensure seed is positive 32b int
+		local rand = (1103515245 * rand_seed + 12345) % 0x10000
+		return 1 + rand % length
+	end
+end
+
+-- Prepare for making an answer from scratch.  (Return the packet for convenience.)
+local function answer_clear(req)
+	-- If we're in postrules, previous resolving might have chosen some RRs
+	-- for inclusion in the answer, so we need to avoid those.
+	-- *_selected arrays are in mempool, so explicit deallocation is not necessary.
+	req.answ_selected.len = 0
+	req.auth_selected.len = 0
+	req.add_selected.len = 0
+
+	-- Let's be defensive and clear the answer, too.
+	local pkt = req.answer
+	pkt:clear_payload()
+	return pkt
+end
+
 function policy.DENY_MSG(msg)
 	if msg and (type(msg) ~= 'string' or #msg >= 255) then
 		error('DENY_MSG: optional msg must be string shorter than 256 characters')
@@ -479,7 +610,7 @@ function policy.DENY_MSG(msg)
 
 	return function (_, req)
 		-- Write authority information
-		local answer = req.answer
+		local answer = answer_clear(req)
 		ffi.C.kr_pkt_make_auth_header(answer)
 		answer:rcode(kres.rcode.NXDOMAIN)
 		answer:begin(kres.section.AUTHORITY)
@@ -493,28 +624,101 @@ function policy.DENY_MSG(msg)
 		return kres.DONE
 	end
 end
+
+local function free_cb(func)
+	func:free()
+end
+
+local debug_logline_cb = ffi.cast('trace_log_f', function (_, msg)
+	jit.off(true, true) -- JIT for (C -> lua)^2 nesting isn't allowed
+	-- msg typically ends with newline
+	io.write(ffi.string(msg))
+end)
+ffi.gc(debug_logline_cb, free_cb)
+
+local debug_logfinish_cb = ffi.cast('trace_callback_f', function (req)
+	jit.off(true, true) -- JIT for (C -> lua)^2 nesting isn't allowed
+	ffi.C.kr_log_req(req, 0, 0, 'dbg',
+		'following rrsets were marked as interesting:\n' ..
+		req:selected_tostring())
+	ffi.C.kr_log_req(req, 0, 0, 'dbg',
+		'answer packet:\n' ..
+		tostring(req.answer))
+end)
+ffi.gc(debug_logfinish_cb, free_cb)
+
+-- log request packet
+function policy.REQTRACE(_, req)
+	ffi.C.kr_log_req(req, 0, 0, 'dbg', 'request packet:\n%s',
+		tostring(req.qsource.packet))
+end
+
+function policy.DEBUG_ALWAYS(state, req)
+	policy.QTRACE(state, req)
+	req:trace_chain_callbacks(debug_logline_cb, debug_logfinish_cb)
+	policy.REQTRACE(state, req)
+end
+
+local debug_stashlog_cb = ffi.cast('trace_log_f', function (req, msg)
+	jit.off(true, true) -- JIT for (C -> lua)^2 nesting isn't allowed
+
+	-- stash messages for conditional logging in trace_finish
+	local stash = req:vars()['policy_debug_stash']
+	table.insert(stash, ffi.string(msg))
+end)
+ffi.gc(debug_stashlog_cb, free_cb)
+
+-- buffer verbose logs and print then only if test() returns a truthy value
+function policy.DEBUG_IF(test)
+	local debug_finish_cb = ffi.cast('trace_callback_f', function (cbreq)
+		jit.off(true, true) -- JIT for (C -> lua)^2 nesting isn't allowed
+		if test(cbreq) then
+			debug_logfinish_cb(cbreq)  -- unconditional version
+			local stash = cbreq:vars()['policy_debug_stash']
+			io.write(table.concat(stash, ''))
+		end
+	end)
+	ffi.gc(debug_finish_cb, function (func) func:free() end)
+
+	return function (state, req)
+		req:vars()['policy_debug_stash'] = {}
+		policy.QTRACE(state, req)
+		req:trace_chain_callbacks(debug_stashlog_cb, debug_finish_cb)
+		policy.REQTRACE(state, req)
+		return
+	end
+end
+
+policy.DEBUG_CACHE_MISS = policy.DEBUG_IF(
+	function(req)
+		return not req:all_from_cache()
+	end
+)
+
 policy.DENY = policy.DENY_MSG() -- compatibility with < 2.0
 
-function policy.DROP(_, _)
+function policy.DROP(_, req)
+	answer_clear(req)
 	return kres.FAIL
 end
 
 function policy.REFUSE(_, req)
-	local answer = req.answer
+	local answer = answer_clear(req)
 	answer:rcode(kres.rcode.REFUSED)
 	answer:ad(false)
 	return kres.DONE
 end
 
 function policy.TC(state, req)
-	local answer = req.answer
-	if answer.max_size ~= 65535 then
-		answer:tc(1) -- ^ Only UDP queries
-		answer:ad(false)
-		return kres.DONE
-	else
+	-- Skip non-UDP queries
+	if req.answer.max_size == 65535 then
 		return state
 	end
+
+	local answer = answer_clear(req)
+	answer:tc(1)
+	answer:ad(false)
+	return kres.DONE
 end
 
 function policy.QTRACE(_, req)
@@ -542,30 +746,6 @@ function policy.evaluate(rules, req, query, state)
 	return
 end
 
--- Top-down policy list walk until we hit a match
--- the caller is responsible for reordering policy list
--- from most specific to least specific.
--- Some rules may be chained, in this case they are evaluated
--- as a dependency chain, e.g. r1,r2,r3 -> r3(r2(r1(state)))
-policy.layer = {
-	begin = function(state, req)
-		-- Don't act on "resolved" cases.
-		if bit.band(state, bit.bor(kres.FAIL, kres.DONE)) ~= 0 then return state end
-
-		req = kres.request_t(req)
-		return policy.evaluate(policy.rules, req, req:current(), state) or
-		       policy.evaluate(policy.special_names, req, req:current(), state) or
-		       state
-	end,
-	finish = function(state, req)
-		-- Don't act on "resolved" cases.
-		if bit.band(state, bit.bor(kres.FAIL, kres.DONE)) ~= 0 then return state end
-
-		req = kres.request_t(req)
-		return policy.evaluate(policy.postrules, req, req:current(), state) or state
-	end
-}
-
 -- Add rule to policy list
 function policy.add(rule, postrule)
 	-- Compatibility with 1.0.0 API
@@ -713,6 +893,8 @@ local private_zones = {
 	'a.e.f.ip6.arpa.',
 	'b.e.f.ip6.arpa.',
 	'8.b.d.0.1.0.0.2.ip6.arpa.',
+	-- RFC8375
+	'home.arpa.',
 }
 policy.todnames(private_zones)
 
@@ -720,6 +902,7 @@ policy.todnames(private_zones)
 policy.rules = {}
 policy.postrules = {}
 policy.special_names = {
+	-- XXX: beware of special_names_optim() when modifying these filters
 	{
 		cb=policy.suffix_common(policy.DENY_MSG(
 			'Blocking is mandated by standards, see references on '
@@ -737,6 +920,7 @@ policy.special_names = {
 				todname('test.'),
 				todname('onion.'),
 				todname('invalid.'),
+				todname('local.'), -- RFC 8375.4
 			}),
 		count=0
 	},
@@ -753,4 +937,44 @@ policy.special_names = {
 	},
 }
 
+-- Return boolean; false = no special name may apply, true = some might apply.
+-- The point is to *efficiently* filter almost all QNAMEs that do not apply.
+local function special_names_optim(req, sname)
+	local qname_size = req.qsource.packet.qname_size
+	if qname_size < 9 then return true end -- don't want to special-case bad array access
+	local root = sname + qname_size - 1
+	return
+		-- .a???. or .t???.
+		(root[-5] == 4 and (root[-4] == 97 or root[-4] == 116))
+		-- .on???. or .in?????. or lo???. or *ost.
+		or (root[-6] == 5 and root[-5] == 111 and root[-4] == 110)
+		or (root[-8] == 7 and root[-7] == 105 and root[-6] == 110)
+		or (root[-6] == 5 and root[-5] == 108 and root[-4] == 111)
+		or (root[-3] == 111 and root[-2] == 115 and root[-1] == 116)
+end
+
+-- Top-down policy list walk until we hit a match
+-- the caller is responsible for reordering policy list
+-- from most specific to least specific.
+-- Some rules may be chained, in this case they are evaluated
+-- as a dependency chain, e.g. r1,r2,r3 -> r3(r2(r1(state)))
+policy.layer = {
+	begin = function(state, req)
+		-- Don't act on "finished" cases.
+		if bit.band(state, bit.bor(kres.FAIL, kres.DONE)) ~= 0 then return state end
+		local qry = req:initial() -- same as :current() but more descriptive
+		return policy.evaluate(policy.rules, req, qry, state)
+			or (special_names_optim(req, qry.sname)
+					and policy.evaluate(policy.special_names, req, qry, state))
+			or state
+	end,
+	finish = function(state, req)
+		-- Optimization for the typical case
+		if #policy.postrules == 0 then return state end
+		-- Don't act on failed cases.
+		if bit.band(state, kres.FAIL) ~= 0 then return state end
+		return policy.evaluate(policy.postrules, req, req:initial(), state) or state
+	end
+}
+
 return policy
