@@ -1,6 +1,7 @@
+-- SPDX-License-Identifier: GPL-3.0-or-later
 -- Module interface
 local ffi = require('ffi')
-local prefixes = {}
+local prefixes_global = {}
 
 -- Create subnet prefix rule
 local function matchprefix(subnet, addr)
@@ -28,7 +29,7 @@ end
 
 -- Add subnet prefix rewrite rule
 local function add_prefix(subnet, addr)
-	table.insert(prefixes, matchprefix(subnet, addr))
+	table.insert(prefixes_global, matchprefix(subnet, addr))
 end
 
 -- Match IP against given subnet or record owner
@@ -61,11 +62,10 @@ local function renumber_record(tbl, rr)
 end
 
 -- Renumber addresses based on config
-local function rule()
+local function rule(prefixes)
 	return function (state, req)
 		if state == kres.FAIL then return state end
-		req = kres.request_t(req)
-		local pkt = kres.pkt_t(req.answer)
+		local pkt = req.answer
 		-- Only successful answers
 		local records = pkt:section(kres.section.ANSWER)
 		local ancount = #records
@@ -83,7 +83,7 @@ local function rule()
 			end
 		end
 		-- If not rewritten, chain action
-		if not changed then return end
+		if not changed then return state end
 		-- Replace section if renumbering
 		local qname = pkt:qname()
 		local qclass = pkt:qclass()
@@ -120,7 +120,7 @@ end
 
 -- Layers
 M.layer = {
-	finish = rule(),
+	finish = rule(prefixes_global),
 }
 
 return M
