@@ -1,22 +1,10 @@
 /*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 /**
  * @file stats.c
- * @brief Storage for various counters and metrics from query resolution. 
+ * @brief Storage for various counters and metrics from query resolution.
  *
  * You can either reuse this module to compute statistics or store custom metrics
  * in it via the extensions.
@@ -28,6 +16,7 @@
 #include <ccan/json/json.h>
 #include <contrib/cleanup.h>
 #include <arpa/inet.h>
+#include <lua.h>
 
 #include "lib/layer/iterate.h"
 #include "lib/rplan.h"
@@ -35,18 +24,13 @@
 #include "lib/layer.h"
 #include "lib/resolve.h"
 
-/** @internal Compatibility wrapper for Lua < 5.2 */
-#if LUA_VERSION_NUM < 502
-#define lua_rawlen(L, obj) lua_objlen((L), (obj))
-#endif
-
 /* Defaults */
 #define VERBOSE_MSG(qry, ...) QRVERBOSE(qry, "stat",  __VA_ARGS__)
 #define FREQUENT_PSAMPLE  10 /* Sampling rate, 1 in N */
 #ifdef LRU_REP_SIZE
  #define FREQUENT_COUNT LRU_REP_SIZE /* Size of frequent tables */
 #else
- #define FREQUENT_COUNT  5000 /* Size of frequent tables */ 
+ #define FREQUENT_COUNT  5000 /* Size of frequent tables */
 #endif
 #ifndef UPSTREAMS_COUNT
  #define UPSTREAMS_COUNT  512 /* Size of recent upstreams */
@@ -60,6 +44,8 @@
 	X(answer,aa) X(answer,tc) X(answer,rd) X(answer,ra) X(answer, ad) X(answer,cd) \
 	X(answer,edns0) X(answer,do) \
 	X(query,edns) X(query,dnssec) \
+	X(request,total) X(request,udp) X(request,tcp) \
+	X(request,dot) X(request,doh) X(request,internal) \
 	X(const,end)
 
 enum const_metric {
@@ -186,6 +172,33 @@ static int collect_rtt(kr_layer_t *ctx, knot_pkt_t *pkt)
 	return ctx->state;
 }
 
+static int collect_transport(kr_layer_t *ctx)
+{
+	struct kr_request *req = ctx->req;
+	struct kr_module *module = ctx->api->data;
+	struct stat_data *data = module->data;
+
+	stat_const_add(data, metric_request_total, 1);
+	if (req->qsource.dst_addr == NULL) {
+		stat_const_add(data, metric_request_internal, 1);
+		return ctx->state;
+	}
+
+	/**
+	 * Count each transport only once,
+	 * i.e. DoT does not count as TCP.
+	 */
+	if (req->qsource.flags.http)
+		stat_const_add(data, metric_request_doh, 1);
+	else if (req->qsource.flags.tls)
+		stat_const_add(data, metric_request_dot, 1);
+	else if (req->qsource.flags.tcp)
+		stat_const_add(data, metric_request_tcp, 1);
+	else
+		stat_const_add(data, metric_request_udp, 1);
+	return ctx->state;
+}
+
 static int collect(kr_layer_t *ctx)
 {
 	struct kr_request *param = ctx->req;
@@ -430,25 +443,29 @@ static char* dump_upstreams(void *env, struct kr_module *module, const char *arg
 	return ret;
 }
 
-/*
- * Module implementation.
- */
-
-KR_EXPORT
-const kr_layer_api_t *stats_layer(struct kr_module *module)
-{
-	static kr_layer_api_t _layer = {
-		.consume = &collect_rtt,
-		.finish = &collect,
-	};
-	/* Store module reference */
-	_layer.data = module;
-	return &_layer;
-}
-
 KR_EXPORT
 int stats_init(struct kr_module *module)
 {
+	static kr_layer_api_t layer = {
+		.consume = &collect_rtt,
+		.finish = &collect,
+		.begin = &collect_transport,
+	};
+	/* Store module reference */
+	layer.data = module;
+	module->layer = &layer;
+
+	static const struct kr_prop props[] = {
+	    { &stats_set,     "set", "Set {key, val} metrics.", },
+	    { &stats_get,     "get", "Get metrics for given key.", },
+	    { &stats_list,    "list", "List observed metrics.", },
+	    { &dump_frequent, "frequent", "List most frequent queries.", },
+	    { &clear_frequent,"clear_frequent", "Clear frequent queries log.", },
+	    { &dump_upstreams,  "upstreams", "List recently seen authoritatives.", },
+	    { NULL, NULL, NULL }
+	};
+	module->props = props;
+
 	struct stat_data *data = malloc(sizeof(*data));
 	if (!data) {
 		return kr_error(ENOMEM);
@@ -482,21 +499,6 @@ int stats_deinit(struct kr_module *module)
 	return kr_ok();
 }
 
-KR_EXPORT
-struct kr_prop *stats_props(void)
-{
-	static struct kr_prop prop_list[] = {
-	    { &stats_set,     "set", "Set {key, val} metrics.", },
-	    { &stats_get,     "get", "Get metrics for given key.", },
-	    { &stats_list,    "list", "List observed metrics.", },
-	    { &dump_frequent, "frequent", "List most frequent queries.", },
-	    { &clear_frequent,"clear_frequent", "Clear frequent queries log.", },
-	    { &dump_upstreams,  "upstreams", "List recently seen authoritatives.", },
-	    { NULL, NULL, NULL }
-	};
-	return prop_list;
-}
-
 KR_MODULE_EXPORT(stats)
 
 #undef VERBOSE_MSG
