@@ -1,7 +1,8 @@
 #!/bin/sh -e
+# SPDX-License-Identifier: GPL-3.0-or-later
 # Create a distribution tarball, like 'make dist' from autotools.
 cd "$(git rev-parse --show-toplevel)"
-ver=$(git describe | sed 's/^v//' | sed 's/-/\./g')
+ver="$(git describe | sed 's/^v//')"
 test 0 -ne $(git status --porcelain | wc -l) && \
 	echo "Git working tree is dirty, make it clean first" && \
 	exit 1
@@ -11,5 +12,5 @@ git submodule status --recursive | grep -q '^[^ ]' && \
 
 # 'git ls-files --recurse-submodules' works only if modules are initialized
 name="knot-resolver-$ver"
-tar caf "$name.tar.xz" -h --no-recursion --transform "s|^|$name/|" -- $(git ls-files --recurse-submodules)
+tar caf "$name.tar.xz" --no-recursion --transform "s|^|$name/|" -- $(git ls-files --recurse-submodules)
 echo "$name.tar.xz"
