@@ -1,25 +1,16 @@
 #!/bin/bash
+# SPDX-License-Identifier: GPL-3.0-or-later
 set -o errexit -o nounset -o xtrace
 
-# Run with -s to include *.symbols files.
+cd "$(dirname ${0})/.."
+pkgdir="build_dist/meson-dist"
 
 package=knot-resolver
-withsymbols=false
-
-while getopts "s" o; do
-	case "${o}" in
-		s)
-			withsymbols=true
-			;;
-		*)
-			;;
-	esac
-done
-shift $((OPTIND-1))
 
 
-cd "$(git rev-parse --show-toplevel)"
+pushd ${pkgdir}
 version=$(ls ${package}*.tar.xz | sed "s/${package}-\(.*\).tar.xz/\1/")
+popd
 
 # Check version for invalid characters
 if [[ $(echo "${version}" | grep '^[[:alnum:].]$') -ne 0 ]]; then
@@ -28,7 +19,7 @@ if [[ $(echo "${version}" | grep '^[[:alnum:].]$') -ne 0 ]]; then
 fi
 
 # Fill in VERSION field in distribution specific files
-files="distro/rpm/${package}.spec distro/deb/changelog distro/arch/PKGBUILD"
+files="distro/rpm/${package}.spec distro/deb/changelog distro/arch/PKGBUILD distro/turris/Makefile"
 for file in ${files}; do
 	sed -i "s/__VERSION__/${version}/g" "${file}"
 done
@@ -36,18 +27,13 @@ done
 # Rename archive to debian format
 pkgname="${package}-${version}"
 debname="${package}_${version}.orig"
-mv "${pkgname}.tar.xz" "${debname}.tar.xz"
+cp "${pkgdir}/${pkgname}.tar.xz" "${debname}.tar.xz"
 
 # Prepare clean debian-specific directory
 tar -xf "${debname}.tar.xz"
 pushd "${pkgname}" > /dev/null
 cp -arL ../distro/deb debian
 
-# Optionally remove symbols file
-if [ "$withsymbols" = false ]; then
-    rm -f debian/*.symbols
-fi
-
 # Create debian archive and dsc
 dpkg-source -b .
 popd > /dev/null
