@@ -1,3 +1,4 @@
+# SPDX-License-Identifier: GPL-3.0-or-later
 from contextlib import contextmanager
 import random
 import ssl
@@ -98,18 +99,18 @@ def ping_alive(sock, msgid=None):
 
 @contextmanager
 def expect_kresd_close(rst_ok=False):
-    with pytest.raises(BrokenPipeError, message="kresd didn't close the connection"):
+    with pytest.raises(BrokenPipeError):
         try:
             time.sleep(0.2)  # give kresd time to close connection with TCP FIN
             yield
         except ConnectionResetError:
             if rst_ok:
                 raise BrokenPipeError
-            else:
-                pytest.skip("kresd closed connection with TCP RST")
+            pytest.skip("kresd closed connection with TCP RST")
+        pytest.fail("kresd didn't close the connection")
 
 
-def make_ssl_context(insecure=False, verify_location=None):
+def make_ssl_context(insecure=False, verify_location=None, extra_options=None):
     # set TLS v1.2+
     context = ssl.SSLContext(ssl.PROTOCOL_TLS)
     context.options |= ssl.OP_NO_SSLv2
@@ -117,6 +118,10 @@ def make_ssl_context(insecure=False, verify_location=None):
     context.options |= ssl.OP_NO_TLSv1
     context.options |= ssl.OP_NO_TLSv1_1
 
+    if extra_options is not None:
+        for option in extra_options:
+            context.options |= option
+
     if insecure:
         # turn off certificate verification
         context.check_hostname = False
